#ident "@(#)ddefs.c	35.4"
/*
 *
 *	ddefs: manages disk defs information
 *		also permits editing of a disk's vid/config block
 *
 *	Three modes:
 *		1. To create a new disk definition.
 *		2. To edit an existing disk definition.
 *		3. To read an existing disk definition. (read-only)
 *		4. To print an existing disk definition.
 *		5. To convert an old style entry.
 *
 *	New style diskdefs file:
 *
 *		bytes		contents
 *		0 to 255	volumeid info with:
 *				vid_dta == 0 if no diag
 *				vid_dta != 0 if diag
 *				vid_btl == maxbad
 *				vid_strat == bad spot strategy
 *				vid_cas == root filesystem size
 *				vid_vol	== /usr filesystem size
 *				vid_osa	== /usr filesystem slice
 *				vid_oss == swap block size
 *				vid_osl == swap slice
 *				vid_ose == reserved blocks
 *		256 to 511	configuration info
 *		512 to x	alternates (int's) (if any)
 *		x to y		format program (ascii)
 */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/dk.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <signal.h>

#define	DEBUG
#undef	DEBUG
#define	DEBUGC
#undef	DEBUGC

char *Usage[] = {
	"[-d ddefsdir] -n [diskname]",			/* create new */
	"[-d ddefsdir] -e diskname",			/* edit */
	"[-d ddefsdir] -r diskname",			/* read-only */
	"[-d ddefsdir] -p diskname",			/* print */
	"[-d ddefsdir] -c [diskdefs] [-a altsdir]",	/* convert old */
#ifdef	DDEFSINIT
	"[-d ddefsdir] -i",				/* initialize */
#endif
	NULL,
};

#ifdef	DDEFSINIT
#define	OPTS	"a:d:ce:np:r:i"
#else
#define	OPTS	"a:d:ce:np:r:"
#endif

#define	DK_NULL		0
#define	DK_NEW		1
#define	DK_EDIT		2
#define	DK_RDONLY	3
#define	DK_PRINT	4
#define	DK_CONVERT	5
#ifdef	DDEFSINIT
#define	DK_INIT		6
#endif

/*
 *	globals
 */

int	action	= DK_NULL;		/* current action */
char	*altsdir = NULL;		/* name of alts directory */
char	*ddefsdir = NULL;		/* name of ddefs directory */
char	*diskname = NULL;		/* disk type name */
char	*diskdefs = NULL;		/* disk defs file */
int	chgd = 0;			/* file changed */

int	diskdev = 0;			/* edit file is disk device */

/*
 *	alternates list
 */

struct alts {				/* linked list of alternates */
	struct alts *a_next;
	unsigned a_altb;		/* begin of alternate sequence */
	unsigned a_alte;		/* end of alternate sequence */
} *alts = NULL;

/*
 *	work areas
 */

char	fname[BUFSIZ];			/* filename (temporary) */
char	dname[BUFSIZ];			/* disk type name */
char	d1name[BUFSIZ];			/* disk type name */
char	fmtprog[BUFSIZ];		/* format program */
char	comment[BUFSIZ];		/* comment line */
char	rdprog[BUFSIZ];			/* read bad list program */

/*
 *	volume ID and config blocks ... see dk.h
 */

struct	volumeid vid;
struct	dkconfig ios;

/*
 *	Edit commands
 */

#define	C_COMMENT	0
#define	C_TYPE		1
#define	C_FORMAT	2
#define	C_DIAGNOSTIC	3
#define	C_STRATEGY	4
#define	C_MAXBAD	5
#define	C_SECTORS	6
#define	C_SECSIZE	7
#define	C_SECPERTRK	8
#define	C_CYLINDERS	9
#define	C_HEADS		10
#define	C_PRECOMP	11
#define	C_INTERLEAVE	12
#define	C_SPIRAL	13
#define	C_STEPRATE	14
#define	C_STARTING	15
#define	C_ECC		16
#define	C_ATRMASK	17
#define	C_EXTAMASK	18
#define	C_ATRWORD	19
#define	C_EXTAWORD	20
#define	C_GAP1		21
#define	C_GAP2		22
#define	C_GAP3		23
#define	C_GAP4		24
#define	C_CATR		25
#define	C_USS		26
#define	C_SECSLIP	27
#define	C_SLICES	28
#define	C_ROOTOFFSET	29
#define	C_ROOTSIZE	30
#define	C_USRSIZE	31
#define	C_USRSLICE	32
#define	C_SWAPSIZE	33
#define	C_SWAPSLICE	34
#define	C_RESERVED	35
#define	C_ALTERNATES	36
#define	C_RDPROG	37

/*
 *	Edit information
 */

#define	E_NULL		0
#define	E_CHANGED	(1<<0)	/* field has been changed */
#define	E_DSKNOACC	(1<<1)	/* field is accessible on disk */
#define	E_DSKCHG	(1<<2)	/* field is changeable on disk */

struct edit {
	short	e_flags;
	char	*e_prompt;
	char	*e_fmt;
	char	*e_help;
} edit[] = {

{/* 0 */	E_CHANGED|E_DSKNOACC, "Comment", "%s",
"This information is for general comments. It cannot contain more than\n\
one line. Usually, the comment information is a verbal description of the\n\
drive type supported by this file. If no comment is wanted, type `none'.\n"
},

{/* 1 */	E_DSKCHG, "Disk type", "%d",
"This value is any unique integer that is used to identify the disk\n\
drive type. Each size drive on each controller should have a\n\
different type.\n"
},

{/* 2 */	E_DSKNOACC, "Format command", "%s",
"The format program, if specified, is called by dinit(1M) to format\n\
a disk drive. The format program line should specify all options necessary\n\
to format the disk. If the driver is a new style driver (MVME321), then\n\
dinit(1M) can issue the format requests directly. In this case, enter `none'.\n"
},

{/* 3 */	E_NULL, "Diagnostic tracks", "%s",
"To reserve diagnostic tracks type `yes'. If not, type `no'.\n"
},

{/* 4 */	E_NULL, "Bad spot strategy", "%s",
"If the controller only supports perfect media, type `perfect'.\n\
If bad TRACK replacement is done by the SYSTEM V/68 bad track replacement\n\
software, type `software'. If the controller automatically performs bad\n\
TRACK replacement, type `hardware'. If the controller supports automatic\n\
bad SECTOR replacement, and requires the cylinder, head, and sector of\n\
each bad SECTOR, then type `sector'. If the controller supports automatic\n\
bad SPOT replacement, and requires the cylinder, head, and byte offset of\n\
each bad SPOT, then type `spot'.\n"
},

{/* 5 */	E_DSKNOACC, "Maximum number of bad spots", "%d",
"This value is the maximum number of bad spots expected on a disk of\n\
this type. This many alternate spots (sectors or tracks) will be allocated\n\
for this drive.\n"
},

{/* 6 */	E_DSKCHG, "Number of sectors", "%d",
"This value is the total number of sectors on the disk drive.\n"
},

{/* 7 */	E_DSKCHG, "Sector size (in bytes)", "%d",
"This value is the disk sector size (specified in bytes). Under current\n\
SYSTEM V/68 implementations, the sector size must be 128, 256 or 512 bytes.\n"
},

{/* 8 */	E_DSKCHG, "Sectors per track", "%d",
"This value is the number of usable sectors per track on the formatted \n\
media.  This value should NOT include sectors reserved for 'sector slip'\n\
or 'runt' sectors.\n"
},

{/* 9 */	E_DSKCHG, "Cylinders", "%d",
"This value is the total number of cylinders on the disk media.\n"
},

{/* 10 */	E_DSKCHG, "Heads", "%d",
"This value is the number of read/write heads on the drive. It is\n\
equivalent to the number of tracks per cylinder.\n"
},

{/* 11 */	E_DSKCHG, "Precompensation cylinder", "%d",
"This value is the disk cylinder number to start write precompensation.\n"
},

{/* 12 */	E_DSKCHG, "Sector interleave", "%d",
"This value is the sector interleave factor used during disk formatting.\n\
For no interleave (or one-to-one interleave), type `1'. Some controllers\n\
will automatically select an appropriate interleave factor when given an\n\
interleave value of `0'.\n"
},

{/* 13 */	E_DSKCHG, "Spiral offset", "%d",
"This value is the spiral offset applied when formatting disks. If no\n\
spiral offset is wanted, then type `0'.\n"
},

{/* 14 */	E_DSKCHG, "Step rate", "%d",
"This value is the seek step rate used when accessing the disk. Some\n\
controllers will automatically select an appropriate step rate when given\n\
a step rate of `0'.\n"
},

{/* 15 */	E_DSKCHG, "Starting head number", "%d",
"This value is the starting head number of the drive. Most drives and\n\
controllers require a starting head number of zero.\n"
},

{/* 16 */	E_DSKCHG, "ECC error length", "%d",
"This value is the error correcting code data burst length.\n"
},

{/* 17 */	E_DSKCHG, "Attributes mask (hex)", "%x",
"This value is the disk attributes mask. Bits in this mask determine\n\
which bits in the attributes word are valid.\n"
},

{/* 18 */	E_DSKCHG, "Extended attributes mask (hex)", "%x",
"This value is the extended attributes mask.\n"
},

{/* 19 */	E_DSKCHG, "Attributes word (hex)", "%x",
"This value is the disk attributes word. The (hexadecimal) bit\n\
definitions are:\n\n\
NAME	BIT	FIELD USE			BIT OFF		BIT ON\n\
ATWAS	0x0400	Alternate sectors?		no		yes\n\
ATWFS	0x0200	Floppy size			5.25\"		8\"\n\
ATWPC	0x0100	Precomp style			pre-write	post-read\n\
ATWSK	0x0080	Seek after head change? 	no		yes\n\
ATWDD	0X0040	Track density of drive		single		double\n\
ATWEN	0X0020	Encoding method			FM		MFM\n\
ATWDT	0X0010	Disk type			floppy		hard\n\
ATWSN	0X0008	Sector Numbering		Motorola	IBM\n\
ATWDS	0x0004	Number of sides			single		double\n\
ATWTD	0x0002	Track density of floppy		8\" floppy	5.25\" floppy\n\
ATWMF	0x0001	Data density of medium		single		double\n"
},

{/* 20 */	E_DSKCHG, "Extended attributes word (hex)", "%x",
"This value is the extended attributes word.\n"
},

{/* 21 */	E_DSKCHG, "Gap byte 1 (hex)", "%x",
"This value is the first `gap byte' required for formatting a disk.\n\
This parameter is controller and drive specific and may not be used by\n\
some controllers.\n"
},

{/* 22 */	E_DSKCHG, "Gap byte 2 (hex)", "%x",
"This value is the second `gap byte' required for formatting a disk.\n\
This parameter is controller and drive specific and may not be used by\n\
some controllers.\n"
},

{/* 23 */	E_DSKCHG, "Gap byte 3 (hex)", "%x",
"This value is the third `gap byte' required for formatting a disk.\n\
This parameter is controller and drive specific and may not be used by\n\
some controllers.\n"
},

{/* 24 */	E_DSKCHG, "Gap byte 4 (hex)", "%x",
"This value is the fourth `gap byte' required for formatting a disk.\n\
This parameter is controller and drive specific and may not be used by\n\
some controllers.\n"
},

{/* 25 */	E_DSKCHG, "Controller attribute (hex)", "%x",
"This value is a controller and drive specific information parameter and may\n\
not be used by some controllers.\n"
},

{/* 26 */	E_DSKCHG, "Unformatted sector size", "%d",
"This value is the unformatted sector size on the disk including the\n\
headers, gaps, ECC, etc., and data. This parameter is controller and drive\n\
specific, and may not be used by some controllers.\n"
},

{/* 27 */	E_DSKCHG, "Sector slip count", "%d",
"This value is the sector slip count used while formatting to implement\n\
controller supported sector slipping. This count is the number of `slip'\n\
sectors per track.\n"
},

{/* 28 */	E_NULL, "Slice count", "%d",
"This value is the dynamic slice count. If zero, then the driver\n\
will not use dynamic slicing. Legal nonzero slice counts are 8, 16,\n\
32, 64, and 128. Some controllers only support a subset of these legal\n\
slice counts.\n"
},

{/* 29 */	E_NULL, "Root filesystem offset", "%d",
"This value is the 1024-byte block offset of the SYSTEM V/68 root filesystem.\n"
},

{/* 30 */	E_DSKNOACC, "Root filesystem size", "%d",
"This value, if non-zero, is the size of the root filesystem (in 512-byte\n\
blocks) created on slice zero of the drive after it is formatted. If no\n\
filesystem is desired, then type `0'.\n"
},

{/* 31 */	E_DSKNOACC, "/usr filesystem size", "%d",
"This value, if non-zero, is the size of the /usr filesystem (in 512-byte\n\
blocks) created in slice one or two after the drive is formatted. If no\n\
/usr filesystem is desired then type `0'. The /usr filesystem is created\n\
only if a root filesystem is also created.\n"
},

{/* 32 */	E_DSKNOACC, "/usr filesystem slice", "%d",
"This value specifies the slice for the /usr filesystem. This value must not\n\
be zero but may coincide with the swap slice.\n"
},

{/* 33 */	E_DSKNOACC, "Swap size", "%d",
"This value, if non-zero, specifies the size of the system swap space (in\n\
512-byte blocks). The swap space will be placed following the root\n\
filesystem if the swap slice is set to zero. If the swap slice and the\n\
/usr filesystem slice are both one, then swap will follow the /usr filesystem\n\
in slice one.\n"
},

{/* 34 */	E_DSKNOACC, "Swap slice", "%d",
"This value specifies the swap slice. It may be set to zero (to share\n\
the root filesystem slice, or to one (to occupy its own slice or share with\n\
the /usr filesystem), or to two (to occupy its own slice following the\n\
/usr filesystem slice).\n"
},

{/* 35 */	E_DSKNOACC, "End-of-disk reserved area", "%d",
"This value is the number of blocks reserved for system use at the end\n\
of the disk. This area may be used for bad track alternates or for\n\
diagnostic tracks.\n"
},

{/* 36 */	E_DSKNOACC, "Alternates list", "%s",
"To add to the alternates list, type `add #[-#]'. To delete from the\n\
alternates list, type `delete #[-#]'. The `#' is any non-zero number.\n\
The sequence `#-#' specifies a range of nonzero numbers (including the\n\
end-points. To have no alternates, type `none'. The value printed in\n\
parentheses is the total number of alternates in the list.\n"
},

{/* 37 */	E_DSKNOACC, "Read bad list command", "%s",
"This program, if specified, is called by dinit(1M) to read the \n\
manufacturers defect list from the disk drive. The command line should \n\
specify all options necessary.  Only ESDI drives currently support this \n\
command.\n"
},

{/* end */	NULL, NULL, 0, NULL
},

};

/*
 *	usage: print usage information and exit
 */

usage(err)
{	char **u;

	fprintf( stderr, "Usage:" );

	for( u = Usage; *u != NULL; u++ )
		fprintf( stderr, "\tddefs %s\n", *u );

	exit(err);
}

/*
 *	main: perform ddefs
 */

main( argc, argv )
register argc;
register char *argv[];
{	register c;
	extern	char	*optarg;
	extern	int	optind;

	while(( c = getopt( argc, argv, OPTS )) != EOF ) {
		switch(c) {
		case 'a':	/* alts dir */
			if( action != DK_CONVERT )
				usage(EINVAL);
			altsdir = optarg;
			break;
		case 'c':	/* convert */
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_CONVERT;
			diskdefs = NULL;
			break;
		case 'd':	/* ddefs dir */
			ddefsdir = optarg;
			break;
		case 'e':	/* edit */
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_EDIT;
			diskname = optarg;
			break;
		case 'n':	/* new */
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_NEW;
			diskname = NULL;
			break;
		case 'p':	/* print */
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_PRINT;
			diskname = optarg;
			break;
		case 'r':
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_RDONLY;
			diskname = optarg;
			break;
#ifdef	DDEFSINIT
		case 'i':
			if( action != DK_NULL )
				usage(EINVAL);
			action = DK_INIT;
			break;
#endif
		default:
			usage(EINVAL);
		}
	}

	if( action == DK_NULL )	/* needs -n, -e, -p, or -c */
		usage(EINVAL);

	if(( action == DK_NEW ) || ( action == DK_CONVERT )) {
		if( optind == argc-1 )
			diskname = argv[optind++];
	}

	if( optind != argc ) {
		fprintf( stderr, "ddefs: unexpected arguments\n" );
		usage(E2BIG);
	}

	/*
	 *	set default directories
	 */

	if( ddefsdir == NULL )
		ddefsdir = "/etc/dskdefs";

	if( altsdir == NULL )
		altsdir = "/etc/diskalts";

	if( diskdefs == NULL )
		diskdefs = "/etc/diskdefs";

#ifdef	DDEFSINIT

	/*
	 *	If initializing, only initialize
	 */

	if( action == DK_INIT ) {
		initdef();
		exit(0);
	}

#endif

	/*
	 *	If converting, only convert
	 */

	if( action == DK_CONVERT ) {
		convdefs(diskdefs);
		exit(0);
	}

	if( action == DK_PRINT ) {
		if( ! existdef(diskname) ) {
			fprintf( stderr, "ddefs: %s not found\n", fname );
			usage();
		}

		readdef(diskname);
		printdef(diskname);
		exit(0);
	}

	if( action == DK_NEW ) {

		/* create new name */
		while( diskname == NULL ) {
			fprintf( stdout, "New disk type name: " );
			fflush(stdout);
			if( fgets( dname, BUFSIZ, stdin ) == NULL )
				continue;

			if( dname[0] == '\n' )
				continue;

			dname[strlen(dname)-1] = '\0';	/* get rid of `\n' */

			if( strlen(dname) <= DIRSIZ ) {
				diskname = dname;	/* use created name */
				break;
			}

			fprintf( stderr, "ddefs: name too long\n" );
			fprintf( stderr,
			 "ddefs: maximum name length = %d characters\n",DIRSIZ);
		}
	}

	if( existdef(diskname) ) {		/* if the ddefs file exists */
		if( action == DK_NEW ) {
			fprintf( stderr, "ddefs: %s already exists\n", fname );
			fprintf( stderr, "ddefs: using edit mode\n");
			action = DK_EDIT;
		}

	} else if( action != DK_NEW ) {
	
		fprintf( stderr, "ddefs: %s not found\n", fname );

		if( action == DK_RDONLY ) 	/* -r needs name */
			usage();

		fprintf( stderr, "ddefs: creating new definition\n" );
		action = DK_NEW;
	}

	/*
	 *	if a new file is being created, clear the data structures
	 */

	if( action == DK_NEW ) {
		memset( &vid, 0, sizeof(vid) );
		memset( &ios, 0, sizeof(ios) );
		ios.ios_rec = 256;	/* default logical sector size */
		alts = NULL;
		fmtprog[0] = 0;
		comment[0] = 0;
		rdprog[0]  = 0;

		for(;;) {
			fprintf( stdout, "Template disk type name: " );
			fflush(stdout);
			if( fgets( d1name, BUFSIZ, stdin ) == NULL )
				continue;

			if( d1name[0] == '\n' )
				break;

			d1name[strlen(d1name)-1] = '\0';

			if( existdef(d1name) ) {
				readdef(d1name);
				setchanged();
				break;
			}

			fprintf( stderr, "ddefs: %s not found\n", fname );
		}

	/*
	 *	if editing or read-only, read ddefs file
	 */

	} else
		readdef(diskname);

	/*
	 *	enter edit mode ... no changes permitted if action == DK_RDONLY
	 */

	editdef(diskname);
	exit(0);
}

/*
 *	convdefs: convert defs from old style to new style
 */

convdefs(name)
register char *name;
{	register FILE *fp;
	register n;
	static char line[BUFSIZ];
	register lncnt;

	/*
	 *	set default diskdefs name
	 */

	if(( fp = fopen( name, "r" )) == NULL ) {
		fprintf( stderr, "ddefs: %s not found\n", name );
		exit(errno);
	}

	/*
	 *	now point name at dname for the ddefs name
	 */

	name = dname;

	/*
	 *	read the diskdefs file
	 */

	for( lncnt=1; fgets( line, BUFSIZ, fp ) != NULL; lncnt++ ) {

		/*
		 *	make sure that the line isn't too long
		 */

		if( line[strlen(line)-1] != '\n' ) {
			fprintf( stderr, "ddefs: line %d too long\n", lncnt );

			/* read and discard rest of line */
			while( fgets(line, BUFSIZ, fp) != NULL && 
				line[strlen(line)-1] != '\n' );
			continue;
		}
		
		/*
		 *	comment lines start with `*'
		 */

		if( line[0] == '*' )		/* comment line? */
			continue;		/* yes */

		/*
		 *	scan the line for data
		 */

		n = scanline(line);

		/*
		 *	each line must have at least 20 entries and no
		 *		more than 26 entries
		 */

		if(( n < 20 ) || ( n > 26 )) {
			fprintf( stderr, "ddefs: line %d invalid\n", lncnt );
			continue;
		}

		/*
		 *	read the alternates file
		 */

		rdalts(name);			/* read alternate data */

		/*
		 *	write out the data to the ddefs file
		 */

		wrtdef(name,1);			/* write the data */
	}

	fclose(fp);
}

/*
 *	scanline: scan a diskdefs line
 */

scanline(line)
register char *line;
{	char	diag[BUFSIZ/4];
	char	strat[BUFSIZ/4];
	short	ios_spt = 0;
	short	ios_hds = 0;
	short	ios_sof = 0;
	short	ios_ilv = 0;
	short	ios_sr = 0;
	short	ios_gpb1 = 0;
	short	ios_gpb2 = 0;
	short	ios_gpb3 = 0;
	short	ios_gpb4 = 0;
	short	ios_eatw = 0;
	register n;

	/*
	 *	clear the data structures
	 */

	memset( &vid, 0, sizeof(vid) );
	memset( &ios, 0, sizeof(ios) );
	ios.ios_rec = 256;		/* default logical sector size */
	fmtprog[0] = '\0';
	comment[0] = '\0';
	rdprog[0]  = '\0';

		/*	   0  1    2      3  4  5   6  7  8   9		*/
	n = sscanf( line, "%s %hd '%[^']' %s %s %hd %d %d %hd %hd\
			   %hd %hd %hd %hd %hd %hd %hd %hd %hd %hd\
			   %d %hd %hd %hd %hd %hd",
		dname,			/* 0 - disk name */
		&vid.vid_typ,		/* 1 - disk type */
		fmtprog,		/* 2 - format program */
		diag,			/* 3 - diagnostic indicator */
		strat,			/* 4 - bad spot strategy */
		&vid.vid_maxbad,	/* 5 - maximum bad spots */
		&vid.vid_rfs,		/* 6 - root fs offset */
		&ios.ios_rsz,		/* 7 - number of sectors */
		&ios.ios_psm,		/* 8 - physical sector size */
		&ios_spt,		/* 9 - sectors per track */
		&ios.ios_trk,		/* 10 - cylinders */
		&ios_hds,		/* 11 - heads */
		&ios.ios_pcom,		/* 12 - precomp cylinder */
		&ios_ilv,		/* 13 - interleave */
		&ios_sof,		/* 14 - spiral offset */
		&ios_sr,		/* 15 - step rate */
		&ios.ios_shn,		/* 16 - starting head number */
		&ios.ios_ecc,		/* 17 - ecc error length */
		&ios.ios_atm,		/* 18 - attributes mask */
		&ios.ios_atw,		/* 19 - attributes word */
		&vid.vid_rsize, 	/* 20 - size of root filesystem */
		&ios_gpb1,		/* 21 - gap byte 1 */
		&ios_gpb2,		/* 22 - gap byte 2 */
		&ios_gpb3,		/* 23 - gap byte 3 */
		&ios_gpb4,		/* 24 - gap byte 4 */
		&ios_eatw );		/* 25 - extended attribute word */

	/*
	 *	if( vid_dta ) then diagnostics tracks enabled
	 */

	vid.vid_dta = ((diag[0] == 'y') || (diag[0] == 'Y'));

	chgtolower(strat);

	if( substr( strat, "perfect" ) )
		vid.vid_strat = DKPERFECT;

	else if( substr( strat, "software" ) )
		vid.vid_strat = DKSOFTWARE;

	else if( substr( strat, "hardware" ) )
		vid.vid_strat = DKHARDWARE;

	else if( substr( strat, "sector" ) )
		vid.vid_strat = DKHARDSEC;

	else if( substr( strat, "spot" ) )
		vid.vid_strat = DKHARDSPOT;

	else {
		fprintf( stderr,
			"ddefs: invalid bad spot strategy, using `PERFECT'\n");
		vid.vid_strat = DKPERFECT;
	}


	ios.ios_spt = ios_spt;		/* assign byte field */
	ios.ios_hds = ios_hds;		/* assign byte field */
	ios.ios_sof = ios_sof;		/* assign byte field */
	ios.ios_ilv = ios_ilv;		/* assign byte field */
	ios.ios_sr = ios_sr;		/* assign byte field */
	ios.ios_gpb1 = ios_gpb1;	/* assign byte field */
	ios.ios_gpb2 = ios_gpb2;	/* assign byte field */
	ios.ios_gpb3 = ios_gpb3;	/* assign byte field */
	ios.ios_gpb4 = ios_gpb4;	/* assign byte field */
	ios.ios_eatw = ios_eatw;	/* assign byte field */

#ifdef	DEBUG
	printf("%s", line );
	printf( "%s %d '%s' %s %s %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n",
		dname,			/* 0 - disk name */
		vid.vid_typ,		/* 1 - disk type */
		fmtprog,		/* 2 - format program */
		diag,			/* 3 - diagnostic indicator */
		strat,			/* 4 - bad spot strategy */
		vid.vid_maxbad,		/* 5 - maximum bad spots */
		vid.vid_rfs,		/* 6 - root fs offset */
		ios.ios_rsz,		/* 7 - number of sectors */
		ios.ios_psm,		/* 8 - physical sector size */
		ios.ios_spt,		/* 9 - sectors per track */
		ios.ios_trk,		/* 10 - cylinders */
		ios.ios_hds,		/* 11 - heads */
		ios.ios_pcom,		/* 12 - precomp cylinder */
		ios.ios_ilv,		/* 13 - interleave */
		ios.ios_sof,		/* 14 - spiral offset */
		ios.ios_sr,		/* 15 - step rate */
		ios.ios_shn,		/* 16 - starting head number */
		ios.ios_ecc,		/* 17 - ecc error length */
		ios.ios_atm,		/* 18 - attributes mask */
		ios.ios_atw,		/* 19 - attributes word */
		vid.vid_rsize, 		/* 20 - size of root filesystem */
		ios.ios_gpb1,		/* 21 - gap byte 1 */
		ios.ios_gpb2,		/* 22 - gap byte 2 */
		ios.ios_gpb3,		/* 23 - gap byte 3 */
		ios.ios_gpb4,		/* 24 - gap byte 4 */
		ios.ios_eatw );		/* 25 - sector slip count */
#endif
	return(n);
}

/*
 *	chgtolower: change string to lower case
 */

chgtolower(s)
register char *s;
{
	for( ; *s; s++ )
		*s = tolower(*s);
}

/*
 *	rdalts: read alternates file
 */

rdalts(name)
register char	*name;
{	register lncnt;
	register FILE *fp;
	unsigned n;

	/*
	 * 	open alternates file
	 */

	sprintf( fname, "%s/%s", altsdir, name );

	if(( fp = fopen( fname, "r" )) == NULL ) {
		fprintf( stderr, "ddefs: %s not found\n", fname );
		return;
	}

	alts = NULL;
	ios.ios_alts = 0;

	/*
	 *	read the alternates information into a linked list
	 */

	while( fscanf( fp, " %d", &n) > 0 )
		addalts(n);

	fclose(fp);
}

/*
 *	readdef: read new style def entry
 */

readdef(name)
register char *name;
{	register FILE *fp;
	register acnt;
	register i;
	register char *c;
	unsigned n;

	/*
	 *	open ddefs file
	 */

	if( diskdev )
		strcpy( fname, name );
	else
		sprintf( fname, "%s/%s", ddefsdir, name );

	if(( fp = fopen( fname, "r" )) == NULL ) {
		fprintf( stderr, "ddefs: %s not found\n", fname );
		exit(0);
	}

	/*
	 *	initialize data structures
	 */

	fmtprog[0] = '\0';
	comment[0] = '\0';
	rdprog[0]  = '\0';

	/*
	 *	read vid and ios blocks
	 */

	if( fread( &vid, sizeof(vid), 1, fp ) != 1 )	/* read vid */
		memset( &vid, 0, sizeof(vid) );
	if( fread( &ios, sizeof(ios), 1, fp ) != 1 ) {	/* read ios */
		memset( &ios, 0, sizeof(ios) );
		ios.ios_rec = 256;	/* default logical sector size */
	}

	if( diskdev ) {			/* if char special disk, then done */
		fclose(fp);
		return;
	}

	/*
	 *	read alternates
	 */

	acnt = ios.ios_alts;

	for( i=0, alts=NULL, ios.ios_alts=0; i < acnt; i++ ) {
		fread( &n, sizeof(n), 1, fp );
		addalts(n);
	}

	/* read disk formatting program */
	if( fgets( fmtprog, BUFSIZ, fp ) == NULL )
		fmtprog[0] = '\0';
	else
		fmtprog[strlen(fmtprog)-1] = '\0';	/* kill \n */

	/* read comment line */
	if( fgets( comment, BUFSIZ, fp ) == NULL )
		comment[0] = '\0';
	else
		comment[strlen(comment)-1] = '\0';	/* kill \n */

	/* read read manufacturer bad list program */
	if( fgets( rdprog, BUFSIZ, fp ) == NULL )
		rdprog[0] = '\0';
	else
		rdprog[strlen(rdprog)-1] = '\0';	/* kill \n */

	fclose(fp);
}

/*
 *	addalts: add alternate to alternates list
 */

addalts(n)
register n;
{	register struct alts *ap;
	register struct alts *lst;
	register struct alts *new;
	char	*malloc();

	if( n == 0 ) {
		fprintf( stderr, "ddefs: zero alternates not allowed\n" );
		return;
	}

	/*
	 *	look for place to add to alternates list
	 */

	for( lst=NULL, ap=alts; ap != NULL; lst=ap, ap=ap->a_next ) {

		if(( ap->a_altb <= n ) && ( n <= ap->a_alte ))
			return;	/* duplicate */

		/*
		 *	at beginning or end of current block ?
		 */

		if( ap->a_altb == n+1 ) {
			ios.ios_alts++;
			ap->a_altb = n;
			if(( lst != NULL ) && ( lst->a_alte == n-1 )) {
				lst->a_next = ap->a_next;
				lst->a_alte = ap->a_alte;
				free(ap);
			}
			return;
		}
		
		if( ap->a_alte == n-1 ) {
			ios.ios_alts++;
			ap->a_alte = n;
			lst = ap->a_next;
			if( lst->a_altb == n+1 ) {
				ap->a_next = lst->a_next;
				ap->a_alte = lst->a_alte;
				free(lst);
			}
			return;
		}

		if( ap->a_alte > n )
			break;
	}

	/*
	 *	no adjacent range
	 */

	if(( new = (struct alts *)malloc(sizeof(*ap))) == NULL ) {
		fprintf( stderr, "ddefs: out of space\n");
		return;
	}

	new->a_altb = n;
	new->a_alte = n;
	new->a_next = ap;

	if( lst == NULL )
		alts = new;
	else
		lst->a_next = new;

	ios.ios_alts++;
}

/*
 *	delalts: delete alternate from alternates list
 */

delalts(n)
register n;
{	register struct alts *ap;
	register struct alts *lst;
	char	*malloc();

	if( n == 0 ) {		/* delete all alternates if n == 0 */
		ios.ios_alts = 0;
		for( ap = alts; ap != NULL; ap = lst ) {
			lst = ap->a_next;
			free(ap);
		}
		alts = NULL;
		return;
	}

	for( lst = NULL, ap = alts; ap != NULL; lst = ap, ap = ap->a_next ) {

		/* if beyond end, then continue */
		if( ap->a_alte < n )
			continue;

		/* if before beginning, then exit */
		if( ap->a_altb > n )
			break;

		ios.ios_alts--;

		/* if element only contains this alternate, remove it */
		if( ap->a_altb == ap->a_alte ) {
			if( lst == NULL )
				alts = ap->a_next;
			else
				lst->a_next = ap->a_next;
			free(ap);
			break;
		}

		/* if element at the beginning, remove from the range */
		if( ap->a_altb == n ) {
			ap->a_altb++;
			break;
		}

		/* if element at the end, remove from the range */
		if( ap->a_alte == n ) {
			ap->a_alte--;
			break;
		}

		/* element in the middle, then make a new range */
		if(( lst = (struct alts *)malloc(sizeof(*lst))) == NULL ) {
			fprintf( stderr, "ddefs: out of space\n" );
			break;
		}

		lst->a_altb = n+1;
		lst->a_alte = ap->a_alte;
		lst->a_next = ap->a_next;

		ap->a_alte = n-1;
		ap->a_next = lst;
		break;
	}
}

/*
 *	wrtdef: write new style def entry
 */

wrtdef(name,rel)
register char *name;
register rel;
{	register FILE *fp;
	register struct alts *ap;
	register struct alts *nxt;
	unsigned n;

	/*
	 *	open ddefs file
	 */

	if( diskdev )
		strcpy( fname, name );
	else
		sprintf( fname, "%s/%s", ddefsdir, name );

	if(( fp = fopen( fname, "w" )) == NULL ) {
		fprintf( stderr, "ddefs: can't write %s\n", fname );
		exit(errno);
	}

	checkcon();

	fwrite( &vid, sizeof(vid), 1, fp );		/* write vid */
	fwrite( &ios, sizeof(ios), 1, fp );		/* write ios */

	if( diskdev ) {		/* all done if char special */
		fclose(fp);
		return;
	}

	/* write alternates */
	for( ap=alts; ap != NULL; ap = nxt ) {
		for( n=ap->a_altb; n <= ap->a_alte; n++ ) {
			fwrite( &n, sizeof(n), 1, fp );
		}
		nxt = ap->a_next;
		if( rel )
			free(ap);	/* free space */
	}

	if( rel ) {
		alts = NULL;
		ios.ios_alts = 0;
	}

	/* write disk formatting program */
	if( fmtprog[0] != '\0' )
		fprintf( fp, "%s\n", fmtprog );

	/* write comment */
	if( comment[0] != '\0' ) {
		if( fmtprog[0] == '\0' )
			fprintf( fp, "\n" );
		fprintf( fp, "%s\n", comment );
	}

	/* write manufacturer defect list program */
	if( rdprog[0] != '\0' )
		fprintf( fp, "%s\n", rdprog );

	fclose(fp);
}

/*
 *	existdef: see if name already exists
 */

existdef(name)
register char *name;
{	struct stat stbuf;

	if( name == NULL )
		return(0);

	/*
	 *	stat file
	 */

	sprintf( fname, "%s/%s", ddefsdir, name );

	if( stat( fname, &stbuf) != -1 )
		return(1);

	/*
	 *	check for character device file
	 */

	if( stat( name, &stbuf ) == -1 )
		return(0);			/* not a file */

	strcpy( fname, name );			/* save this name */
	return( diskdev = (( stbuf.st_mode&S_IFMT ) == S_IFCHR ) );
}

/*
 *	editdef: edit new style def entry
 */

editdef(name)
register char *name;
{	register struct edit *ep = edit;
	register changed = chgd;
	char	line[BUFSIZ];

	/* edit forever */
	for(;;) {
		if( diskdev ) {
			while( ep->e_flags&E_DSKNOACC ) {
				if( (++ep)->e_prompt == NULL )
					ep = edit;
			}
		}

		fprintf( stdout, "%30.30s", ep->e_prompt );

		if(( action != DK_NEW ) || (ep->e_flags&E_CHANGED) ) {
			printf(": ");
			prfield(ep,stdout);	/* print field to stdout */
			fprintf( stdout, " " );
		} else
			printf("? ");

		fflush( stdout );

		line[0] = '\0';

		if( fgets( line, BUFSIZ, stdin ) == NULL )
			continue;

		if(( line[0] == '\n' ) || ( line[0] == '\0' )) {
			if( (++ep)->e_prompt == NULL )
				ep = edit;
			continue;
		}

		if( line[0] == '!' ) {	/* shell escape */
			shesc(&line[1]);
			continue;
		}

		if( strcmp( line, "-\n" ) == 0 ) {
			do {
				if( ep == edit )
					while( (++ep)->e_prompt != NULL );
				--ep;
			} while( diskdev && (ep->e_flags&E_DSKNOACC) );
			continue;
		}

		if( strcmp( line, "f\n" ) == 0 ) {
			printf("\nDisk name: %s", diskname );
			if( changed )
				printf(" (modified)");
			else if( action == DK_RDONLY )
				printf(" (read only)");
			printf("\n\n");
			continue;
		}

		if( strcmp( line, "q\n" ) == 0 ) {
			if( unchanged() ) {
				printf( "\n(q! overrides)\n\n" );
				continue;
			}
			if( changed ) {
				printf( "\nNo write since last change\n" );
				printf( "(q! overrides)\n\n" );
				continue;
			}
			return;
		}

		if( strcmp( line, "q!\n" ) == 0 )
			return;

		if( strcmp( line, "w\n" ) == 0 ) {
			if( unchanged() ) {
				checkcon();
				continue;
			}
			if( ! changed ) {
				checkcon();
				continue;
			}
			checkcon();
			wrtdef(name,0);
			printf("\n`%s' written\n\n", fname );
			changed = 0;
			continue;
		}

		if( strcmp( line, "w!\n" ) == 0 ) {
			unchanged();
			checkcon();
			wrtdef(name,0);
			printf("\n`%s' written\n\n", fname );
			changed = 0;
			continue;
		}

		if( strcmp( line, "wq\n" ) == 0 ) {
			if( unchanged() ) {
				checkcon();
				continue;
			}
			if( changed ) {
				wrtdef(name,0);
				printf("\n`%s' written\n\n", fname );
			} else
				checkcon();
			return;
		}

		if( strcmp( line, "?\n" ) == 0 ) {
			printf("\n%s\n", ep->e_help );
			continue;
		}

		if( strcmp( line, "help\n" ) == 0 ) {
			printf("\nLegal commands:\n");
			printf("\t<newline>\tstep to next field\n");
			printf("\t-\t\tstep to previous field\n");
			printf("\tf\t\tprint file (disk) name\n");
			printf("\tw\t\twrite new ddefs file\n");
			printf("\tw!\t\twrite new ddefs file (no checks)\n");
			printf("\tq\t\tquit ddefs edit\n");
			printf("\tq!\t\tquit ddefs edit (no write)\n");
			printf("\twq\t\twrite ddefs file and quit\n");
			printf("\t?\t\tprint information about field\n");
			printf("\t!\t\texecute interactive shell\n");
			printf("\t!cmd\t\texecute shell command\n");
			printf("\thelp\t\tprint this help summary\n\n");
			continue;
		}

		if( action == DK_RDONLY ) {
			printf( "\nRead-only mode, no changes allowed\n\n");
			continue;
		}

		if( diskdev && ((ep->e_flags&E_DSKCHG) == 0) ) {
			printf( "\nField read-only, no changes allowed\n\n");
			continue;
		}

		/*
		 *	 do change
		 */

		if( dochange(ep,line) ) {
			ep->e_flags |= E_CHANGED;
			changed++;
			if( (ep-edit) != C_ALTERNATES )
				if( (++ep)->e_prompt == NULL )
					ep = edit;
		}
	}
}

/*
 *	shesc: shell escape
 */

shesc(cmd)
register char *cmd;
{	register pid;
	register w;
	int	status;

	cmd[strlen(cmd)-1] = '\0';	/* kill newline */

	if((pid = fork()) == 0) {
		if( *cmd == '\0' )
			execl("/bin/sh", "sh", 0);
		else
			execl("/bin/sh", "sh", "-c", cmd, 0);
		exit(0);
	}

	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	while(((w = wait(&status)) != pid) && (w != -1));

	fprintf( stdout, "\nHit return to continue ");
	fflush(stdout);

	while( fgetc( stdin ) != '\n' );
	printf("\n");

	signal(SIGINT, SIG_DFL);
	signal(SIGQUIT, SIG_DFL);
}

/*
 *	prfield: print field contents
 */

prfield(ep,fp)
register struct edit *ep;
register FILE *fp;
{
	register int slice, reserved;
#ifdef m88k
	slice = (int) vid.vid_u_uslice;
	slice = ((slice << 16) | vid.vid_l_uslice);
	reserved = (int) vid.vid_u_reserved;
	reserved = ((reserved << 16) | vid.vid_l_reserved);
#else
	slice = vid.vid_uslice;
	reserved = vid.vid_reserved;
#endif
	if(( ep == NULL ) || ( ep->e_prompt == NULL ))
		return;

	switch(ep-edit) {
	case C_COMMENT:
		if( comment[0] != '\0' )
			fprintf( fp, ep->e_fmt, comment );
		else
			fprintf( fp, "(none)" );
		break;
	case C_TYPE:
		fprintf( fp, ep->e_fmt, vid.vid_typ );
		break;
	case C_FORMAT:
		if( fmtprog[0] != '\0' )
			fprintf( fp, ep->e_fmt, fmtprog );
		else
			fprintf( fp, "(none)" );
		break;
	case C_RDPROG:
		if( rdprog[0] != '\0' )
			fprintf( fp, ep->e_fmt, rdprog );
		else
			fprintf( fp, "(none)" );
		break;
	case C_DIAGNOSTIC:
		fprintf( fp, ep->e_fmt, vid.vid_dta? "yes" : "no" );
		break;
	case C_STRATEGY:
		switch( vid.vid_strat ) {
		default:
			vid.vid_strat = DKPERFECT;
		case DKPERFECT:
			fprintf( fp, ep->e_fmt, "PERFECT" );
			break;
		case DKSOFTWARE&0xff:
			fprintf( fp, ep->e_fmt, "SOFTWARE" );
			break;
		case DKHARDWARE:
			fprintf( fp, ep->e_fmt, "HARDWARE" );
			break;
		case DKHARDSEC:
			fprintf( fp, ep->e_fmt, "SECTOR" );
			break;
		case DKHARDSPOT:
			fprintf( fp, ep->e_fmt, "SPOT" );
			break;
		}
		break;
	case C_MAXBAD:
		fprintf( fp, ep->e_fmt, vid.vid_maxbad );
		break;
	case C_SECTORS:
		fprintf( fp, ep->e_fmt, ios.ios_rsz );
		break;
	case C_SECSIZE:
		fprintf( fp, ep->e_fmt, ios.ios_psm );
		break;
	case C_SECPERTRK:
		fprintf( fp, ep->e_fmt, ios.ios_spt );
		break;
	case C_CYLINDERS:
		fprintf( fp, ep->e_fmt, ios.ios_trk );
		break;
	case C_HEADS:
		fprintf( fp, ep->e_fmt, ios.ios_hds );
		break;
	case C_PRECOMP:
		fprintf( fp, ep->e_fmt, ios.ios_pcom );
		break;
	case C_INTERLEAVE:
		fprintf( fp, ep->e_fmt, ios.ios_ilv );
		break;
	case C_SPIRAL:
		fprintf( fp, ep->e_fmt, ios.ios_sof );
		break;
	case C_STEPRATE:
		fprintf( fp, ep->e_fmt, ios.ios_sr );
		break;
	case C_STARTING:
		fprintf( fp, ep->e_fmt, ios.ios_shn );
		break;
	case C_ECC:
		fprintf( fp, ep->e_fmt, ios.ios_ecc );
		break;
	case C_ATRMASK:
		fprintf( fp, ep->e_fmt, ios.ios_atm );
		break;
	case C_EXTAMASK:
		fprintf( fp, ep->e_fmt, ios.ios_eatm );
		break;
	case C_ATRWORD:
		fprintf( fp, ep->e_fmt, ios.ios_atw );
		break;
	case C_EXTAWORD:
		fprintf( fp, ep->e_fmt, ios.ios_eatw );
		break;
	case C_GAP1:
		fprintf( fp, ep->e_fmt, ios.ios_gpb1 );
		break;
	case C_GAP2:
		fprintf( fp, ep->e_fmt, ios.ios_gpb2 );
		break;
	case C_GAP3:
		fprintf( fp, ep->e_fmt, ios.ios_gpb3 );
		break;
	case C_GAP4:
		fprintf( fp, ep->e_fmt, ios.ios_gpb4 );
		break;
	case C_CATR:
#ifdef m88k
		fprintf( fp, ep->e_fmt, (ios.ios_catr_l | (ios.ios_catr_u << 16)) );
#else
		fprintf( fp, ep->e_fmt, ios.ios_catr );
#endif
		break;
	case C_USS:
		fprintf( fp, ep->e_fmt, ios.ios_uss );
		break;
	case C_SECSLIP:
		fprintf( fp, ep->e_fmt, ios.ios_ssc );
		break;
	case C_SLICES:
		fprintf( fp, ep->e_fmt, ios.ios_slccnt );
		break;
	case C_ROOTOFFSET:
		fprintf( fp, ep->e_fmt, vid.vid_rfs );
		break;
	case C_ROOTSIZE:
		fprintf( fp, ep->e_fmt, vid.vid_rsize );
		break;
	case C_USRSIZE:
		fprintf( fp, ep->e_fmt, vid.vid_usize );
		break;
	case C_USRSLICE:
		fprintf( fp, ep->e_fmt, slice );
		break;
	case C_SWAPSIZE:
		fprintf( fp, ep->e_fmt, vid.vid_ssize );
		break;
	case C_SWAPSLICE:
		fprintf( fp, ep->e_fmt, vid.vid_sslice );
		break;
	case C_RESERVED:
		fprintf( fp, ep->e_fmt, reserved );
		break;
	case C_ALTERNATES:
		pralts();
		break;
	default:
		fprintf( fp, "unknown field");
		break;
	}
}

/*
 *	pralts: print alternates
 */

pralts()
{	register struct alts *ap;

	printf("(%d) ", ios.ios_alts );

	for( ap=alts; ap != NULL; ap = ap->a_next ) {
		if( ap != alts )
			printf(",");
		printf( "%d", ap->a_altb );
		if( ap->a_altb != ap->a_alte )
			printf("-%d", ap->a_alte );
	}
}

/*
 *	dochange: execute change in field
 */

dochange(ep,line)
register struct edit *ep;
register char *line;
{	register chg = 0;
	register help = 0;
	unsigned n;

	if(( ep == NULL ) || ( ep->e_prompt == NULL ))
		return(0);

	if( action == DK_RDONLY ) {
		printf("\nRead-only mode, no changes permitted\n\n");
		return(0);
	}

	switch( ep->e_fmt[1] ) {
	case 'd':	/* decimal */
	case 'x':	/* hex */
		chg = sscanf( line, ep->e_fmt, &n );
		break;
	case 's':	/* string */
		chg = 1;
		line[strlen(line)-1] = '\0';	/* kill \n */
		break;
	default:
		chg = 0;
		break;
	}

	if( chg == 0 ) {	/* scanf failure or blank line */
		printf("\n%s\n\n", ep->e_help );
		return(chg);
	}

	switch(ep-edit) {
	case C_COMMENT:
		if(( strcmp( line, "none" ) == 0) ||
		   ( strcmp( line, "(none)" ) == 0 ))
			comment[0] = '\0';
		else
			strcpy( comment, line );
		break;
	case C_TYPE:
		if( n == 0 ) {
			printf("\nType must be non-zero\n\n");
			chg = 0;
		} else
			vid.vid_typ = n;
		break;
	case C_FORMAT:
		if(( strcmp( line, "none" ) == 0) ||
		   ( strcmp( line, "(none)" ) == 0 ))
			fmtprog[0] = '\0';
		else
			strcpy( fmtprog, line );
		break;
	case C_RDPROG:
		if(( strcmp( line, "none" ) == 0) ||
		   ( strcmp( line, "(none)" ) == 0 ))
			rdprog[0] = '\0';
		else
			strcpy( rdprog, line );
		break;
	case C_DIAGNOSTIC:
		chgtolower(line);
		if( substr( line, "yes" ) )
			vid.vid_dta = 1;
		else if( substr( line, "no" ) )
			vid.vid_dta = 0;
		else
			help++;
		break;
	case C_STRATEGY:
		chgtolower(line);
		if( substr( line, "perfect" ) )
			vid.vid_strat = DKPERFECT;
		else if( substr( line, "software" ))
			vid.vid_strat = DKSOFTWARE;
		else if( substr( line, "hardware" ))
			vid.vid_strat = DKHARDWARE;
		else if( substr( line, "sector" ))
			vid.vid_strat = DKHARDSEC;
		else if( substr( line, "spot" ))
			vid.vid_strat = DKHARDSPOT;
		else {
			help++;
			chg = 0;
		}
		break;
	case C_MAXBAD:
		vid.vid_maxbad = n;
		break;
	case C_SECTORS:
		if( n == 0 ) {
			help++;
			chg = 0;
		} else
			ios.ios_rsz = n;
		break;
	case C_SECSIZE:
		if(( n == 128 ) || ( n == 256 ) || ( n == 512 ))
			ios.ios_psm = n;
		else {
			help++;
			chg = 0;
		}
		break;
	case C_SECPERTRK:
		if( n == 0 ) {
			printf("\nThe sectors per track must be nonzero\n\n");
			chg = 0;
		} else
			ios.ios_spt = n;
		break;
	case C_CYLINDERS:
		if( n == 0 ) {
			printf("\nThe cylinder count must be nonzero\n\n");
			chg = 0;
		} else
			ios.ios_trk = n;
		break;
	case C_HEADS:
		if( n == 0 ) {
			printf("\nThe heads count must be nonzero\n\n");
			chg = 0;
		} else
			ios.ios_hds = n;
		break;
	case C_PRECOMP:
		ios.ios_pcom = n;
		break;
	case C_INTERLEAVE:
		ios.ios_ilv = n;
		break;
	case C_SPIRAL:
		ios.ios_sof = n;
		break;
	case C_STEPRATE:
		ios.ios_sr = n;
		break;
	case C_STARTING:
		ios.ios_shn = n;
		break;
	case C_ECC:
		ios.ios_ecc = n;
		break;
	case C_ATRMASK:
		ios.ios_atm = n;
		break;
	case C_EXTAMASK:
		ios.ios_eatm = n;
		break;
	case C_ATRWORD:
		ios.ios_atw = n;
		break;
	case C_EXTAWORD:
		ios.ios_eatw = n;
		break;
	case C_GAP1:
		ios.ios_gpb1 = n;
		break;
	case C_GAP2:
		ios.ios_gpb2 = n;
		break;
	case C_GAP3:
		ios.ios_gpb3 = n;
		break;
	case C_GAP4:
		ios.ios_gpb4 = n;
		break;
	case C_CATR:
#ifdef m88k
		ios.ios_catr_l = (short) (n);
		ios.ios_catr_u = (short) (n >> 16);
#else
		ios.ios_catr = n;
#endif
		break;
	case C_USS:
		ios.ios_uss = n;
		break;
	case C_SECSLIP:
		ios.ios_ssc = n;
		break;
	case C_SLICES:
		if(( n == 0 ) || ( n == 8 ) || ( n == 16 ) || ( n == 32 )
		|| ( n == 64 ) || ( n == 128 ))
			ios.ios_slccnt = n;
		else {
			help++;
			chg = 0;
		}
		break;
	case C_ROOTOFFSET:
		if( n == 0 ) {
			printf("\nRoot filesystem offset must be nonzero\n\n");
			chg = 0;
		} else
			vid.vid_rfs = n;
		break;
	case C_ROOTSIZE:
		vid.vid_rsize = n;
		break;
	case C_USRSIZE:
		vid.vid_usize = n;
		break;
	case C_USRSLICE:
		if((n == 0) && (vid.vid_usize == 0)) {
#ifdef m88k
			vid.vid_u_uslice = 0;
			vid.vid_l_uslice = 0;
#else
			vid.vid_uslice = 0;
#endif
			break;
		}
		if(( n == 1 ) || ( n == 2 ))
		{
#ifdef m88k
			vid.vid_l_uslice = (short) (n);
			vid.vid_u_uslice = (short) (n >> 16);
#else
			vid.vid_uslice = n;
#endif
		}
		else {
			help++;
			chg = 0;
		}
		break;
	case C_SWAPSIZE:
		vid.vid_ssize = n;
		break;
	case C_SWAPSLICE:
		if(( n >= 0 ) && ( n <= 2 ))
			vid.vid_sslice = n;
		else {
			help++;
			chg = 0;
		}
		break;
	case C_RESERVED:
#ifdef m88k
		vid.vid_l_reserved = (short) (n);
		vid.vid_u_reserved = (short) (n >> 16);
#else
		vid.vid_reserved = n;
#endif
		break;
	case C_ALTERNATES:
	{	register char *c;
		char *r;
		register cnt;
		char ch;
		unsigned m;

		chgtolower(line);

		if( strcmp( line, "none" ) == 0 ) {
			delalts(0);
			break;
		}

		for( c = line; *c && isspace(*c); c++ );
		if( *c == '\0' ) {
			help++;
			chg = 0;
			break;
		}
#ifdef	DDEFSINIT
		if( *c == '(' ) {	/* scan for close paren */
			for( ; *c && (*c != ')' ); c++ );
			if( *c == '\0' )
				break;
			/* skip spaces again */
			for( ; *c && isspace(*c); c++ );
			*line = 'a';	/* do `add' in this case */

		} else
#endif
		{
			/* skip command */
			while( *c && (! isspace(*c))) c++;
			if( *c == '\0' )
				break;
			*c = '\0';
			if((! substr( line, "add" )) 
			  && (! substr( line, "delete" ))) {
				help++;
				chg = 0;
				break;
			}
		}

		chg = 0;

		for(;;) {
			for( ++c; *c && isspace(*c); c++ );
			if( *c == '\0' )
				break;
			for( r=c; *c && (*c != ','); c++ );
			if(( cnt = sscanf( r, "%d-%d", &n, &m )) == 0 )
				break;
			if(( cnt < 1 ) || ( cnt > 2 )) {
				help++;
				break;
			}
			if( cnt == 1 )
				m = n;
			else if( n > m ) {
				cnt = n;
				n = m;
				m = cnt;
			}
			chg = 1;
			for( cnt=n; cnt <= m; cnt++ ) {
				if( *line == 'a' )
					addalts(cnt);
				else
					delalts(cnt);
			}
			if( *c == '\0' )
				break;
		}
		break;
	}
	default:
		break;
	}

	if( help )
		printf("\n%s\n", ep->e_help );

	return(chg);
}

/*
 *	substr: returns true of s1 is a substring of s2
 */

substr( s1, s2 )
register char *s1;
register char *s2;
{
	while( *s1 == *s2++ )
		if( *s1++ == '\0' )
			return(1);
	
	return( *s1 == '\0' );
}

/*
 *	printdef: print out current fields
 */

printdef(name)
register char *name;
{	register struct edit *ep;

	printf("\t\tDisk definitions for `%s'\n", name );

	for( ep = edit; ep->e_prompt != NULL; ep++ ) {
		if( diskdev && (ep->e_flags&E_DSKNOACC) )
			continue;	/* inaccessible */
		printf( "%30.30s: ", ep->e_prompt );
		prfield(ep,stdout);
		printf("\n");
	}
}

/*
 *	unchanged: return true if there are uninitialized fields
 */

unchanged()
{	register struct edit *ep;
	register emsg = 0;

	if( action != DK_NEW )	/* only worry about new entries */
		return(0);

	for( ep = edit; ep->e_prompt != NULL; ep++ ) {
		if( diskdev && (ep->e_flags&E_DSKNOACC) )
			continue;
		if(( ep->e_flags&E_CHANGED) == 0 ) {
			if((( ep-edit ) == C_ALTERNATES )
			&& ( vid.vid_strat == DKPERFECT ))
				continue;
			if( emsg++ == 0 )
				printf("\nFields still uninitialized:\n");
			printf("\t%s\n", ep->e_prompt );
		}
	}

	if( emsg )
		printf("\n");

	return(emsg);
}

/*
 *	setchanged: mark all elements as changed
 */

setchanged()
{	register struct edit *ep;

	for( ep = edit; ep->e_prompt != NULL; (ep++)->e_flags |= E_CHANGED );

	/*
	 *	each of the following fields must be reset
	 */

	edit[C_TYPE].e_flags &= ~E_CHANGED;
	edit[C_MAXBAD].e_flags &= ~E_CHANGED;
	edit[C_CYLINDERS].e_flags &= ~E_CHANGED;
	edit[C_HEADS].e_flags &= ~E_CHANGED;
	edit[C_ROOTOFFSET].e_flags &= ~E_CHANGED;
	edit[C_RESERVED].e_flags &= ~E_CHANGED;

	chgd++;
}

/*
 *	checkcon: check for consistency
 */

checkcon()
{	register unsigned sectors;
	register unsigned blocks;
	register unsigned rootfs;
	register unsigned tno;
	register struct alts *ap;
	register errs = 0;

	register int slice, reserved;
#ifdef m88k
	slice = (int) vid.vid_u_uslice;
	slice = ((slice << 16) | vid.vid_l_uslice);
	reserved = (int) vid.vid_u_reserved;
	reserved = ((reserved << 16) | vid.vid_l_reserved);
#else
	slice = vid.vid_uslice;
	reserved = vid.vid_reserved;
#endif
	if( diskdev )
		return;

	/*
	 *	check total sectors consistency
	 */


	/* compute number of 256 byte LOGICAL sectors */
	sectors = ios.ios_spt * ios.ios_hds * ios.ios_trk;

	switch( ios.ios_psm ) {
	case 128:	sectors /= 2;	break;
	case 512:	sectors *= 2;	break;
	default:			break;
	}

	if( sectors != ios.ios_rsz ) {
		printf("\nWarning: number of sectors inconsistent");
		printf(" with computed value.\n");
		printf("Given: %d\tComputed: %d\n\n", ios.ios_rsz, sectors );
	}

	/* 
	 *	check slice table consistency
	 */

	if( ios.ios_slccnt == 0 )
		return;			/* no slice table */

	blocks = sectors/2;		/* convert to 512-byte blocks */
	rootfs = vid.vid_rfs;		/* get root offset (1K blocks) */
	rootfs *= 2;			/* convert to 512-byte blocks */

#ifdef	DEBUGC
	printf("Block count: %d\n", blocks );
	printf("Root offset: %d\n", rootfs );
	printf("Root size: %d\n", vid.vid_rsize );
	printf("/usr size: %d\n", vid.vid_usize );
	printf("/usr slice: %d\n", slice );
	printf("Swap size: %d\n", vid.vid_ssize );
	printf("Swap slice: %d\n", vid.vid_sslice );
	printf("Reserved size: %d\n", reserved );
#endif

	if( rootfs >= blocks )
	    printf("\nWarning: root filesystem offset too large.\n\n");

	else if( rootfs+vid.vid_rsize > blocks )
	    printf("\nWarning: root filesystem runs off end of disk.\n\n");

	else if( rootfs+vid.vid_rsize > blocks-reserved )
	    printf("\nWarning: root filesystem runs into reserved area.\n\n");

	else if( vid.vid_usize == 0 ) {		/* no /usr filesystem */

	    if( vid.vid_ssize && (vid.vid_sslice > 1))
		printf("\nWarning: swap must be in slice 0 or 1\n\n");

	    if( rootfs+vid.vid_rsize+vid.vid_ssize > blocks )
		printf("\nWarning: swap space runs off end of disk\n\n");

	    else if( rootfs+vid.vid_rsize+vid.vid_ssize >
						blocks-reserved )
		printf("\nWarning: swap space runs into reserved area.\n\n");

	/* /usr filesystem */
	} else if( slice == 0 )
		printf("\nWarning: /usr filesystem slice must be nonzero\n\n");
	
	else if( vid.vid_ssize == 0 ) {		/* no swap space */

	    if( rootfs+vid.vid_rsize+vid.vid_usize > blocks )
		printf("\nWarning: /usr filesystem runs off end of disk\n\n");

	    else if( rootfs+vid.vid_rsize+vid.vid_usize >
						blocks-reserved )
		printf("\nWarning: /usr filesystem runs into reserved area.\n\n");

	} else if(( slice > 1 ) && ( vid.vid_sslice != 1 ))
		printf("\nWarning: /usr filesystem not in slice 1\n\n");

	/* if swap area not at end */
	else if( vid.vid_sslice < slice ) {

	    if( rootfs+vid.vid_rsize+vid.vid_ssize+vid.vid_usize > blocks )
		printf("\nWarning: /usr filesystem runs off end of disk\n\n");

	    else if( rootfs+vid.vid_rsize+vid.vid_ssize+vid.vid_usize >
						blocks-reserved )
		printf("\nWarning: /usr filesystem runs into reserved area.\n\n");

	} else if( rootfs+vid.vid_rsize+vid.vid_ssize+vid.vid_usize > blocks )
		printf("\nWarning: swap runs off end of disk\n\n");

	else if( rootfs+vid.vid_rsize+vid.vid_ssize+vid.vid_usize >
						blocks-reserved )
		printf("\nWarning: swap runs into reserved area.\n\n");

	/* make sure alternates are not in usable area */

	blocks -= reserved;	/* last usable block */

	/* compute blocks per track */
	sectors /= (ios.ios_hds * ios.ios_trk);
	sectors /= 2;

#ifdef	DEBUGC
	printf("Blocks per track: %d\n", sectors );
	printf("Beginning track: %d\n", rootfs/sectors );
	printf("Ending track: %d\n", blocks/sectors );
#endif
	
	for( ap=alts; ap != NULL; ap = ap->a_next ) {
		for( tno = ap->a_altb; tno <= ap->a_alte; tno++ ) {
			if(( tno*sectors >= rootfs ) &&
			   ( tno*sectors < blocks )) {
			    if( errs == 0 )
				printf("\n");
			    printf("Warning: invalid alternate: %d\n", tno );
			    errs++;
			}
		}
	}

	if( errs )
		printf("\n");

#ifdef	SLCDEBUG
	prtslice();
#endif
}

#ifdef	DDEFSINIT

initdef()
{	register struct edit *ep;
	static char line[BUFSIZ];
	register char *c;
	register char *l;
	char *matchpat();

	action = DK_NEW;		/* make a new entry */

	/*
	 *	initialize structures
	 */

	memset( &vid, 0, sizeof(vid) );
	memset( &ios, 0, sizeof(ios) );
	ios.ios_rec = 256;		/* default logical sector size */
	alts = NULL;
	fmtprog[0] = 0;
	comment[0] = 0;
	rdprog[0]  = 0;

	/*
	 *	read disk definitions line
	 */

	if( fgets( line, BUFSIZ, stdin ) == NULL ) {
		fprintf( stderr, "ddefs: unexpected EOF\n" );
		exit(EINVAL);
	}

	/*
	 *	get pointer to argument
	 */

	if(( c = matchpat( line, "Disk definitions for ")) == NULL ) {
		fprintf( stderr, "ddefs: missing definitions line\n");
		exit(EINVAL);
	}

	/*
	 *	get disk name
	 */

	if( *c++ != '`' ) {
		fprintf( stderr, "ddefs: missing ` in definitions line\n" );
		exit(EINVAL);
	}

	for( l=c; *c && (*c != '\''); c++ );	/* scan for ' */

	if( *c != '\'' ) {
		fprintf( stderr, "ddefs: missing ' in definitions line\n" );
		exit(EINVAL);
	}

	*c = '\0';
	strcpy( diskname = dname, l );

	/*
	 *	now read data lines
	 */

	while( fgets( line, BUFSIZ, stdin ) != NULL ) {
		for( ep = edit; ep->e_prompt != NULL; ep++ ) {
			if(( c = matchpat( line, ep->e_prompt )) == NULL )
				continue;

			if( *c++ != ':' )
				continue;
			
			if( *c++ != ' ' )
				continue;

			if( dochange(ep,c) ) {
				ep->e_flags |= E_CHANGED;
				chgd++;
			}
		}
	}

	unchanged();
	wrtdef(diskname,1);
	printf("\n`%s' written\n\n", fname );
}

/*
 *	matchpat: match input line pattern
 */

char *matchpat( line, pat )
register char *line;
register char *pat;
{
	/*
	 *	skip white space
	 */

	while( isspace(*line) )
		line++;

	/*
	 *	now match the pattern
	 */

	while( *line++ == *pat )
		if( *pat++ == '\0' )
			return(NULL);

	return((*pat == '\0')? --line: NULL );
}

#endif

#ifdef	SLCDEBUG

#ifndef	BSIZE
#define	BSIZE	512
#endif

/*
 *	globals for slice table printout
 */

struct	slice	*slctab;

/*
 *	prtslice: print slice table
 */

prtslice()
{	register struct slice *sp;
	register i;

	slctab = NULL;
	mkslice();

	if( slctab == NULL )
		return;

	printf("Slice Table Definition\n");
	printf("Slice\tLength\tOffset\n");
	for( sp=slctab, i=0; i<ios.ios_slccnt; i++, sp++ )
		printf("%d\t%d\t%d\n", i, sp->nblocks, sp->blkoff );
	free(slctab);
}

/*
 *	mkslice: allocate and initialize a slice table
 */

mkslice()
{	register nblocks;
	register slcbcnt;		/* slice table block count */
	register ssize = vid.vid_ssize;	/* swap size */
	register i;

	register int slice, reserved;
#ifdef m88k
	slice = (int) vid.vid_u_uslice;
	slice = ((slice << 16) | vid.vid_l_uslice);
	reserved = (int) vid.vid_u_reserved;
	reserved = ((reserved << 16) | vid.vid_l_reserved);
#else
	slice = vid.vid_uslice;
	reserved = vid.vid_reserved;
#endif
	/* if no slice table, just return */

	if ( ios.ios_slccnt == 0 )
		return;

	/*
	 *	compute slice table block count
	 */

	slcbcnt = ((sizeof(struct slice)*ios.ios_slccnt)+BSIZE-1)/BSIZE;

	/*
	 *	allocate slice table
	 */

	if(( slctab = (struct slice *)malloc(slcbcnt*BSIZE)) == NULL ) {
		fprintf( stderr, "dinit: out of space\n" );
		return;
	}

	/*
	 *	zero slice table
	 */

	memset( slctab, 0, slcbcnt*BSIZE );

	/*
	 *	compute blocks on disk
	 */

	nblocks =(ios.ios_spt*ios.ios_trk*ios.ios_hds)/(BSIZE/ios.ios_psm);

	/*
	 *	setup last slice and reinitialize "slctab"
	 */

	slctab[ios.ios_slccnt-1].nblocks = nblocks;	/* set up last slice */

	/*
	 *	set up slice zero
	 */

	slctab[0].blkoff = vid.vid_rfs*(1024/BSIZE);	/* rfs in K-blocks */
	slctab[0].nblocks = nblocks-slctab[0].blkoff-reserved;

	/*
	 *	filesystem on slice 0?
	 */

	if( vid.vid_rsize == 0 )
		return;		/* "-m option"  OR no filesystem  -done- */

	/*
	 *	set up unused slice 1
	 */

	slctab[1] = slctab[0];
	slctab[0].nblocks = vid.vid_rsize;
	slctab[1].nblocks -= vid.vid_rsize;
	slctab[1].blkoff += vid.vid_rsize;

	/*
	 *	if no /usr and no swap, return
	 */

	if(( ssize == 0 ) && ( vid.vid_usize == 0 ))
		return;

	/*
	 *	compute slice 0 in with root
	 */

	if( ssize && ( vid.vid_sslice == 0 )) {
		slctab[1].nblocks -= ssize;
		slctab[1].blkoff += ssize;
		slctab[0].nblocks += ssize;
		if( vid.vid_usize == 0 )
			return;	/* done ... no /usr */
		ssize = 0;	/* no more swap */
	}

	/*
	 *	compute /usr filesystem
	 */

	if( vid.vid_usize && ( slice == 1 )) {
		slctab[2].nblocks = slctab[1].nblocks - vid.vid_usize;
		slctab[2].blkoff = slctab[1].blkoff + vid.vid_usize;
		slctab[1].nblocks = vid.vid_usize;

		if( ssize ) {		/* if swap */
			if( vid.vid_sslice == 1 ) {/* swap shares /usr slice */
				slctab[2].nblocks -= ssize;
				slctab[2].blkoff += ssize;
				slctab[1].nblocks += ssize;
			} else {	/* swap is private slice 2 */
				slctab[3].nblocks = slctab[2].nblocks - ssize;
				slctab[3].blkoff = slctab[2].blkoff + ssize;
				slctab[2].nblocks = ssize;
			}
		}

	/*
	 *	check for swap in slice 1
	 */

	} else if( ssize && ( vid.vid_sslice == 1 )) {
		slctab[2].nblocks = slctab[1].nblocks - ssize;
		slctab[2].blkoff = slctab[1].blkoff + ssize;
		slctab[1].nblocks = ssize;

		/*
		 *	check for /usr in slice 2
		 */

		if( vid.vid_usize && ( slice == 2 )) {
			slctab[3].nblocks = slctab[2].nblocks - vid.vid_usize;
			slctab[3].blkoff = slctab[2].blkoff + vid.vid_usize;
			slctab[2].nblocks = vid.vid_usize;
		}
	}
}

#endif
