# @(#)portcnfg.sh	35.6
#	
# Determine what serial I/O board types exist and attempt to
# create /dev/tty{module}{port} nodes to match the back panel
# transition board configuration.
#
# In the case of the MVME336 Cluster Controller Board,
# /dev/tty entries must be appended the /etc/inittab file.
#
# Portconfig ASSUMES THAT:
#
#	a) For every installed board type, the corresponding driver
#		has been sysgen'd into the kernel, and thus appears
#		in ${MASTER}.
#		
#	b) The proper transition modules are installed for each
#		I/O board and are configured as per the "Rear Panel
#		Card Placement" rules outlined in the VME Delta SeriesTM
#		Board Strapping and Configuration Guide.
#		Specifically, that the smaller the number of ports
#		provided by a transition module, the farther
#		to the left (increasing relative module numbers) it
#		resides.  New modules are inserted to the left
#		of existing, equivalent modules; and to the right of
#		existing, smaller modules; bumping the smaller modules
#		to the left by one relative module, and requiring any
#		connected cables to be moved to the right, preserving
#		their {module}{port} name.
#
#	c) 332xt modules are always placed after (to the left of)
#		plain vme332 modules.  ie.  332's are grouped together
#		to the right of any 332xt's.
#
#	d) Attempting to open a device will fail if the corresponding 
#		board is not installed
#
# We manufacture 1 node for each possible occurrence of each possible
# board type, and use these names to search for configured boards.


PATH=$PATH:/usr/lbin
export PATH

print=/dev/null 	# dummy arg for pr(1) in $DISP

TMPDEV=/tmp/dev$$
DEV=${_DEV:-/tmp/dev}   # by default, mknod somewhere other than /dev
DISP=${_DISP:-/tmp/portdisplay}
MASTERD=${_MASTER:-/usr/src/uts/m68k/cf}
MASTER=${MASTERD}/master
INITTAB=${_INITTAB:-/tmp/inittab}
ADM=/usr/adm/ports

if [ $# != 1 -o \( "$1" != "-m" -a "$1" != "-t" \) ]
then
    cat 1>&2 <<!!!
Syntax:  $0 -t   (puts changes in $DEV, $DISP, and, ${INITTAB})
  -or-   $0 -m   (puts changes in /dev, /etc/portdisplay and /etc/inittab)

$0	creates the necessary tty* device entries for all
		configured serial I/O drivers and boards.
!!!
    exit 1
fi

# -m (modify) option puts nodes in /dev.  
# and puts new configuration display command in /etc

if [ "$1" = "-m" ]
then
    DEV=/dev
    DISP=/etc/portdisplay
    INITTAB=/etc/inittab
else
    cp /etc/inittab ${INITTAB}
fi

if [ "$DEV" = /dev ]
then
    echo "\nTHIS WILL REMOVE EXISTING tty NODES IN  '/dev'  AND"
    echo "CREATE NEW ONES BASED ON THE DRIVERS CONFIGURED INTO THE"
    echo "MOST RECENTLY BUILT KERNEL IN  '$MASTERD' AND THE CURRENT"
    echo "HARDWARE CONFIGURATION.  ARE YOU SURE THIS IS WHAT"
    echo "YOU WANT TO DO?"
    echo "(default = n) [y, n]: \c"

    read resp
    case "$resp" in
	y )  ;;

	* )   exit 2 ;;
    esac
else
    echo "MODS WILL BE MADE TO  '$DEV'  AND  '$DISP', BASED ON"
    echo "CONFIGURED DRIVERS IN '$MASTER'"
fi

sleep 1

if [ ! -f ${MASTER} ]
then
    echo "$0: Cannot access ${MASTER}: Make sure /usr is mounted ..."
    exit 1
fi

getmajor() {
    Major=`grep "^$1" ${MASTER} | awk '{ print $8 }'`
#masterline=`grep "^$1" ${MASTER}`
#if [ -n "$masterline" ]
#then
#    Major=`(set \`echo $masterline\` ; echo $8)`
#else
#   Major=
#fi
}

vme331=NULL
m332xt=NULL
vme333=NULL
vme335=NULL
mpcc=NULL
sio336=NULL
m564con=NULL
z8530=NULL
m681con=NULL

# device name in master for vme332 is vme331.  We use it as though
# it always refers to a vme332 (ie. has 8 ports).
# device name for vme050 serial ports is mpcc
# Don't know exactly what name for vme333 serial ports will be

for bdtype in vme331 m332xt vme335 mpcc sio336 m564con z8530 m681con
do
    getmajor ${bdtype}
    if [ -n "$Major" ]
    then
	eval $bdtype=${bdtype}:${Major}
    fi
done

# maketnodes() - Make 'test' nodes for each possible board of this type
# syntax: maketnodes <bdname> <majdev> <maxbds> <minorcnt>

maketnodes() {
    name=$1
    major=$2
    numbds=$3
    numdevs=$4	# number of minor devices to modulo by 
    bdno=0
    minor=0
    while [ $bdno -lt $numbds ]
    do
	mknod $TMPDEV/${name}_${bdno} c ${major} ${minor}
	bdno=`expr ${bdno} + 1`
	minor=`expr ${minor} + ${numdevs}`
    done
    }
#
#
makemodemtnodes() {
    name=$1
    major=$2
    numbds=$3
    numdevs=$4	# number of minor devices to modulo by 
    bdno=0
    mknod $TMPDEV/${name}_${bdno} c ${major} ${minor}
    }

# makettynodes() - Make the /dev/tty{module}{port} nodes for a board type
# Syntax: makettynodes  <majdev> <boardno> <portcount> <moduleno>

makettynodes() {

    bdmajor=$1
    bdno=$2
    portcnt=$3
    numdevs=$4
    module=$5
    port=1
    minor=`expr $bdno \* $numdevs`

    # Remove all existing entries for this module
    echo "Removing existing ${DEV}/tty entries for module ${module}
    rm -f $DEV/tty${module}[0123456789]"

    rm -f $DEV/tty${module}[0123456789]

    echo "Making tty entries for board ${bdno}"
    echo "Ports appear on backpanel module ${module}"
    echo "$DEV:"

    while [ $port -le $portcnt ]
    do
	mknod $DEV/tty${module}${port} c ${bdmajor} ${minor}
	echo "tty${module}${port} \c"
	minor=`expr ${minor} + 1`
	port=`expr ${port} + 1`
    done
    echo
}

configmodem() {

    bdno=0
	if openodel $TMPDEV/${bdname}_${bdno} 2>/dev/null
	then
	    echo "Found a configured board (board ${bdno})"
	    rm -f $DEV/tty0[123]
	    # corresponds to scc2 on mvme147
	    mknod $DEV/tty01 c $bdmajor $minor
	    case $bdname in
		z8530) cpuboard
			case $? in
				2) cpuname=z8530 ;; # mvme134
				6) cpuname=vme143
				   # third port on mvme143
	    			   mknod $DEV/tty02 c $bdmajor 1 ;;
				7) cpuname=m147con
				   # corresponds to scc3 for mvme147
	    			   mknod $DEV/tty02 c $bdmajor 3
				   # corresponds to scc4 on mvme147
	    			   mknod $DEV/tty03 c $bdmajor 2 ;;
				*) echo "unknown cpu board"
				   exit -1 ;;
			esac
	    		print="$ADM/$cpuname $print" ;;
		m681con) cpuboard
			case $? in
				3) cpuname=vme136 ;;
				4) cpuname=vme140 ;;
				5) cpuname=vme141 ;;
				*) echo "unknown cpu board"
				   exit -1 ;;
			esac
	    		print="$ADM/$cpuname $print" ;;
		*) print="$ADM/$bdname $print" ;;
	   esac
	fi
}
# configure() - Configure the ports for a board type
# uses all global vars from the caller: bdname, maxbds, bdmajor, portcnt,
# minorcnt, module.   Updates $module

configure() {

    bdno=0
    while [ $bdno -lt $maxbds ]
    do
	if openodel $TMPDEV/${bdname}_${bdno} 2>/dev/null
	then
	    echo "Found a configured board (board ${bdno})"
	    makettynodes $bdmajor $bdno $portcnt $minorcnt $module
	    module=`expr $module + 1`
	    print="$ADM/$bdname $print"
	fi
	bdno=`expr $bdno + 1`
    done
}

# configure336() - Configure the ports for a vme336 cluster controller.
# uses all global vars from the caller: bdname, maxbds, bdmajor, portcnt,
# minorcnt.  Updates nothing; 336 does not use the same naming convention as
# other tty's.  Ports are named  tty[link][port], where the link IDs are
# a - f and ports are numbered 1 - 16.
# Logic is sort of here for 2 boards, though only one is really supported,
# both from a software and naming standpoint; all transition modules are
# labeled a - f, and we don't really want to stick a module number ahead
# of the letter (tty2a16 ?)....

configure336() {

    found336=
    bdno=0
    while [ $bdno -lt $maxbds ]
    do
	present336=false
	if openodel $TMPDEV/${bdname}_${bdno} 2>/dev/null
	then
	    echo "Found a configured board (board ${bdno})"
	    make336nodes $bdmajor $bdno $portcnt $minorcnt
	    print="$ADM/$bdname $print"
	    found336=YES
	    present336=true
	fi
	inittab336
	bdno=`expr $bdno + 1`
    done
    unset present336
    return 0
}

# make336nodes() - Make the /dev/tty{link}{port} nodes for vme336 board type
# Syntax: make336nodes  <majdev> <boardno> <portcount>

make336nodes() {

    bdmajor=$1
    bdno=$2
    numdevs=$4
    portcnt=16
    minor=`expr $bdno \* $numdevs`


    echo "Making tty entries for vme336 board ${bdno}"
    echo "$DEV:"

    for link in a b c d e f
    do
	port=1
	prefx="0"
        while [ $port -le $portcnt ]
	do
	    mknod $DEV/tty${link}${prefx}${port} c ${bdmajor} ${minor}
	    echo "tty${link}${prefx}${port} \c"
	    if [ ${port} = 9 ] ; then  echo ; prefx= ; fi ;
	    minor=`expr ${minor} + 1`
	    port=`expr ${port} + 1`
	done
	echo
    done
    echo
}

# inittab336() - Add or delete MVME336 port associated entries in inittab.

inittab336()
{
    tmpinittab=/tmp/pcit$$
    m336pat='[ 	]tty[a-f][0-1][0-9][ 	]'

    if grep 1>/dev/null -c "${m336pat}" ${INITTAB}
    then
	has336entries=true
        grep 1>${tmpinittab} -v "${m336pat}" ${INITTAB}
    else
	has336entries=false
	cp ${INITTAB} ${tmpinittab}
    fi

    if ${present336}
    then
    	echo "Adding MVME336 Controller associated entries to ${INITTAB}."

	O_IFS="${IFS}"
	IFS=':'
	set -- `grep '^[0-9]*' ${tmpinittab} | sort -n | sed -n -e '$p'`
	IFS="${O_IFS}"

	first_entry=`expr ${1} + 1`
	if [ ${first_entry} -le 48 ]
	then
		first_entry=49
	fi
	last_entry=`expr ${first_entry} + 95`

	initentry ()
	{
		expr "000${1}" : '.*\(...\)'
	}

	echo "These entries will be numbered `initentry ${first_entry}`" \
		"through `initentry ${last_entry}`."

	entry="${first_entry}"
    	for server in a b c d e f
    	do
	    for port in 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 16
	    do
		cat <<!!!
`initentry ${entry}`:23:off:getty tty${server}${port} 9600 \
# MVME336 Server: ${server} Port: ${port}
!!!
		entry=`expr ${entry} + 1`
	    done
    	done 1>>${tmpinittab}
	mv ${tmpinittab} ${INITTAB}
    	unset port server entry initentry last_entry first_entry O_IFS
    else
	if ${has336entries}
	then
	    echo \
"Removing MVME336 Controller associated entries from ${INITTAB}."
	    mv ${tmpinittab} ${INITTAB}
	else
	    rm -f ${tmpinittab}
	fi
    fi
    unset has336entries m336pat tmpinittab
    return 0
}

if [ ! -d $TMPDEV ]
then
	mkdir $TMPDEV || exit 1
fi

if [ ! -d $DEV ]
then
	mkdir $DEV || exit 2
fi

module=1 		# Prime the module counter

# Do not change the order of these tests
# The 331 (actually 332) must be checked first, the 050 last
# to properly build up the configuration map and tty nodes

for i in $vme331 $m332xt $vme335 $mpcc $sio336 $m564con $z8530  $m681con
do
    IFS=:
    set $i

    IFS=" 	"

    if [ "$1" = NULL ]
    then
	continue
    fi

    bdname=$1
	
    bdmajor=$2

    case $bdname in

    vme331 )
	echo "\n-----------------------------------------------"
	echo "Found a vme332 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=8
	minorcnt=8
	maxbds=4
	maketnodes $bdname $bdmajor $maxbds $minorcnt
	configure
	;;

    m332xt )
	echo "\n-----------------------------------------------"
	echo "Found a vme332xt driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=9	# 9th port is parallel port, make ttyx9 for now...
	minorcnt=16
	maxbds=8
	maketnodes $bdname $bdmajor $maxbds $minorcnt
	configure
	;;

    vme335 )
	echo "\n-----------------------------------------------"
	echo "Found a vme335 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=4
	minorcnt=4
	maxbds=2
	maketnodes $bdname $bdmajor $maxbds $minorcnt
	configure
	;;

    mpcc )
	echo "\n-----------------------------------------------"
	echo "Found a vme050 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=2
	minorcnt=2
	maxbds=1
	maketnodes $bdname $bdmajor $maxbds $minorcnt
	configure
	;;

    sio336 )
	echo "\n-----------------------------------------------"
	echo "Found a vme336 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=96
	minorcnt=96
	maxbds=1
	maketnodes $bdname $bdmajor $maxbds $minorcnt
	configure336
	;;

    m564con )
	echo "\n-----------------------------------------------"
	echo "Found a m564 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=1
	minorcnt=1
	maxbds=1
	minor=1
	makemodemtnodes $bdname $bdmajor $maxbds $minorcnt $minor
	configmodem
	;;

    z8530 )
	echo "\n-----------------------------------------------"
	echo "Found a z8530 driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=1
	minorcnt=1
	maxbds=1
	minor=0
	makemodemtnodes $bdname $bdmajor $maxbds $minorcnt $minor
	configmodem
	;;

    m681con )
	echo "\n-----------------------------------------------"
	echo "Found a m681con driver entry in ${MASTER} (major # ${bdmajor})"
	portcnt=1
	minorcnt=1
	maxbds=1
	minor=1
	makemodemtnodes $bdname $bdmajor $maxbds $minorcnt $minor
	configmodem
	;;

    esac
done


echo "\n----------------------------------------------"
echo "Removing $DEV/tty entries for non-configured back panel modules:\n"

while [ ${module} -le 9 ]
do
    echo "\trm -f $DEV/tty${module}[123456789]"
    rm -f $DEV/tty${module}[123456789]
    module=`expr ${module} + 1`
done

echo "\n----------------------------------------------"
if [ -z "${found336}" -a -c "$DEV/ttya01" ]
then
    echo "Removing existing nodes for non-configured vme336 board (board 0)"
    echo "rm -fr $DEV/tty[a-f][0-1][0-9]"
    rm -fr $DEV/tty[a-f][0-1][0-9]
fi

echo "\nUse 'sysadm ttymgmt' to turn these entries 'off' in /etc/inittab"

cat << ! >$DISP
echo "\nCurrent back panel layout should be as follows:"
pr -tms' ' $print
!
chmod 755 $DISP
$DISP

# rm -fr $TMPDEV
