/* @(#)ci.c	35.1 */
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci.c
**
**  DESCRIPTION:	This module contains the functions composing
**			the Common Interface Library.
**
**  FUNCTIONS:		ci_register,	ci_deregister,
**			ci_dereg,	ci_read
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  06/05/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.2	Changed names for generic 37x.
**
***********************************************************************/

#include <fcntl.h>
#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>

extern void	close();	/* Close device */
extern int	open();		/* Open device */
extern int	read();		/* Read from device */
extern int	write();	/* Write to device */

extern int	errno;		/* Error number */

static CE_EPB	epb = {0};	/* Used for all requests to CI */

static char	*dev_tbl[] =	/* MVME37x device table */
{
	"/dev/ci/m37x_2",	/* MVME37x - Minor device 2 */
	"/dev/ci/m37x_3",	/* MVME37x - Minor device 3 */
	"/dev/ci/m37x_4",	/* MVME37x - Minor device 4 */
	"/dev/ci/m37x_5",	/* MVME37x - Minor device 5 */
	"/dev/ci/m37x_6",	/* MVME37x - Minor device 6 */
	"/dev/ci/m37x_7",	/* MVME37x - Minor device 7 */
	"/dev/ci/m37x_8"	/* MVME147 - Minor device 8 */
};

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_register()	
**
**  DESCRIPTION:	Register the caller with the designated device
**			with the indicated IP or allow the CI to
**			assign the IP if the user doesn't indicate one.
**
**  CONTENTS:
**
**  PARAMETERS:		
**
**	Input:		dev		- 37x CPU number,
**			id		- ID pointer,
**			fd		- File descriptor address.
**
**	Output:		NONE
**
**	Returns:	0		- process is registered,
**					  otherwise, it is not.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  06/05/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.2	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

int
ci_register(dev, id, fd)

register unsigned short	dev;			/* 37x CPU number */
register unsigned long	*id;			/* ID pointer */
register int	  *fd;				/* File descriptor address */
{
	register unsigned long	src_id;		/* Source ID */
	register int	  n;			/* Read count */
	register int	  lfd;			/* File descriptor */
	register CE_EPB	  *epbptr;		/* CE EPB pointer */

	/*
	**	Check the minor device number.
	*/

	if ((dev < CPU_OFFSET) || (dev > NUM_CPU_ENTRIES))
		return (-1);

	/*
	**	Open the 37x device
	*/

	if ((lfd = open(dev_tbl[dev - 2], O_RDWR)) < 0)
		return (errno);

	/*
	**	Read the default registration response EPB
	*/

	(epbptr = &epb)->hdr.epb_size = sizeof(CE_EPB);

	epbptr->hdr.buf = (BUF *)0;

	if ((n = ci_read(epbptr, lfd)) != 0)
		return (n);

	/*
	**	If asked, demand the indicated IP.
	*/

	if ((src_id = *id & 0xFFFF) != 0)
	{
		/*
		**	Deregister from the default IP.
		*/

		if ((n = ci_dereg(epbptr->hdr.dst_id, lfd)) != 0)
			return (n);

		/*
		**	Format the registration request EPB.
		*/

		epbptr->hdr.next     = (EPB *)0;
		epbptr->hdr.mem_id   = 0;
		epbptr->hdr.src_id   = src_id;
		epbptr->hdr.dst_id   = PUT_ID(LOCAL_CPU, CE_IP);
		epbptr->hdr.epb_size = sizeof(CE_EPB);
		epbptr->hdr.type     = CE_REGISTRATION_REQUEST;
		epbptr->hdr.status   = NO_ERROR;
		epbptr->hdr.buf      = (BUF *)0;
		epbptr->hdr.user_def = 0;

		/*
		**	Write the request to the CI.
		*/

		while ((n = write(lfd, epbptr, epbptr->hdr.epb_size)) <= 0)
			if (n < 0)
				return (errno);

		/*
		**	Read the response from the CI.
		*/

		if ((n = ci_read(epbptr, lfd)) != 0)
			return (n);
	}

	/*
	**	Set the users variables.
	*/

	*fd = lfd;
	*id = epbptr->hdr.dst_id;

	return (0);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_deregister()	
**
**  DESCRIPTION:	Deregister the caller from the indicated
**			device.
**
**  PARAMETERS:		
**
**	Input:		id		- indicated device,
**			fd		- File descriptor.
**
**	Output:		NONE
**
**	Returns:	0		- process is registered,
**					  otherwise, it is not.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  06/05/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.2	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

int
ci_deregister(id, fd)

register unsigned long	id;		/* Indicated device */
register int 	  fd;			/* File descriptor */
{
	register int	n;		/* Deregistration return */

	/*
	**	Deregister the device and close it.
	*/

	if ((n = ci_dereg(id, fd)) == 0)
		close(fd);

	return (n);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_dereg()	
**
**  DESCRIPTION:	Deregister the caller from the indicated device.
**			Do not close the device.
**
**  PARAMETERS:		
**
**	Input:		id		- Indicated device,
**			fd		- File descriptor.
**
**	Output:		NONE
**
**	Returns:	0		- process is deregistered,
**					  otherwise, it is not.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:		This function is only used internally. It is
**			not available to the user.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  06/05/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.2	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

static int
ci_dereg(id, fd)

register unsigned long	 id;		/* Indicated device */
register int	  fd;			/* File descriptor */
{
	register CE_EPB	 *epbptr;	/* CE EPB pointer */
	register int	 n;		/* Return value */

	/*
	**	Format the deregistration request EPB.
	*/

	epbptr = &epb;

	epbptr->hdr.next     = (EPB *)0;
	epbptr->hdr.mem_id   = 0;
	epbptr->hdr.src_id   = id;
	epbptr->hdr.dst_id   = PUT_ID(LOCAL_CPU,CE_IP);
	epbptr->hdr.epb_size = sizeof(CE_EPB);
	epbptr->hdr.type     = CE_DEREGISTRATION_REQUEST;
	epbptr->hdr.status   = NO_ERROR;
	epbptr->hdr.buf      = (BUF *)0;
	epbptr->hdr.user_def = 0;

	/*
	**	Write the request to the CI.
	*/

	while ((n = write(fd, epbptr, epbptr->hdr.epb_size)) <= 0)
		if (n < 0)
			return (errno);

	return (0);
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_read()	
**
**  DESCRIPTION:	Read the response EPB and check for valid
**			status.
**
**  PARAMETERS:		
**
**	Input:		epbptr		- CE EPB pointer,
**			fd		- File descriptor address.
**
**	Output:		NONE
**
**	Returns:	0		- valid response EPB,
**					  otherwise, it is not.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:		This function is only used internally. It is
**			not available to the user.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  06/05/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.2	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

static int
ci_read(epbptr, fd)

register CE_EPB	 *epbptr;		/* CE EPB pointer */
register int	 fd;			/* File descriptor */
{
	register int	n;		/* Return value */

	/*
	**	Read response EPB.
	*/

	while ((n = read(fd, epbptr, epbptr->hdr.epb_size)) <= 0)
		if (n < 0)
			return (errno);

	/*
	**	Check for valid response.
	*/

	if (epbptr->hdr.type != CE_REGISTRATION_RESPONSE)
		return (-1);
	else
		if (epbptr->hdr.status != 0)
			return ((int)epbptr->hdr.status);

	return (0);
}
