/*@(#)bp_get_mpb.c	35.1*/
/*
**  This is a prototype of a possible BPPIR buffer management routine
**  which allocates MPBs (Message Parameter Blocks) using a modified 
**	form of the EPB style of buffer management.  It supports optional
**  allocation limit "windows" for flow control.
**
**	The design of this routine assumes that some higher level function
**	will provide the required mapping to specific pool ID and window
**	ID.  In the most rudimentary case, window ID could simply be the
**	same as IP.  The use of a window ID (wid) that is separate and distinct
**	from the IP allows a greater variety of flow control requirements to
**  be supported at the expense of only slightly increased overhead.
**	Windows may be mapped in a one-to-one correspondence to IPs.  They
**	could also be set up individually by pool_id and IP, allowing a 
**  separate window for each pool and IP which uses it.  Or several IPs
**  could share a common window, thus limiting the combined acquisition
**  of buffers by those IPs.
**
*/
#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/mvme147/bpp_local.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/bm.h>

BP_MPB_T	*bp_get_mpb(pool_id, wid, func_code, err, func_arg)
	POOL_ID		pool_id;
	BM_WID_T	wid;
	ULONG		func_code;
	ULONG		*err;
	ULONG		func_arg;		/* optional argument when function is used */
{
	extern	 int		bpp_disable_level;
	extern	 BM_PCB_T	*bp_pcb_array[];
	extern	 BM_WINDOW	bm_windows[];
	extern	 BM_PEND_T	*bp_get_pend();

	register BM_PCB_T	*pcb;
	register BP_SYSHDR_T	*mpb;
	register BP_MPB_T	*check_ptr;
	register BM_WINDOW	*wp;
	register BM_WID_T	rel_wid;
	BM_PEND_T			*pptr;
	int					pri;

	/*
	**	Check input parameters
	**
	**  NOTE:  	This function returns the address of a single MPB if one
	**			is available.  An optional interface routine could be 
	**			provided to allocate a chain of MPBs, if necessary.
	*/

	if (pool_id >= BP_MAX_POOLS )
	{
		*err = BM_BAD_POOL_ID;
		return ((BP_MPB_T *) NULL_PTR);
	}
	if (wid > BP_MAX_WINDOWS)
	{
		*err = BM_INVALID_WINDOW;
		return ((BP_MPB_T *) NULL_PTR);
	}
	*err = BM_SUCCESS;			/* return status is initially ok */

	pcb = bp_pcb_array[pool_id];

	/*
	**	NOTE:	This function call traverses the list of MPB's released by
	**			the remote processor to update window counts.  We do this
	**			before trying to allocate the buffer to make sure that the
	**			window status is current.
	*/

	bp_rel_wind(pool_id);

	bp_wakeup(pool_id);

	/*
	**  Check limits for window that wants to get a buffer.  If the
	**  window is used up, we deny the request, either returning or
	**  queueing the request in the window control block depending upon
	**  the calling arguments for the request.  The window is checked first
	**  since requests blocked because of window availability are queued
	**  to the window control block and are moved to the service queue
	**  in the pcb when released buffers are processed in bp_rel_wind().
	**	If the request is blocked because of buffer availability rather
	**  than window, it is placed directly on the pcb service queue.
	*/

	wp = &bm_windows[wid];
	DISABLE(pri, bpp_disable_level);
	if ( wp->maximum != 0 )
	{
		if ((wp->current + 1) >= wp->maximum)
		{
			if (func_code == NO_PEND)
			{ /* no waiting -- return error */
				ENABLE(pri);
				*err = BM_NO_MPB_AVAIL;
				return ((BP_MPB_T *)NULL_PTR);
			}
			else if (func_code == STD_PEND)
			{ /* suspend processing until buffer is available */
				ENABLE(pri);
				return ((BP_MPB_T *)NULL_PTR);
			}
			else
			{ /* we have a function to call when buffer is available */
				if (pptr = (BM_PEND_T *) bp_get_pend())
				{
					pptr->wind_id = wid;
					pptr->func_code = (PTF) func_code;
					pptr->func_arg = func_arg;
					pptr->next_pend = (BM_PEND_T *) NULL_PTR;
					if (wp->next_pend == (BM_PEND_T *) NULL_PTR)
					{
						wp->next_pend = pptr;
						wp->pend_tail = pptr;
					}
					else
					{
						wp->pend_tail->next_pend = pptr;
						wp->pend_tail = pptr;
					}
				}
				else
				{
					*err = BM_NO_MPB_AVAIL;
				}
				ENABLE(pri);
				return ((BP_MPB_T *)NULL_PTR);
			}
		}
	}

	/*
	**	Check to see if we have a buffer available
	*/
	mpb = (BP_SYSHDR_T *) LOCAL_ADDRESS(pcb->get_ptr);

	if ((mpb = mpb->bmh_link) == (BP_SYSHDR_T *) NULL)
	{
		if (func_code == NO_PEND)
		{ /* no waiting -- return error */
			ENABLE(pri);
			*err = BM_NO_MPB_AVAIL;
			return ((BP_MPB_T *)NULL_PTR);
		}
		else if (func_code == STD_PEND)
		{ /* suspend processing until buffer is available */
			ENABLE(pri);
			return ((BP_MPB_T *)NULL_PTR);
		}
		else
		{ /* we have a function to call when buffer is available */
			if (pptr = bp_get_pend())
			{
				pptr->wind_id = wid;
				pptr->func_code = (PTF) func_code;
				pptr->func_arg = func_arg;
				pptr->next_pend = (BM_PEND_T *) NULL_PTR;
				if (pcb->pend_list.next_pend == (BM_PEND_T *) NULL_PTR)
				{
					pcb->pend_tail->next_pend = pptr;
					pcb->pend_list.next_pend = pptr;
				}
				else
				{
					pcb->pend_tail->next_pend = pptr;
					pcb->pend_tail = pptr;
				}
			}
			else
			{
				*err = BM_NO_MPB_AVAIL;
			}
			ENABLE(pri);
			return ((BP_MPB_T *)NULL_PTR);
		}
	}

	mpb = (BP_SYSHDR_T *) LOCAL_ADDRESS(mpb);
	wp->current++;
	pcb->free_count--;
	pcb->get_ptr = mpb->bmh_link;
	ENABLE(pri);

	mpb->buf_status &= ~ BUF_IN_POOL;
	mpb->window_id = wid;
	return ((BP_MPB_T *) mpb);			/* mpb is mapped to local address space */
}
