/*	@(#)creat_pool.c	35.1	*/
/* creat_pool.c
**
**	FUNCTION:	bp_create_mpb_pool()
**
**	ARGUMENTS:	memory_addr	= 32-bit address of place to build pool
**				memory_size	= 32-bit size of memory area for pool
**				buffer_size	= 16-bit size of buffers to build
**				remote_cpu	= 16-bit remote CPU number
**				remote_bufs	= number of buffers for remote CPU's subpool
**				local_bufs	= number of buffers for local  CPU's subpool
**				pool_id		= address of 8-bit variable containing the
**							  requested pool ID number or zero. If zero,
**							  pool ID will be assigned by Buffer Manager
**							  and will be returned in pool_id variable.
**
**	RETURNS:	32-bit return code value
**
**	PROCEDURE:	This function attempts to create a buffer pool for the
**				requester.  If the remote_cpu argument is non-zero, the
**				pool is assumed to be a shared memory buffer pool.  If
**				the memory_addr argument is non-zero, it is assumed to 
**				be the address of an area of shared or private memory in
**				which the requester wants the pool to be created, in 
**				which case, the memory_size argument is verified to be
**				large enough to support the number and size of local and
**				remote buffers specified.  If the memory_addr argument
**				is null, then this function calls a driver interface 
**				function, get_memory(), to acquire the amount of memory 
**				needed for the buffer pool and its control blocks.
**
**				Once the required memory is available, it is formatted
**				into the specified number of local and remote buffers.
**				If a remote_cpu was specified, then one of the local
**				buffers is used to format a BM_INIT_POOL_REQ message to
**				the desired CPU.  The message is sent (via pipe_write)
**				and then the message "type" field is polled until the
**				remote CPU sets it to BM_INIT_POOL_RSP.  The response_wait
**				and retry_count fields of the pipe control table are
**				used to determine how long this function will wait for
**				the remote CPU to process the initialize pool request.
*/

#include <sys/bppv2/bpptypes.h>
#include <sys/bppv2/mvme147/bpp_local.h>
#include <sys/bppv2/bpp.h>
#include <sys/bppv2/ccb.h>
#include <sys/bppv2/bm.h>

#define	PRIVATE		0x01		/* private buffer pool definition */
#define	SHARED		0x02		/* shared buffer pool definition */
ULONG	bp_create_mpb_pool(memory_addr,memory_size,buffer_size,remote_cpu,
							remote_bufs,local_bufs,pool_id)
ULONG	memory_addr;
ULONG	memory_size;
USHORT	buffer_size;
USHORT	remote_cpu;
USHORT	remote_bufs;
USHORT	local_bufs;
POOL_ID *pool_id;
{
	extern		BM_PCB_T		*bp_pcb_array[];
	extern		CPU_ID			bpp_local_cpu;
	extern		PCT_T			bp_local_pipe_tbl[];
	extern		LONG			bp_get_pipe_status();
	extern		UBYTE			*bp_get_memory();
	extern		void			bp_free_memory();

	register	BM_INREQ_T		*mpb;
	register	BP_SYSHDR_T		*buf_ptr;
	register	UBYTE			*mem_ptr;
	register	BM_PCB_T		*pcb = 0;
	register	ULONG		memory_needed;
	register	POOL_ID		pool = *pool_id;
	register	UBYTE		buf_status= SHARED_BUF + BUF_IN_POOL;
				UBYTE		pool_type = PRIVATE;
				ULONG		remote_resp;
	register	int			i;

	/*
	**	See if caller wants a specific pool ID
	*/
	if (pool > BP_MAX_POOLS)
	{ /* not configured for this many pools */
		return(BM_BAD_POOL_ID);
	}
	if (pool != NULL)
	{
		if (bp_pcb_array[pool] != (BM_PCB_T *)NULL)
			return(BM_BAD_POOL_ID);		/* pool ID already in use */
	}
	else
	{
		/*
		**	Make sure there's a pool ID available on this CPU
		*/
		for (; pool < BP_MAX_POOLS; pool++)
		{
			if ((pcb = bp_pcb_array[pool]) == (BM_PCB_T *)NULL)
				break;
		}
	}
	if (pcb)
		return(BM_BAD_POOL_ID);

	/*
	**	Check that buffer size is large enough for minimum MPB
	*/
	if (buffer_size < sizeof(BP_MPB_T))
		return(BM_CREATE_PARM);

	/*
	**	Calculate amount of memory required for pool and PCB
	*/
	if (local_bufs == 0)
		local_bufs = 1;		/* must have at least one buffer in a pool */
	memory_needed = sizeof(BM_PCB_T) + (buffer_size*(local_bufs+remote_bufs));

	if (remote_cpu != NULL)
	{
		pool_type = SHARED;
		buf_status = SHARED + BUF_IN_POOL;
		/*
		**	Make sure pipe exists to remote CPU
		*/
		if (bp_get_pipe_status(remote_cpu | BP_CPU_INDICATOR) & (ALIVE | RUNNING) 
				!= (ALIVE + RUNNING))
		{
			return(BPP_NO_PIPE);
		}

		/*
		**  Make sure user-provided memory is shared memory
		*/
		if (remote_cpu)
		{ /* if a remote cpu is involved, must use shared memory */
			if (memory_addr != NULL)
			{
				if (bp_is_private(memory_addr))
				{
					return(BM_CREATE_PARM);
				}
				if (bp_is_private(memory_addr+memory_size))
				{
					return(BM_CREATE_PARM);
				}
			}
		}
	}

	/*
	**  If user provided memory for pool
	*/
	if (memory_addr != NULL)
	{
		/*
		**	Verify that user-provided memory is big enough
		*/
		if (memory_size < memory_needed)
		{
			return(BM_INSUFF_MEMORY);
		}
		mem_ptr = (UBYTE *)memory_addr;
	}
	else
	{
		/*
		**  User did not provide memory for pool: go try to get it
		*/
		if ((mem_ptr = (bp_get_memory(pool_type, memory_needed))) == NULL)
		{
			return(BM_INSUFF_MEMORY);
		}

		/*
		**	We got it OK
		*/
	}

	/*
	**	We now have the memory required to build the buffer pool
	*/
	*pool_id = pool;		/* tell user what pool ID we've assigned */
	pcb = (BM_PCB_T *)mem_ptr;
	bp_pcb_array[pool] = pcb;
	mem_ptr += sizeof(BM_PCB_T);
	buf_ptr = (BP_SYSHDR_T *)mem_ptr;

	/* 
	**	Initialize Pool Control Block
	*/
	pcb->pool.cpu = remote_cpu;		/* remote cpu -- zero, if local */
	pcb->pool.pool_id 		= pool;
	pcb->buf_size 	= buffer_size;
	pcb->get_ptr  	= buf_ptr;
	pcb->start	  	= (BYTE *) buf_ptr;
	pcb->last_proc 	= buf_ptr;
	pcb->pool_size 	= local_bufs*buffer_size;
	pcb->orig_count = local_bufs;
	pcb->free_count = local_bufs - 1;
	pcb->last_alloc = (BP_SYSHDR_T *)NULL;

	/*
	**  Format local MPB's first
	*/
	for (i = 0; i < local_bufs; i++)
	{
		buf_ptr->creator.cpu 	 = bpp_local_cpu;
		buf_ptr->creator.pool_id = pool;
		buf_ptr->manager.cpu 	 = bpp_local_cpu;
		buf_ptr->manager.pool_id = pool;
		buf_ptr->partnercpu		 = remote_cpu;
		buf_ptr->buf_status	 		 = buf_status;
		buf_ptr->mpb_size 		 = buffer_size;
		buf_ptr->window_id  	 = NULL;
		buf_ptr->bmh_link 		 = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
		buf_ptr = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
	}
	buf_ptr->bmh_link = NULL;
	pcb->rel_ptr = buf_ptr;
	buf_ptr = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
	pcb->end = (BYTE *)buf_ptr;

	if ( remote_cpu > 0 && remote_bufs > 0 )
	{
		/*
		**	Use first local buffer for BM_INIT_POOL_REQ MPB
		*/
		mpb = (BM_INREQ_T *) pcb->get_ptr;
		pcb->get_ptr = mpb->hdr.sys_hdr.bmh_link;
		pcb->last_alloc	= (BP_SYSHDR_T *) mpb;
		/*
		**  Format MPB as an BM_INIT_POOL_REQ
		*/
		mpb->hdr.dst_id.cpu  	 = remote_cpu;
		mpb->hdr.dst_id.ip  	 = BP_BFM_IP;
		mpb->hdr.src_id.cpu 	 = bpp_local_cpu;
		mpb->hdr.src_id.ip  	 = BP_BFM_IP;
		mpb->hdr.msg_status 		 = 0;
		mpb->hdr.data_size 		 = sizeof(BM_INIT_POOL_T)-sizeof(BP_MPB_T);
		mpb->hdr.data	  		 = (UBYTE *)(mpb + sizeof(BP_MPB_T));
		mpb->type	   		 = BM_INIT_POOL_REQ;
		mpb->creator_pool_id.cpu   	 = bpp_local_cpu;
		mpb->creator_pool_id.pool_id = pool;
		mpb->manager_pool_id.cpu   	 = remote_cpu;
		mpb->manager_pool_id.pool_id = 0;
		mpb->hdr.sys_hdr.buf_status &= ~BUF_IN_POOL;
		mpb->resp_code	   	 = BM_SUCCESS;
		mpb->buf_size = buffer_size;
		mpb->count 	 = remote_bufs;
		mpb->get_buf		 = (BP_MPB_T *) buf_ptr;
		mpb->first_buf = (BP_MPB_T *) buf_ptr;

		/*
		**  Format remote MPB's if this is not just a local pool
		*/
		for (i = 0; i < remote_bufs; i++)
		{
			buf_ptr->creator.cpu 	 = bpp_local_cpu;
			buf_ptr->creator.pool_id = pool;
			buf_ptr->manager.cpu 	 = remote_cpu;
			buf_ptr->manager.pool_id = pool;
			buf_ptr->partnercpu	 	 = bpp_local_cpu;
			buf_ptr->buf_status	 		 = buf_status;
			buf_ptr->mpb_size	 	 = buffer_size;
			buf_ptr->window_id		 = NULL;
			buf_ptr->bmh_link 		 = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
			buf_ptr = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
		}
		buf_ptr->bmh_link = NULL;
		mpb->ret_buf  = (BP_MPB_T *) pcb->rel_ptr;
		pcb->rel_ptr  = buf_ptr;
		buf_ptr = (BP_SYSHDR_T *) ((BYTE *) buf_ptr + buffer_size);
		mpb->last_buf	  = (BP_MPB_T *) buf_ptr;

		/*
		**	Send BM_INIT_POOL_REQ to remote processor
		*/
		bp_pwrite(bp_local_pipe_tbl[remote_cpu], mpb);

		/*
		**	Wait for remote CPU to process request and respond
		*/
		while (mpb->type != BM_INIT_POOL_RSP
				&& mpb->hdr.msg_status == BPP_SUCCESSFUL)
		{
		/*	
		**	Do something here to delay a reasonable length of time
		**	to allow the remote CPU to process the request.  However,
		**	don't wait forever.  If he doesn't respond in the specified
		**	time, with possibly a retry option, then give up and return
		**	the negative response to the caller.
		*/
		}

		if ((remote_resp = mpb->resp_code) != BM_SUCCESS)
		{
			if (memory_addr == 0)	/* Did we provide memory? */
			{
				bp_free_memory(pcb, memory_needed);
			}
			return(remote_resp);

		}
	} /* End of conditional block for remote pool creation */

	return(BM_SUCCESS);
}
