/*	@(#)ci_bufmgt.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_bufmgt.c
**
**  DESCRIPTION:	This module contains the buffer management
**			functions of the UNIX driver for MVME37x IPCs
**			using the BPP and BPPIR protocols.
**
**  FUNCTIONS:		ci_get_bpe, ci_rel_bpe,
**			ci_get_epb, ci_rel_epb,
**			ci_get_buf, ci_rel_buf
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  09/09/87	F. Christiansen 2.0	Changed spl's to use intr_level.
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	 DEFAULT	 splx();	/* Set priority level */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_get_bpe()	
**
**  DESCRIPTION:	Get a shared BPE from the shared BPE pool.
**
**  CONTENTS:		This function will return the address of a
**			BPE from the shared BPE pool. If a BPE is not
**			available, a NULL pointer is returned.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry.
**
**	Output:		NONE
**
**	Returns:	Pointer to BPE.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

BPE *
ci_get_bpe( c )
	REG	CPU_ENT	*c;		/* CPU table entry address */
{
	REG	BPE	*bpe;		/* BPE pointer */
	REG	DEFAULT	pri;		/* Current priority level */
	
	/*
	**	Set default pointer to NULL.
	*/

	bpe = ( BPE * ) NULL;

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Check BPE flags. If this is not the last BPE
	**	then return it.
	*/

	if( ( c->lc.get_bpe->flags & 0x0001 ) == 0 )
	{
		/*
		**	Update list pointer.
		*/

		bpe = c->lc.get_bpe;

		c->lc.get_bpe = bpe->next;

		/*
		**	Fill in BPE contents.
		*/

		bpe->next   = ( BPE * ) NULL;
		bpe->msg    = ( EPB * ) NULL;
		bpe->dst_id = 0;
		bpe->src_id = 0;
		bpe->flags  = 0;
	}

	/*
	**	Enable interrupts and return BPE address.
	*/

	splx( pri );

	return( bpe );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_rel_bpe()	
**
**  DESCRIPTION:	Return a shared BPE to the shared BPE pool.
**
**  CONTENTS:		This function will return the BPE to the
**			shared BPE pool.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			bpe		- BPE address.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_rel_bpe( c, bpe )
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	BPE	*bpe;		/* BPE address */
{
	REG	DEFAULT pri;		/* Current priority level */

	/*
	**	Clear BPE link and mark it as new terminator
	*/

	bpe->next   = ( BPE * ) NULL;
	bpe->flags |= 0x0001;

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Attach the BPE to the pool, unmark old terminator, and
	**	move the tail pointer.
	*/

	c->lc.ret_bpe->next   = bpe;
	c->lc.ret_bpe->flags &= 0xFFFE;
	c->lc.ret_bpe         = bpe;

	/*
	**	Enable interrupts.
	*/

	splx( pri );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_get_epb()	
**
**  DESCRIPTION:	Get a shared EPB from the local shared EPB pool.
**
**  CONTENTS:		This function will return the address of a
**			EPB from the local shared EPB pool. If a EPB
**			is not available, a NULL pointer is returned.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry.
**
**	Output:		NONE
**
**	Returns:	Pointer to EPB.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

EPB *
ci_get_epb( c )
	REG	CPU_ENT *c;		/* CPU table entry address */
{
	REG	EPB	*epb;		/* EPB pointer */
	REG	DEFAULT pri;		/* Current priority level */
	
	/*
	**	Set default pointer to NULL.
	*/

	epb = ( EPB * ) NULL;

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Check EPB flags. If this is not the last EPB
	**	then return it.
	*/

	if( c->lr.epb_head->user_def != TERMINATOR )
	{
		/*
		**	Update list pointer.
		*/

		epb = c->lr.epb_head;

		c->lr.epb_head = epb->next;

		/*
		**	Fill in EPB contents.
		*/

		epb->next     = ( EPB * ) NULL;
		epb->dst_id   = 0;
		epb->src_id   = 0;
		epb->epb_size = 0;
		epb->type     = 0;
		epb->status   = 0;
		epb->buf      = ( BUF * ) NULL;
		epb->user_def = 0;
	}

	/*
	**	Enable interrupts and return EPB address.
	*/

	splx( pri );

	return( epb );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_rel_epb()	
**
**  DESCRIPTION:	Return a shared EPB to the remote shared
**			EPB pool tail or to the local shared EPB
**			pool.
**
**  CONTENTS:		This function will return the EPB to the
**			remote shared EPB pool if it belongs to
**			that pool or it will return it to the local
**			shared EPB pool if it belongs to that pool. 
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			epb		- EPB address.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_rel_epb( c, epb )
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	EPB	*epb;		/* EPB address */
{
	REG	DEFAULT pri;		/* Current priority level */

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Use remote portion to determine which of the remote
	**	or local pools it goes into. Update the list and
	**	mark it as a member or terminator.
	*/

	if( OWNER( epb->mem_id ) == c->lc.local_cpu )
	{
		epb->user_def = MEMBER;
		epb->next     = c->lr.epb_head;

		c->lr.epb_head = epb;
	}
	else
	{
		epb->user_def = TERMINATOR;
		epb->next     = (EPB *)NULL;

		c->lr.epb_tail->next     = epb;
		c->lr.epb_tail->user_def = MEMBER;
		c->lr.epb_tail           = epb;
	}

	/*
	**	Enable interrupts.
	*/

	splx( pri );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_get_buf()	
**
**  DESCRIPTION:	Get a shared BUF from the shared BUF pool.
**
**  CONTENTS:		This function will return the address of a
**			BUF from the shared BUF pool. If a BUF is not
**			available, a NULL pointer is returned.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry.
**
**	Output:		NONE
**
**	Returns:	Pointer to BUF.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

BUF *
ci_get_buf(c)
	REG	CPU_ENT	*c;		/* CPU table entry address */
{
	REG	BUF	*buf;		/* BUF pointer */
	REG	DEFAULT pri;		/* Current priority level */
	
	/*
	**	Set default pointer to NULL.
	*/

	buf = ( BUF * ) NULL;

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	If this is not the last BUF then return it.
	*/

	if( c->lr.buf_head->next != ( BUF * ) NULL)
	{
		/*
		**	Update list pointer.
		*/

		buf = c->lr.buf_head;

		c->lr.buf_head = c->lr.buf_head->next;

		/*
		**	Fill in BUF contents.
		*/

		buf->next   = ( BUF * ) NULL;
		buf->length = 0;
	}

	/*
	**	Enable interrupts and return BUF address.
	*/

	splx( pri );

	return( buf );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_rel_buf()	
**
**  DESCRIPTION:	Return a shared BUF to the shared BUF pool.
**
**  CONTENTS:		This function will return the BUF to the
**			shared BUF pool.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			buf		- BUF address.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_rel_buf( c, buf )
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	BUF	*buf;		/* BUF address */
{
	REG	DEFAULT pri;		/* Current priority level */

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Attach the BUF to the pool and move the tail pointer.
	*/

	c->lr.buf_tail->next = buf;
	c->lr.buf_tail       = buf;

	buf->next = (BUF *)NULL;

	/*
	**	Enable interrupts.
	*/

	splx( pri );
}
