/*MH*********************** MODULE HEADER *****************************
**
** @(#)ci_driver.c	35.2
**
**  MODULE NAME:	ci_driver.c
**
**  DESCRIPTION:	This module contains the functions composing 
**			the Unix driver for MVME37x IPCs using the
**			BPP and BPPIR protocols.
**
**  RESTRICTIONS:	This driver cannot be compiled with the -T
**			(truncate) option.  It must be compiled and
**			linked seperately.
**
**  FUNCTIONS:		c37xinit,  c37xopen,  c37xclose, c37xread,
**			c37xwrite, c37xioctl, c37xintr
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  03/20/86	Dana Holgate	1.1	Initial Coding.
**  05/01/87	E. Skinner	1.2	Port to System V/68 Release 3.
**  05/16/87	E. Skinner	1.3	Modified c37xioctl.
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x.
**  06/09/87	J. Richey	1.5	Modified c37xread, c37xintr,
**					and c37xioctl.
**  06/29/87	R. Robinson	1.6	Modified c37xwrite.
**  07/15/87	R. Robinson	1.7	Modified c37xopen, c37xclose,
**					c37xioctl, and c37xintr.
**  11/04/87	F. Christiansen		(see SCCS comments)
**
***********************************************************************/

#include <sys/errno.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/signal.h>
#include <sys/dir.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/sysmacros.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/mvmecpu.h>
#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	EPB	*ci_get_epb();		/* Get an EPB */
extern	BUF	*ci_get_buf();		/* Get a BUF */
extern	EPB	*ci_pread();		/* Pipe read */
extern	EPB	*ci_uread();		/* Read from user's queue */
extern	VOID	ci_uwrite();		/* Write to user's queue */
extern	VOID	ci_unread();		/* Unread from the user' queue */
extern	VOID	ci_add_retry();		/* Something's wrong */
extern	VOID	ci_swap_id();		/* Swap EPB IDs */
extern	VOID	ci_ret_rel();		/* Return/release EPBs and BUFs */
extern	VOID	ci_rel_epb();		/* Release EPB */
extern	VOID	ci_rel_buf();		/* Release BUF */
extern	VOID	ci_ret_send();		/* Return EPB to sender */
extern	VOID	ci_reg_init();		/* Initialize registration table */
extern	VOID	ci_rc_init();		/* Initialize remote channel */
extern	VOID	ci_mem_init();		/* Initialize CI memory */
extern	DEFAULT	ci_pwrite();		/* Pipe write */
extern	DEFAULT	ci_reg();		/* Register/Deregister process */
extern	BOOLEAN	ci_malloc();		/* Allocate CI memory */
extern	BOOLEAN	ci_pipe_init();		/* Initialize pipe */
extern	BOOLEAN	ci_dst_ok();		/* Check destination process */
extern	BOOLEAN	ci_find_empty();	/* Find empty registration entry */
extern	BOOLEAN	ci_is_reg();		/* Check for registered process */

extern	DEFAULT	splx();			/* Restore interrupts */
extern	VOID	delay();		/* Delay N seconds */
extern	VOID	psignal();		/* Signal user */
extern	DEFAULT	copyin();		/* Copy data from user's space */
extern	DEFAULT	copyout();		/* Copy data to user's space */
extern	BOOLEAN	bprobe();		/* Probe address */
extern  int     cpuboard;               /* cpu board type */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xinit()	
**
**  DESCRIPTION:	Initializes the 37x buffer pools, tables,
**			and pipe.
**
**  CONTENTS:		This function will allocate memory for the
**			BPE, EPB and BUF pools as well as the remote
**			channel table. It will then initialize the
**			channel and registration tables. It will also
**			"bprobe" for the board, and if found, initialize
**			the pipe.
**
**  PARAMETERS:		
**
**	Input:		NONE
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:	If this function is being compiled on a VME10 system,
**		it will call the function ci_mvme10() function to adjust
**		the VME10's control registers.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**
**  01/19/87	J. Holbrook	1.5	Added hooks for VME147 support.
**********************************************************************/

VOID
c37xinit()
{
	register DEFAULT i;		/* Just a counter */
	extern	UBYTE	*bppv2_local_cpu;

	/*
	**	Next, loop through the CPU entries in the CPU
	**	table, filling each entry.  We never initialize the special
	**  mvme147 CPU entry from here, since it is in the same CPU
	**  and has no PIPE initialization in th c37x driver sense.  The
	**	entry for it is initialized from the BPPV2 initialization.
	*/

	for (i = 0; i < NUM_CPU_ENTRIES - CPU_OFFSET; ++i)
	{
		/*
		**	Check for missing device.
		*/

		if (i < c37x_cnt)
			ci_cpu[i].boardadr = (ULONG)c37x_add[i];
		if (ci_debug)
			printf("c37x: looking for cpu %d at %x\n", 
				i, ci_cpu[i].boardadr);
		
		if ((i < c37x_cnt) && (bprobe(ci_cpu[i].boardadr, -1) == FALSE))
		{
			/* probe the registers to determine board type - use
			   374 as a starting point.  DO NOT try to access board
			   before regbase is initialized. Subtract 0x80 to ensure
			   that we don't probe interrupting ram on 374.
			 */
			if(bprobe(ci_cpu[i].boardadr + SROFFSET + MVME374_OFFSET - 0x80, -1) == FALSE)
			  ci_cpu[i].regbase = MVME374_OFFSET;
			else
			  ci_cpu[i].regbase = MVME372_OFFSET;

			/*
			**	Allocate memory for 2 BPE pools, 2 EPB
			**	pools, 1 BUF pool and remote channel
			**	table.
			*/

			if (ci_malloc(i, (ULONG)MEM_POOL_SZ) == SUCCESS)
			{
				ci_cpu[i].no_ipc = FALSE;

				/*
				**	Next, create the BPE, EPB and BUF
				**	pools as well as the remote channel
				**	table.
				*/

				ci_mem_init(i);

				/*
				**	Initialize the channel tables and
				**	the registration table.
				*/

				ci_rc_init(i);

				ci_reg_init(i);

				/*
				**	Finally, create the message channel
				**	with the device.
				*/

				if (ci_pipe_init(i) == SUCCESS)
				{
					ci_cpu[i].no_com = FALSE;

					if (ci_debug) 
						printf("%sinit: CPU %d initialization complete\n", ci_dev, (i + CPU_OFFSET));
				}
				else
					ci_cpu[i].no_com = TRUE;
			}
			else
			{
				printf("%sinit: Unable to allocate memory for CPU %d\n", ci_dev, (i + CPU_OFFSET));

				ci_cpu[i].no_ipc = TRUE;
			}
		}
		else
		{
			if (ci_debug)
				printf("%sinit: c37x board %d is missing\n",
					ci_dev, (i + CPU_OFFSET));

			ci_cpu[i].no_ipc = ci_cpu[i].no_com = TRUE;	
		}
	}
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xopen()	
**
**  DESCRIPTION:	Registers the user's process ID in the
**			37x registration table.
**
**  CONTENTS:		See comments.
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  NOTES:		The spl()'s protect data than can get changed
**			from interrupt level.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  07/15/87	R. Robinson	1.7	Removed no_com flag check.
**  08/28/87	F. Christiansen x.x	Split off c37kopen code.
**
**********************************************************************/

VOID
c37xopen( dev )
	REG	DEFAULT	dev;		/* Minor device number */
{
	REG	CPU_ENT	*c;		/* CPU table entry address */
		BYTE	c37kopen();	/* Kernel level interface */
		USHORT	ip;		/* Interface point */
	REG	USHORT	pid;		/* Process id */
	extern	VOID	c37_rxint();	/* Receive handler */
	extern	UBYTE	ci_first_open;	/* driver first open flag */

	dev = minor( dev );		/* Get minor device number */

	if( ci_debug )
		printf("%sopen: opening minor device %d\n", ci_dev, dev );

	/*
	**	Generate CPU table entry address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	** If this is the first open, check whether the mvme147 lance support
	** library has declared its presence by initializing reference
	** variables used by the c37x driver.  If it hasn't, set the ci_cpu
	** entry for the 147 cpu to indicate no_ipc. 
	*/

	if (ci_first_open && (bppv2_local_cpu == 0))
	{
		ci_cpu[NUM_CPU_ENTRIES - CPU_OFFSET].no_ipc = ci_cpu[NUM_CPU_ENTRIES - CPU_OFFSET].no_com = TRUE;
		ci_first_open = FALSE;
	}

	/*
	**	Check to see if we can communicate with the IPC.  If the
	**	IPC is OK, check if the user is registered.  If registered,
	**	it is an error; if not, attempt to register.
	*/

	if( c->no_ipc == FALSE )
	{
		if( ci_is_reg( c, ( pid = u.u_procp->p_pid ), &ip ) == FALSE )
			u.u_error = c37kopen( c, &ip, pid, c37_rxint );
		else
			u.u_error = EEXIST;
	}
	else
		u.u_error = ENXIO;

	return;
}

BYTE
c37kopen( c, ip, pid, rx_int )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	*ip;		/* Interface point */
	REG	USHORT	pid;		/* Process id */
		VOID	(*rx_int)();	/* Receive handler */
{
	REG	BYTE	err;		/* Error code */
	REG	DEFAULT	pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration entry address */

	/*
	**	If there is an empty registration table entry, we can register.
	*/

	if( ci_find_empty( c, ip ) == TRUE )
	{
		pri = splx( SPLEV( c->rc->intr_level ) );

		r		= &( c->rt[*ip] );
		r->process_id	= pid;
		r->proc		= u.u_procp;
		r->block	= 0;
		r->reg_notified	= FALSE;
		r->rx_int	= rx_int;

		splx( pri );

		err = 0;
	}
	else
		err = ENOSPC;

	return( err );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xclose()	
**
**  DESCRIPTION:	Clear the 37x registration table.
**
**  CONTENTS:		This function will "close" communication with
**			the 37x IPC by clearing its registration table
**			(since this is the last user-invoked close).
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:		Kernel level users have negative process ID's.
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  07/15/87	R. Robinson	1.7	Removed no_com flag check.
**
**********************************************************************/

VOID
c37xclose( dev )
	REG	DEFAULT	dev;		/* Minor device number */
{
	REG	CPU_ENT *c;		/* CPU table entry address */
		VOID	c37kclose();	/* Kernel level interface */
	REG	USHORT	ip;		/* Interface point */

	dev = minor( dev );		/* Get minor device number */

	if( ci_debug )
		printf("%sclos: closing minor device %d\n", ci_dev, dev );

	/*
	**	Generate CPU table entry address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	**	Clear out all user ip's for this CPU.  This is the last close.
	*/

	if( c->no_ipc == FALSE )
		for( ip = 0; ip < NUM_REG_ENTRIES; ++ip )
			if( ( ( short ) c->rt[ip].process_id ) > 0 )
				c37kclose( c, ip );

	return;
}

VOID
c37kclose( c, ip )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface point */
{
	REG	REG_ENT	*r;		/* Registration table entry address */
	REG	DEFAULT	pri;		/* Current priority level */

	/*
	**	Clear the registration entry specified by ip for CPU c.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	r		= &( c->rt[ip] );
	r->process_id	= EMPTY;
	r->proc		= ( struct proc * ) NULL;
	r->block	= 0;
	r->reg_notified	= FALSE;
	r->rx_int	= NULL;

	splx( pri );

	/*
	**	Release anything pending.
	*/

	ci_ret_rel( c, ip );

	return;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xread()	
**
**  DESCRIPTION:	Read from the 37x IPC.
**
**  CONTENTS:		This function will read an EPB from the user's
**			queue for the given 37x IPC.
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  06/09/87	J. Richey	1.5	Added BLOCK flag check for NULL
**					read.
**
**********************************************************************/

VOID
c37xread( dev )
	REG	DEFAULT	dev;		/* Minor device number */
{
	REG	BOOLEAN	block;		/* Blocking flag */
	REG	BUF	*buf;		/* BUF pointer */
	REG	CPU_ENT *c;		/* CPU table entry address */
		EPB	*c37kread();	/* Kernel level interface */
		VOID	c37krel();	/* Kernel level interface */
	REG	EPB	*epb;		/* EPB pointer */
	REG	DEFAULT	pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */
	REG	ULONG	size;		/* EPB/BUF size */

		USHORT	ip;		/* Interface point */

	dev = minor( dev );		/* Get minor device number */
	
	if( ci_debug )
		printf("%sread: reading minor device %d\n", ci_dev, dev );

	/*
	**	Generate CPU table entry address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	**	Clear return error code.
	*/

	u.u_error = 0;

	/*
	**	Check to see if we can communicate with the IPC.  If the
	**	IPC is OK, check if the user is registered.  If not registered,
	**	it is an error.
	*/

	if( ( c->no_ipc == FALSE ) && ( c->no_com == FALSE ) )
	{
		if( ci_is_reg( c, u.u_procp->p_pid, &ip ) == FALSE )
		{
			u.u_error = EIDRM;
			return;
		}
	}
	else
	{
		u.u_error = ENXIO;
		return;
	}

	/*
	**	See if something is waiting.  If nothing, see if
	**	we should block.
	*/
	
	r = &( c->rt[ip] );

	if( ( epb = c37kread( c, ip ) ) == ( EPB * ) NULL )
	{
		if( ( ( block = r->block ) & ST_SIGNAL ) != 0 )
			block &= ~BLOCK;

		if( ( block & BLOCK ) != 0 )
		{
			sleep( r, ( PZERO + 1 ) );

			if( ( epb = c37kread( c, ip ) ) == ( EPB * ) NULL )
			{
				u.u_error = EINTR;
				return;
			}
		}
		else
			return;
	}

	/*
	**	Handle the registration response on the first read.
	*/

	size = epb->epb_size;

	if( r->reg_notified == FALSE )
	{
		if( u.u_count >= size )
		{
			if( copyout( epb, u.u_base, size ) == 0 )
			{
				pri = splx( SPLEV( c->rc->intr_level ) );

				r->reg_notified = TRUE;

				splx( pri );

				u.u_count -= size;
			}
			else
				u.u_error = EFAULT;
		}
		else
			u.u_error = EINVAL;

		return;
	}

	/*
	**	If we have a BUF, where will it go on copyout?
	**	And what is the total size?
	*/

	if( ( buf = epb->buf ) != ( BUF * ) NULL )
	{
		size += ( buf->length + BUF_HDR_SIZE );
	
		epb->buf = ( BUF * )( u.u_base + epb->epb_size );
	}

	/*
	**	Has the user provided sufficient input space?
	*/

	if( u.u_count < size )
	{
		u.u_error = EINVAL;

		return;
	}

	/*
	**	Copy EPB to user's space.
	*/

	if( copyout( epb, u.u_base, epb->epb_size ) != 0 )
	{
		if( buf != ( BUF * ) NULL )
			epb->buf = buf;

		ci_unread( c, ip, epb );

		u.u_error = EFAULT;

		return;
	}

	/*
	**	Copy BUF to user's space.
	*/

	if( buf != ( BUF * ) NULL )
		if( copyout( buf, epb->buf,
			( buf->length + BUF_HDR_SIZE ) ) != 0 )
		{
			epb->buf = buf;

			ci_unread( c, ip, epb );

			u.u_error = EFAULT;

			return;
		}

	/*
	**	Tell the user how many bytes copied.
	*/

	u.u_count -= size;

	/*
	**	Release the resources.
	*/

	c37krel( c, epb, buf );

	return;
}

EPB *
c37kread( c, ip )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface point */
{
	REG	EPB	*epb;		/* EPB pointer */
		EPB	*ci_uread();	/* Read of BPP user's read queue */

	/*
	**	Special registration response on the first read.
	*/

	if( c->rt[ip].reg_notified == FALSE )
	{
		/*
		**	Fill the static response EPB, other fields
		**	already initialized.
		*/

		epb		= &ci_reg_epb;
		epb->dst_id	= PUT_ID( c->lc.local_cpu, ip );
		epb->src_id	= PUT_ID( c->lc.local_cpu, CE_IP );

		return( epb );
	}

	/*
	**	Handle the regular reading of data from BPP or other
	**	process.
	*/
	
	return( ci_uread( c, ip ) );
}

VOID
c37krel( c, epb, buf )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	EPB	*epb;		/* EPB pointer */
	REG	BUF	*buf;		/* BUF pointer */
{
	if( buf != ( BUF * ) NULL )
		ci_rel_buf( c, buf );

	ci_rel_epb( c, epb );

	return;
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xwrite()	
**
**  DESCRIPTION:	Write to the 37x IPC.
**
**  CONTENTS:		This function will write an EPB to the user's
**			queue for the given 37x IPC.
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  05/01/87	E. Skinner	1.2	Port to System V/68 Release 3.
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  06/29/87	R. Robinson	1.6	Fixed bug in REG./DEREG.
**
**********************************************************************/

VOID
c37xwrite(dev)
	REG	DEFAULT dev;		/* Minor device number */
{
	REG	BUF	*buf;		/* BUF pointer */
	REG	CPU_ENT *c;		/* CPU table entry address */
		BYTE	c37kwrite();	/* Kernel level interface */
	REG	EPB	*epb;		/* EPB pointer */
	REG     BYTE    err;            /* return value from c37kwrite */
	REG	ULONG	mem_id;		/* Memory ID */
	REG	USHORT	pid;		/* Process ID */
	int		pri;

		USHORT	ip;		/* Interface point */

	dev = minor( dev );		/* Get minor device number */

	if( ci_debug )
		printf("%swrit: writing minor device %d\n", ci_dev, dev );

	/*
	**	Generate CPU table entry address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	**	Clear return error code.
	*/

	u.u_error = 0;

	/*
	**	Check to see if we can communicate with the IPC.  If the
	**	IPC is OK, check if the user is registered.  If not registered,
	**	it is an error.  If registered, check for reg notification.
	*/

	if( ( c->no_ipc == FALSE ) && ( c->no_com == FALSE ) )
	{
		if( ci_is_reg( c, ( pid = u.u_procp->p_pid ), &ip ) == TRUE )
		{
			if( c->rt[ip].reg_notified == FALSE )
			{
				u.u_error = EIO;
				return;
			}
		}
		/* else, OK to be unregistered if writing a registration EPB */
	}
	else
	{
		u.u_error = ENXIO;
		return;
	}

	/*
	**	Get a shared EPB
	*/

	if( ( epb = ci_get_epb( c ) ) == ( EPB * ) NULL )
		return;

	/*
	**	Save the EPB's ID
	*/

	mem_id = epb->mem_id;

	/*
	**	Get the header portion of the user's EPB
	*/

	if( copyin( u.u_base, epb, sizeof( EPB ) ) != 0 )
	{
		ci_rel_epb( c, epb );

		u.u_error = EFAULT;

		return;
	}

	/*
	**	Restore the mem ID field and snap any link
	**	put by the user.
	*/

	epb->mem_id	= mem_id;
	epb->next	= ( EPB * ) NULL;

	/*
	**	Check for invalid EPB size.
	*/

	if( ( epb->epb_size > MAX_CEEPB_SIZE ) ||
		( epb->epb_size < sizeof( EPB ) ) )
	{
		ci_rel_epb( c, epb );

		u.u_error = EINVAL;

		return;
	}

	/*
	**	Copy user's EPB data from the user' space.
	*/

	if( epb->epb_size > sizeof( EPB ) )
	{
		if( copyin( ( u.u_base + sizeof( EPB ) ),
			( ( UBYTE * ) epb + sizeof( EPB ) ),
			( epb->epb_size - sizeof( EPB ) ) ) != 0)
		{
			ci_rel_epb( c, epb );

			u.u_error = EFAULT;

			return;
		}
	}

	/*
	**	If there is an attached buffer, copy the BUF's data
	**	from the user's space.
	*/

	if( epb->buf != ( BUF * ) NULL )
	{
		/*
		**	Get a shared BUF.
		*/
		if( ( buf = ci_get_buf( c ) ) == ( BUF * ) NULL )
		{
			ci_rel_epb( c, epb );

			return;
		}

		/*
		**	Copy the BUF's header information.
		*/

		if( copyin( epb->buf, buf, BUF_HDR_SIZE ) != 0 )
		{
			ci_rel_epb( c, epb );
			ci_rel_buf( c, buf );

			u.u_error = EFAULT;

			return;
		}

		/*
		**	Check for valid length size.
		*/

		if( ( buf->length == 0 ) ||
			( ( buf->length + BUF_HDR_SIZE ) > MAX_BUF_SIZE ) )
		{
			ci_rel_epb( c, epb );
			ci_rel_buf( c, buf );

			u.u_error = EINVAL;

			return;
		}

		/*
		**	Get the user data portion of BUF.
		*/
		
		if( copyin( epb->buf->data, buf->data, buf->length ) != 0 )
		{
			ci_rel_epb( c, epb );
			ci_rel_buf( c, buf );

			u.u_error = EFAULT;

			return;
		}

		/*
		**	Correct pointer for kernel space.
		*/

		epb->buf = buf;
	}

	/*
	**	Send them on their way.
	*/

	err = c37kwrite( c, epb, pid );
	
	/*
	**	Tell the user how many bytes copied.
	*/

	switch(err) {
	    case 0:                          /* write ok, no error */
		u.u_error = 0;
		u.u_count = 0;
		break;
	    case NO_BPE:                     /* no write, no error */
		u.u_error = 0;
		break;
	    default:                        /* no write, error */
		u.u_error = err;
		break;
	}

	return;
}

BYTE
c37kwrite( c, epb, pid )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	EPB	*epb;		/* EPB pointer */
	REG	USHORT	pid;		/* Process ID */
{
	REG	BUF	*buf;		/* BUF pointer */
	REG	BYTE	err;		/* Error code */
		USHORT	ip;		/* Interface point */

	/*
	**	Handle registration/deregistration.
	*/

	buf = epb->buf;

	if( ( GET_IP( epb->dst_id ) == CE_IP ) &&
		( ( epb->type == CE_REGISTRATION_REQUEST ) ||
		  ( epb->type == CE_DEREGISTRATION_REQUEST ) ) )
	{
		err = ci_reg( c, epb, pid );

		ci_rel_epb( c, epb );

		if( buf != ( BUF * ) NULL )
			ci_rel_buf( c, buf );

		return( err );
	}

	/*
	**	Is the user registered?  We let unregistered users thru the
	**	check in c37xwrite in the event the EPB is for registration.
	**	Since we're here, we insist upon the user being registered.
	*/

	if( ci_is_reg( c, pid, &ip ) == FALSE )
	{
		ci_rel_epb( c, epb );

		if( buf != ( BUF * ) NULL )
			ci_rel_buf( c, buf );
		
		return( EIDRM );
	}

	/*
	**	Update the source CPU and destination CPU IDs.
	**	We don't allow lying about the src_id.
	*/

	epb->src_id = PUT_ID( c->lc.local_cpu, ip );
	
	if( GET_CPU( epb->dst_id ) == LOCAL_CPU )
		epb->dst_id = PUT_ID( c->lc.local_cpu, GET_IP( epb->dst_id ) );
	
	/*
	**	If it is a local destination, it must be registered.
	**	and the EPB will be placed on the user's queue. Otherwise,
	**	it must be a remote and will be sent to the pipe.
	*/

	if( GET_CPU( epb->dst_id ) == c->lc.local_cpu )
	{
		if( ci_dst_ok( c, GET_IP( epb->dst_id ) ) == FALSE )
		{
			ci_rel_epb( c, epb );

			if( buf != ( BUF * ) NULL )
				ci_rel_buf( c, buf );

			return( ESRCH );
		}

		ci_uwrite( c, GET_IP( epb->dst_id ), epb );
	}
	else
	{
		if( ( err = ci_pwrite( c, epb ) ) != NO_ERROR )
		{
			ci_rel_epb( c, epb );

			if( buf != ( BUF * ) NULL )
				ci_rel_buf( c, buf );

			if( err == NO_PIPE )
				err = EPIPE;

			return( err );
		}
	}

	return( 0 );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xioctl()	
**
**  DESCRIPTION:	I/O Control function for the 37x IPC.
**
**  CONTENTS:		This function will write an EPB to the user's
**			queue for the given 37x IPC.
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number,
**			command		- IOCTL command code,
**			arg		- IOCTL argument pointer,
**			mode		- IOCTL mode code.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  08/28/86	Dana Holgate	1.1	Added ci_debug flag and root
**					user id requirements
**  05/01/87	E. Skinner	1.2	Port to System V/68 Release 3.
**  05/16/87	E. Skinner	1.3	Added "RESET" command passed to
**					MVME372 to REALLY reset it!
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  06/09/87	J. Richey	1.5	Added "ARM_SIGNAL" and
**					"DIS_ARM_SIGNAL".
**  07/15/87	R. Robinson	1.7	Fixed bug with psignal call.
**
**********************************************************************/

VOID
c37xioctl( dev, command, arg, mode )
	REG	DEFAULT dev;		/* Minor device number */
	REG	DEFAULT command;	/* IOCTL command code */
	REG	DEFAULT	*arg;		/* IOCTL argument pointer */
	REG	DEFAULT mode;		/* IOCTL mode code */
{
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	CPU_TAB *cp;		/* CPU description table ptr */
		USHORT ip;		/* Interface point */
	REG	DEFAULT	pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */
		CPU_TAB *c37kcnfg();	/* Fills out CPU description table */
	        short *reset_374_ptr;   /* hardware reset for 374 */

	dev = minor( dev );		/* Get minor device number */

	if( ci_debug )
		printf("%sioct: ioctl on minor device %d, cmd = %d\n",
			ci_dev, dev, command );

	/*
	**	Generate CPU table address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	**	First, we must clear the error.
	*/

	u.u_error = 0;

	/*
	**	If IOCTL command is a reset, reinitialize all.
	*/

	if( command == RESET )
	{
		if (bppv2_local_cpu)
			if ((UBYTE) c->boardadr == *bppv2_local_cpu) 
			{
				u.u_error = ENXIO;
				return;
			}
		/*
		**	Must be root to do this.
		*/
		if( ( u.u_uid == 0 ) || ( u.u_ruid == 0 ) )
		{
			if( c->no_ipc == FALSE )
			{
				/*
				**	Set flag for no communication.
				*/

				c->no_com = TRUE;

				/*
				**	Reset the MVME372 and wait five
				**	seconds.
				*/
				reset_374_ptr = (short *) c->boardadr;
				*reset_374_ptr = 0x800;
				BPPCR( ( dev - CPU_OFFSET ), BPP_HWRESET );

				delay( 7 * HZ );
		
				/*
				**	Reinitialize the memory pools, the
				**	remote channel table, and the
				**	registration table.
				*/

				ci_mem_init( dev - CPU_OFFSET );

				ci_rc_init( dev - CPU_OFFSET );

				ci_reg_init( dev - CPU_OFFSET );

				/*
				**	Try to reinitialize the pipe.
				*/

				if( ci_pipe_init( dev - CPU_OFFSET ) == FAILURE)
					u.u_error = ENXIO;
				else
					c->no_com = FALSE;
			}
			else
				u.u_error = ENXIO;
		}
		else
			u.u_error = EPERM;

		return;
	}

	/*
	**	Check for ioctl CONFIG commands.
	*/

	if(  command == CONFIG  )
	{
		cp = c37kcnfg();
		if( copyout( (char *) cp, arg, sizeof(CPU_TAB)*NUM_CPU_ENTRIES ) != 0 )
			u.u_error = EFAULT;

		return;
	}

	/*
	**	Check for IOCTL debug commands.
	*/

	if( ( command == DEBUG_ON ) || ( command == DEBUG_OFF ) )
	{
		/*
		**	Must be root to do this.
		*/

		if( ( u.u_uid == 0 ) || ( u.u_ruid == 0 ) )
			ci_debug = ( command == DEBUG_ON ) ? TRUE : FALSE;
		else
			u.u_error = EPERM;

		return;
	}

	/*
	**	All other commands must have the process ID be 
	**	registered.
	*/

	if( ( c->no_ipc == FALSE ) && ( c->no_com == FALSE ) )
	{
		if( ci_is_reg( c, u.u_procp->p_pid, &ip ) == TRUE )
		{
			r = &( c->rt[ip] );

			switch (command)
			{
			case WAIT:
				r->block |= BLOCK;
				break;

			case NOWAIT:
				r->block &= ~BLOCK;
				break;

			case ADDR:
				if( copyout( (char *) &c->memadr, arg, sizeof( ULONG ) ) != 0 )
					u.u_error = EFAULT;
				break;

			case ARM_SIGNAL:
				pri = splx( SPLEV( c->rc->intr_level ) );

				r->block |= ST_SIGNAL;
				r->block &= ~BLOCK;

				if( ( r->block & ST_DATA ) != 0 )
				{
					r->block &= ~ST_SIGNAL;
					r->block &= ~ST_DATA;

					psignal( r->proc, SIGUSR1 );
				}

				splx( pri );
				break;

			case DIS_ARM_SIGNAL:
				pri = splx( SPLEV( c->rc->intr_level ) );

				r->block |= BLOCK;
				r->block &= ~ST_SIGNAL;
				r->block &= ~ST_DATA;

				splx( pri );
				break;

			default:
				u.u_error = EINVAL;
			}
		}
		else
			u.u_error = EIDRM;
	}
	else
		u.u_error = ENXIO;
}

CPU_TAB *
c37kcnfg()
{
	REG	int	i;	/* Index */
	REG	CPU_TAB	*cp;	/* CPU description table entry ptr */
	REG	CPU_ENT	*c;	/* CPU table entry pointer */
	REG	CPU_TAB	*rp;	/* CPU description table ptr */

	/*
	**	Initialize the pointers.
	*/

	rp = cp = ci_cputab;
	c  = ci_cpu;

	/*
	**	Fill in the description table based upon ci_cpu table values.
	*/

	for( i = 0; i < NUM_CPU_ENTRIES; ++i, ++cp, ++c )
	{
		cp->cpu_id = i + CPU_OFFSET;

		/* setup for board type */
		switch(c->regbase) {
		    case MVME372_OFFSET:
			cp->cpu_type = 372;
			break;
		    case MVME374_OFFSET:
			cp->cpu_type = 374;
			break;
		    case MVME147_OFFSET:
			cp->cpu_type = 147;
			break;
		    default:
			cp->cpu_type = 0;
			break;
		}

		cp->cpu_present = c->no_ipc ? FALSE : TRUE;
		cp->cpu_running = c->no_com ? FALSE : TRUE;
	}

	/*
	**	Return the pointer to the CPU description table.
	*/

	return( rp );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	c37xintr()	
**
**  DESCRIPTION:	Interrupt service routine for c37x.
**
**  CONTENTS:		This function is the interrupt service routine
**			for all the c37x IPCs. It will read from the 
**			pipe and, if not a Common Environment operation,
**			place the EPB on the user's queue.
**
**  PARAMETERS:		
**
**	Input:		dev		- minor device number.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  01/11/86	Dana Holgate	1.1	Initial Release
**  06/01/87	R. Robinson	1.4	Changed names for generic 37x,
**					added comments.
**  06/09/87	J. Richey	1.5	Added 'psignal'.
**  07/15/87	R. Robinson	1.7	Fixed bug with psignal call.
**
**********************************************************************/

DEFAULT
c37xintr(dev)
	REG	DEFAULT	dev;		/* Minor device number */
{
	REG	BUF	*buf;		/* BUF pointer */
	REG	BUF	*buf_next;	/* BUF pointer */
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	CE_EPB  *epb;		/* EPB pointer */
	REG	USHORT	ip;		/* Interface point */
	REG	ULONG	i;		/* Just a counter */
		REG_ENT	*r;		/* Registration table entry address */
		CE_EPB	*depb;		/* Destination EPB pointer */

	dev = minor(dev);		/* Get minor device number */

	if (ci_debug)
		printf("%sintr: interrupt on minor device %d\n", ci_dev, dev);

	/*
	**	Generate CPU table address.
	*/

	c = &ci_cpu[dev - CPU_OFFSET];

	/*
	**	Do pipe read until no more EPBs.
	*/

	while ((epb = (CE_EPB *)ci_pread(c)) != (CE_EPB *)NULL)
	{
		/*
		**	Get/check interface point of EPB.
		*/

		if ((ip = GET_IP(epb->hdr.dst_id)) == CE_IP)
		{
			/*
			**	Do Common Environment stuff.
			*/

			switch (epb->hdr.type)
			{
			case CE_GETBUF_REQUEST:
				buf_next = (BUF *)NULL;

				for (i = 0; i < epb->epbbuf_count; ++i)
				{
					buf = ci_get_buf( c );
					if( buf  != ( BUF * ) NULL )
					{
						buf->next = buf_next;
						buf_next  = buf;
					}
					else
						break;
				}

				ci_swap_id(epb);

				epb->hdr.type     = CE_GETBUF_RESPONSE;
				epb->hdr.status   = NO_ERROR;
				epb->hdr.epb_size = sizeof(CE_EPB);
				epb->hdr.buf      = buf_next;
				epb->epbbuf_count = i;

				if (ci_pwrite(c, epb) != NO_ERROR)
					ci_add_retry(c, epb);
				break;

			case CE_RETBUF_REQUEST:
				buf = epb->hdr.buf;

				while (buf != (BUF *)NULL)
				{
					buf_next = buf->next;

					ci_rel_buf(c, buf);

					buf = buf_next;
				}

				ci_rel_epb(c, epb);
				break;
			}

			continue;
		}

		/*
		**	First, check to see if this EPB must be
		**	routed to another board.
		*/

		if ((i = GET_CPU(epb->hdr.dst_id)) != CPU_NO)
		{
			/*
			**	Generate CPU table address.
			*/

			c = &ci_cpu[i - CPU_OFFSET];

			if ((c->no_ipc == FALSE) && (c->no_com == FALSE))
			{
				if ((depb = (CE_EPB *)ci_get_epb(c)) != (CE_EPB *)NULL)
				{
					if (((buf = epb->hdr.buf) == (BUF *)NULL) || ((buf = ci_get_buf(c)) != (BUF *)NULL))
					{
						i = depb->hdr.mem_id;

						depb->hdr.mem_id = epb->hdr.mem_id;
						epb->hdr.mem_id  = i;

						if (ci_pwrite(c, epb) != NO_ERROR)
						{
							epb->hdr.mem_id = depb->hdr.mem_id;

							depb->hdr.mem_id = i;

							if (buf != (BUF *)NULL)
								ci_rel_buf(c, buf);

							ci_rel_epb(c, depb);
						}
						else
							(epb = depb)->hdr.buf = buf;
					}
					else
						ci_rel_epb(c, depb);
				}
			}

			c = &ci_cpu[dev - CPU_OFFSET];

			if (epb->hdr.buf != (BUF *)NULL)
				ci_rel_buf(c, epb->hdr.buf);

			ci_rel_epb(c, epb);
		}
		else
		{
			/*
			**	If process is registered, place on user's
			**	queue, else return to sender.
			*/

			if (ci_dst_ok(c, ip) == TRUE)
			{
				ci_uwrite(c, ip, epb);

				r = &( c->rt[ip] );

				if( ( r->block & ST_SIGNAL ) != 0 )
				{
					r->block &= ~ST_SIGNAL;
					r->block &= ~ST_DATA;

					psignal( r->proc, SIGUSR1 );
				}
				else
					r->block |= ST_DATA;
			}
			else
				ci_ret_send(c, epb, DST_NOT_REGISTERED);
		}
	}

	return (SUCCESS);
}
