/*	@(#)ci_user.c	35.1	*/
/*MH*********************** MODULE HEADER *****************************
**
**  MODULE NAME:	ci_user.c
**
**  DESCRIPTION:	This module contains the Common Interface
**			user read and write functions.
**
**  FUNCTIONS:		ci_uwrite, ci_uread, ci_unread
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------   	---	-----------------------------
**  01/01/86	Dana Holgate	1.1	Initial Coding.
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x.
**  06/01/87	J. Richey	1.1	Modified ci_uwrite and ci_uread.
**  07/21/87	J. Richey	1.2	Modified ci_uread.
**  09/10/87	dtk             2.1     Changed location of header files
**
***********************************************************************/

#include <sys/types.h>
#include <sys/param.h>
#include <sys/immu.h>
#include <sys/region.h>
#include <sys/proc.h>

#include <sys/c37x/ci_driver.h>
#include <sys/c37x/ci_const.h>
#include <sys/c37x/ci_struct.h>
#include <sys/c37x/ci_global.h>

extern	VOID	wakeup();		/* Wake up process */
extern	VOID	sleep();		/* Sleep until woken up */
extern	DEFAULT	splx();			/* Set priority level */

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_uwrite()	
**
**  DESCRIPTION:	Write an EPB to the tail of the indicated
**			user's read queue.
**
**  CONTENTS:		This function will place the passed EPB
**			on the tail of the passed user's read
**			queue.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			ip		- Interface point,
**			epb		- EPB pointer.
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**  06/01/87	J. Richey	1.1	Added BLOCK flag check.
**
**********************************************************************/

VOID
ci_uwrite(c, ip, epb)
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface Point */
	REG	EPB	*epb;		/* EPB pointer */
{
	REG	DEFAULT	pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	If the list is non-empty, place the EPB on the
	**	end of the list; else, set the tail and head
	**	pointers to the EPB.
	*/

	r = &( c->rt[ip] );

	if( r->rq_tail != ( EPB * ) NULL )
	{
		r->rq_tail->next = epb;
		r->rq_tail	= epb;
	}
	else
		r->rq_tail	= r->rq_head	= epb;

	epb->next = ( EPB * ) NULL;

	/*
	**	Enable interrupts.
	*/

	splx( pri );

	/*
	**	Call the specified rx_int routine.
	*/

	if( r->rx_int != NULL )
		( *r->rx_int )( c, ip );
}

VOID
c37_rxint( c, ip )
	REG	CPU_ENT	*c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface Point */
{
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	If the user process is sleeping, wake it up.
	*/

	r = &( c->rt[ip] );

	if( ( ( r->block & BLOCK) != 0) && ( r->proc->p_stat == SSLEEP ) )
		wakeup( r );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_uread()	
**
**  DESCRIPTION:	Read an EPB from the head of the indicated
**			user's read queue.
**
**  CONTENTS:		This function will read an EPB from the head
**			of the passed user's read queue and return it.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			ip		- Interface point,
**
**	Output:		NONE
**
**	Returns:	Pointer to EPB.
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**  06/01/87	J. Richey	1.1	Added signal mode check.
**  07/21/87	J. Richey	1.2	Fixed bug in ST_DATA.
**
**********************************************************************/

EPB *
ci_uread( c, ip )
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface point */
{
	REG	EPB	*epb;		/* The returned EPB */
	REG	DEFAULT pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	If the list is empty, forget it.
	*/

	r = &( c->rt[ip] );

	if( ( epb = r->rq_head ) == ( EPB * ) NULL )
		return( ( EPB * ) NULL );

	/*
	**	Get first EPB from the list and adjust the head
	**	pointer.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	if( epb->next == ( EPB * ) NULL )
	{
		r->rq_head	=  r->rq_tail	= ( EPB * ) NULL;
		r->block	&= ~ST_DATA;
	}
	else
	{
		r->rq_head	=  epb->next;

		epb->next	= ( EPB * ) NULL;
	}

	splx( pri );

	return( epb );
}

/*FH*********************** FUNCTION HEADER ***************************
**
**  FUNCTION NAME:	ci_unread()	
**
**  DESCRIPTION:	Unread (replace) an EPB to the head of the
**			indicated user's read queue.
**
**  CONTENTS:		This function will place the passed EPB back
**			on the passed user's read queue.
**
**  PARAMETERS:		
**
**	Input:		c		- CPU table entry,
**			ip		- Interface point,
**			epb		- EPB address
**
**	Output:		NONE
**
**	Returns:	NONE
**
**  COMPLEX LOCALS:
**
**  METHOD:
**
**  NOTES:
**
**  UPDATE LOG:
**  Date	Name		V/R	Comments
**  --------	-----------	---	-----------------------------
**  04/01/86	Dana Holgate	1.1	Initial Release
**  04/20/87	R. Robinson	1.1	Changed names for generic 37x,
**					added comments.
**
**********************************************************************/

VOID
ci_unread(c, ip, epb)
	REG	CPU_ENT *c;		/* CPU table entry address */
	REG	USHORT	ip;		/* Interface point */
	REG	EPB	*epb;		/* EPB address */
{
	REG	DEFAULT pri;		/* Current priority level */
	REG	REG_ENT	*r;		/* Registration table entry address */

	/*
	**	Inhibit interrupts.
	*/

	pri = splx( SPLEV( c->rc->intr_level ) );

	/*
	**	Place EPB back on user's read queue.
	*/

	r = &( c->rt[ip] );

	if( r->rq_head == ( EPB * ) NULL )
	{
		epb->next = (EPB *)NULL;

		r->rq_tail = r->rq_head = epb;
	}
	else
	{
		epb->next = r->rq_head;

		r->rq_head = epb;
	}

	/*
	**	Enable interrupts.
	*/

	splx( pri );
}
