

/*****************************************************************************

	Rev = %I%		Date of 'get'   = %T% %H%
				Date of 'delta' = %U% %G%

	File = %M%

*****************************************************************************/



/**********************************************************************
 *								      *
 *		Copyright (c) Ferranti plc 1988			      *
 *								      *
 *	This is unpublished proprietory source code of Ferranti plc.  *
 *		The copyright of which vests in Ferranti plc.	      *
 **********************************************************************/


/**********************************************************************
 *								      *
 *			FVME250 DEVICE DRIVER			      *
 *    							              *	
 * 		    Written  1988  by P.E.Callow   	              *	
 *    							              *	
 **********************************************************************
 *    							              *	
 *	This program is the driver for the FVME250 card.	      *
 *    							              *	
 **********************************************************************/

/*
 *			NOTES ABOUT THE PROGRAM
 *			=======================
 *
 *	For further information see the relevent program documentation.
 *
 *	Terminology:
 *
 *	The FVME250 is sometimes referred to in the program as a FIOP.
 *
 *	There can be upto 8 controllers on the SCSI bus, controller 7 is the
 *	FVME250.
 *	Each controller can have two devices attatched. And each device can
 *	have upto eight partitions. The minor device number gives this 
 *	information for a particular device.
 *
 *			   bits     7 6 5 4 3 2 1 0
 *				   -----------------
 *	minor device number (mdev) | | | | | | | | |
 *				   -----------------
 *				           
 *	bits 0-2  =  slice field       ( 0-7, use depends upon device type). 
 *      bit   3   =  device number     ( 0 or 1 ).
 *	bits 4-6  =  controller number ( 0-7 ).
 *      bit   7   =  me250 number      ( This allows 2 me250's per major
 *				         device number ).
 *
 *	Main data structures:
 *
 *	bdctab[]       called device table. Its main use is as a head to the
 *		       queue of commands waiting to be serviced.
 *
 *	bdtab[][]      one for each physical device. Used in error logging.
 *
 *	m250stat[][]   contains error statistics for each physical device. 
 *
 *	sd_conf[][]    contains configuration information (eg no. of bytes
 *		       per sector, type of device, device name...) for
 *                     each physical device.
 *
 *	sd_dev[]      contains configuration information ( eg interrupt
 *		      vector, address of id block) for each FVME250.
 *
 *	fiop_tbl[]    contains details about an FVME250.
 */

/*
 *	For the FVME250 the following restrictions apply:
 *
 *	(1) Byte accesses on FVME250 are prohibited.
 *	(2) FVME250 DMA can not transfer data across the bus.
 *	(3) Addresses given to the FVME250 must be local to FVME250 memory.
 */


#undef DEBUG 	     
#undef DINTR          
#undef DSTRATEGY

#include <sys/param.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/dir.h>
#include <sys/reg.h>
#include <sys/signal.h>
#include <sys/user.h>
#include <sys/buf.h>
#include <sys/iobuf.h>
#include <sys/elog.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/map.h>
#include <sys/systm.h>
#include <sys/inode.h>
#include <sys/cmn_err.h>
#include <sys/utsname.h>
#include <sys/erec.h>
#include <sys/mvmecpu.h>
#include <sys/mvme350.h>
#include <stdio.h>
#include <sys/dk.h>
#include <sys/fvme250/f250bits.h>
#include <sys/fvme250/f250buf.h>
#include <sys/fvme250/f250cmds.h>
#include <sys/fvme250/f250dev.h>
#include <sys/fvme250/f250fix.h>
#include <sys/fvme250/f250addr.h>
#include <sys/fvme250/f250error.h>
#include <sys/fvme250/f250legal.h>
#include <sys/fvme250/f250value.h>
#include <sys/fvme250/f250slice.h>


#define MAXFIOP         4        /*   max no of ME250's                */
#define MAXCONTROLLERS  8        /*   max no of controllers per ME250  */
#define MAXDEVS         2        /*   max no of devices per controller */
#define NCMDS           1000      /*   max no of commands available in  */
				 /*   command pool of an ME250         */

/*  QUEUE1 is the queue of commands waiting to be serviced     */
/*  QUEUE2 is the queue of commands being serviced by ME250/2  */

#define NCMDS_QUEUE1 io_s1
#define NCMDS_QUEUE2 io_s2


/*  The global data structures used by this device driver   */
struct iobuf bdctab[MAXFIOP];                /* one for each FIOP            */
					     /* device table.                */

struct iobuf bdtab[MAXFIOP][16];             /* one for each physical device  */
 
struct iotime m250stat[MAXFIOP][16];         /* one for each physical device  */
					     /* contains error statistics for */
					     /* each device.                  */
 

extern struct confg sd_conf[MAXFIOP][16];    /* configuration details for
						each device                  */

extern struct dev_init sd_dev[MAXFIOP];     /* configuration details for
					       each FIOP                     */

struct me250addr fiop_tbl[MAXFIOP];          /* details for each FIOP        */

struct device_info dev_info[MAXFIOP][16];    /* device information           */

extern short legal[ME250CMDS+1];             /*  legality of ioctl commands  */

extern cpuboard;     			     /*  The CPU type in the system  */

extern m250_ivec[];     	             /*  Vector array                */


#define READY    1       /* state of FIOP */
#define NOTREADY 0       /* state of FIOP */


/*
 *	General macros used in the program
 *
 *	ctl = parameter passed by pseudo driver routines
 *	x   = minor device number
 *
 *	getfnum(ctl,x)     -   returns FIOP number
 *	devnum(x)          -   returns device number 
 *			       (combination of controller number and
 *			       logical device number). 
 *	devslice(x)        -   returns slice number
 *	devlog(x)          -   returns logical device number ( 0 or 1 ).
 * 	devcont(x)         -   returns controller number
 *	devtype(ctl,x)     -   returns device type eg WINCH
 *	devname(ctl,x)     -   returns device name eg WINCH160M
 *	devstat(ctl,x)     -   returns device status
 *	secsize(ctl,x)     -   bytes per sector for this device
 *	rewind(ctl,x)      -   should device be rewound on closure
 *	noaction(ctl,x)    -   should any action be taken on closure
 *	writing(ctl,x)     -   has any writing been done on the tape
 *	filemark(ctl,x)    -   has a filemark been detected on reading
 */

/*   macros  for converting to/from bytes/blocks   */
#define dtob(x)    ((x)<<BSHIFT)
#define btod(x)    (((x) + (BSIZE-1)) >> BSHIFT)

	
#define getfnum(ctl,x)  ((x>>7)&0x1)+2*ctl  
#define devnum(x)       ((x>>3)&0xf)         
#define devslice(x)     ((x)&0x7)         
#define devlog(x)       (x>>3)&0x1         
#define devcont(x)      ((x>>4)&0x7)
#define devtype(ctl,x)  sd_conf[getfnum(ctl, x)][devnum(x)].dev_type
#define devname(ctl,x)  sd_conf[getfnum(ctl, x)][devnum(x)].dev_name
#define secsize(ctl,x)  sd_conf[getfnum(ctl, x)][devnum(x)].bytes_per_sector
#define devstat(ctl,x)  dev_info[getfnum(ctl,x)][devnum(x)].status
#define rewind(ctl,x)   (dev_info[getfnum(ctl,x)][devnum(x)].option & BIT_ZERO)
#define noaction(ctl,x) (dev_info[getfnum(ctl,x)][devnum(x)].option & BIT_ONE)
#define writing(ctl,x)  (dev_info[getfnum(ctl,x)][devnum(x)].option & BIT_FOUR)
#define reading(ctl,x)  (dev_info[getfnum(ctl,x)][devnum(x)].option & BIT_SIX)
#define filemark(ctl,x) (dev_info[getfnum(ctl,x)][devnum(x)].option & BIT_SEVEN)

unsigned short fiopno;         /* number of ME250's found on the VME bus  */


char  *lamb = LAMBSTR; 
char  *fiop = FIOPSTR;


int timedout();
void setopflag();
void resetopflag();
void init_cmdbufs();
void create_command();
void init_databufs();
void relcmdbuf();
void me250check();
void me250breakup();
void me250start();
void ps_0strategy();

struct me250cmd *getcmdbuf();
unsigned int me250malloc();
ushort  set_error();

/*
 *	find_me250s() probes through an address range searching for ME250/2s.
 *	It returns the number of ME250/2s it finds.
 */

find_me250s()
{
	short id = 0;                 /* Id number of me250's */
	register unsigned long base;  /* Base address of me250 address space */
	register unsigned long end;   /* End address of me250 address space */
	long me250_lamb;
	long me250_fiop;
	

#ifdef DEBUG
	printf("Inside find_me250s: looking for FIOPS\n");
#endif

	/*
         *  Because the ME250/2 will not allow byte accesses, the strings
         *  for LAMB and FIOP are read into long variables and passed
         *  onto strncmp.
	 */

	switch (cpuboard) {
	case FVME252  :
		base = 	PROBEST_252;
		end  = 	PROBEEND_252;
#ifdef DEBUG
		printf("FVME252 processor\n");
#endif
		break;
	case VME141  :
		base = 	PROBEST_141;
		end  = 	PROBEEND_141;
#ifdef DEBUG
		printf("MVME141 processor\n");
#endif
		break;
	default :
		base = 	PROBEST_132;
		end  = 	PROBEEND_132;
	}

	for ( base ; base < end; base = base + PROBEINC) {
		if ( me250bprobe((base+BRDID), -1) ) {
			continue;
		}

		/* test to see if we have an ME250 */
		me250_lamb = *((long *)(base + BRDID));
		me250_fiop = *((long *)(base + BRDTYPE));

		if (strncmp(lamb, (char *)&me250_lamb, IDLEN))
			continue;
		if (strncmp(fiop, (char *)&me250_fiop, TYLEN))
			continue;
#ifdef DEBUG
	printf("Found FVME250 at %x\n",base);
#endif


		if ( id == MAXFIOP ) {
			printf("Found FVME250 at address %x\n", base);
			printf("But exceeded max no. of fiop's\n");
			continue;
		}

		/* set up fiop_tbl structure for the fiop we have just found */
		fiop_tbl[id].me250_mem   = (long *)base; 
		fiop_tbl[id].me250_idblk = (struct brd_id *)(base + BRDID); 
		fiop_tbl[id].me250_loc   = (short *)(base + LOCADDR); 
		fiop_tbl[id].me250_queue = (long *)(base + REQUEST); 
		fiop_tbl[id].cmd_buf = (struct me250cmd *)(base + STARTPOOL); 
		fiop_tbl[id].data_buf = (char *)(base + DATSTART); 

		/* increment fiop id */
		++id;
	}           
#ifdef DEBUG
	printf("Found %x FIOPS\n",id);
#endif
	return(id);        /* no. of fiop's found */

}




m250init()
{

	unsigned short	id, c, d;
	long	tout = SELFTIMEOUT;
	long	status;
	struct  me250cmd *getcmdbuf();
	struct  me250cmd *cmd;
	char    *bbuf;
	char dev_num;
	long i;
	void initstart();
	short lev;

#ifdef DEBUG
	printf("Inside m250init\n");
#endif

	if (cpuboard == VME131) {
		/* allow d16 across VME */
		*(unsigned char *)0xfffb003b = 0x78;
#ifdef DEBUG
		printf("Found VME131 CPU\n");
#endif
	}

	/*  find no. of FIOP's and set up fiop_tbl[] array  */
	fiopno = find_me250s();

	/*  set up command pool for each ME250 */
	init_cmdbufs();

	/*  set up memory maps for each ME250  */
	init_databufs();

	for ( id = 0; id < fiopno; id++, tout = SELFTIMEOUT) {
		while ((status = *(fiop_tbl[id].me250_mem + (long *)STAT)) == TESTRUN && tout)
			--tout;
		if (!tout) {
			printf("Timeout! FVME250 %d not responding\n", id);
			continue;
		}
		switch (status) {
		case TESTPASS:
			fiop_tbl[id].state = READY;
#ifdef DEBUG
	printf("FVME250 %d passed selftests\n",id);
#endif
			break;
		case TESTFAIL:
			printf("FVME250 %d failed selftests\n",id);
			fiop_tbl[id].state = NOTREADY;
			break;
		case NODEVICE:
#ifdef DEBUG
			printf("FVME250 %d failed to initialise all devices\n",id);
#endif
			fiop_tbl[id].state = READY;
			break;
		default:
			printf("FVME250 %d failed selftests\n",id);
			fiop_tbl[id].state = NOTREADY;
			break;
		}

		if ( fiop_tbl[id].state == NOTREADY)
			continue;

#ifdef DEBUG
	printf("Getting a command buffer\n");
#endif

		/*   get a command buffer  */
		cmd = getcmdbuf(id);

#ifdef DEBUG
	printf("Allocating memory\n");
#endif

		/*   allocate memory in datapool  */
		bbuf = (char *)me250malloc(fiop_tbl[id].datamap,
				       	  sizeof(struct dev_init) );
		
		if (!bbuf) {
			printf("can not get memory for fiop %d\n", id);
			fiop_tbl[id].state = NOTREADY;
			relcmdbuf(id,cmd);
			continue;
		}

		/* change default vector number to that
                   configured in sysgen(1M)		*/
		sd_dev[id].vector = (char)(m250_ivec[id]/4);

		/*   copy data into pool  */
		my_blkcopy( (char *)&sd_dev[id],bbuf , sizeof(struct dev_init) );

		create_command(INIT_CMD, 0, (short)INTERR|DRIV_CMD,-1, MEM0,
			      bbuf-(char *)fiop_tbl[id].me250_mem,cmd);

		add_command(id, cmd);
#ifdef DEBUG
	printf("Calling start\n");
#endif

		/* set the interrupt level to allow the FIOP to interrupt */
		/* the driver */
		lev = spl3();
		initstart(id);

#ifdef DEBUG
	printf("Waiting\n");
#endif
		for (i = EXEC_TIMEOUT; i && bdctab[id].b_active; i--)
			;
		/*  restore the interrupt level */
		splx(lev);

#ifdef DEBUG
	printf("Finished waiting loop\n");
#endif
		if (!i) {
			printf("timed out waiting loop\n");
			fiop_tbl[id].state = NOTREADY;
			relcmdbuf(id, cmd);
			me250mfree( fiop_tbl[id].datamap, sizeof(struct dev_init), bbuf);
			continue;
		}
		
		printf("FVME250 %x INITIALIZED\n",id);

		/*   release command and previously allocated data space */
		relcmdbuf(id, cmd);
		me250mfree( fiop_tbl[id].datamap, sizeof(struct dev_init), bbuf);

		/* if cmd->error != 0 then at the moment do nothing  */
		/* In later versions of the driver, devices not      */
		/* available will be marked as such                  */

		/*  mark each device as available                    */
		for ( c = 0 ; c < sd_dev[id].scsi_cntl; c++) {
			for (d= 0 ; d < sd_dev[id].controller[c].dev_no; d++) {
				dev_num = c<<1;
				dev_num |= d;
				dev_info[id][dev_num].status = AVAILABLE;
			}
		}
	}
#ifdef DEBUG
	printf("Leaving m250init\n");
#endif
}







/* Fill command buffer in */
void
create_command(cnum, id, instreg, ad, cnt, datadr, cmd)
short	        cnum, id;	/* Command number and id */
short           instreg; 	/* Instruction register contents */
unsigned long   ad, cnt;
char            *datadr;
struct me250cmd *cmd;
{
	struct me250cmd cmd_buf;
	struct me250cmd *pcmd;

	pcmd = &cmd_buf;

	/*   
	 *    Because of a hardware limitation,
	 *    it is not possible to access byte quantities
	 *    from ME250/2.
	 *    So no field in command structure is only one byte long.
         */	

	pcmd->cmd = cnum;
	pcmd->cntrlid = id;
	pcmd->intsreg = instreg;
	pcmd->addr = ad;
	pcmd->count = cnt;
	pcmd->buff_addr = datadr;
	pcmd->error = 0;

	my_blkcopy(pcmd, cmd, (sizeof(struct me250cmd)-8) );
}


/*
 *  add_command queues a command request onto the appropiate ME250's
 *  request queue.
 */

add_command(fnum, cmd)
	unsigned short  fnum;
	struct me250cmd *cmd;
{
	struct iobuf    *ip;
	struct me250cmd *pcmd;
	int level;

	/*  only allow one procedure at a time to
	    append the queue. */
	level = spl5();

	/*  we are only adding one request to the queue  */
	cmd->next_cmd = NULL; 

	ip = &bdctab[fnum];

#ifdef DEBUG
	printf("Adding b_actf = %x  b_actl= %x\n",ip->b_actf, ip->b_actl);
#endif

	/*  if the queue is empty */
	if (!ip->b_actf) {
		ip->b_actf = (struct buf *)cmd;
	}
	else {
		pcmd = (struct me250cmd *)ip->b_actl;
		/* change from VME to local address for ME250 */
		pcmd->next_cmd = (struct me250cmd *)((long)cmd
			 - (long)fiop_tbl[fnum].me250_mem);
#ifdef DEBUG
	printf("After Add pmd->next_cmd = %x\n",pcmd->next_cmd);
#endif
	}

	ip->b_actl = (struct buf *)cmd;

	/*  another command on inactive queue  */
	ip->NCMDS_QUEUE1++;

	splx(level);

#ifdef DEBUG
	printf("After Add b_actf = %x  b_actl= %x\n",ip->b_actf, ip->b_actl);
#endif

}



/*
 *  remove_command dequeues a command request from the appropiate ME250's
 *  request queue.
 */

remove_command(fnum)
	unsigned short  fnum;
{
	struct iobuf    *ip;
	struct me250cmd *pcmd;

	ip = &bdctab[fnum];
	pcmd       = (struct me250cmd *)ip->b_forw;

#ifdef DEBUG
	printf("BEfore remove b_forw = %x  b_back= %x\n",ip->b_forw, ip->b_back);
	printf("Before remove pcmd->next_cmd = %x\n",pcmd->next_cmd);
#endif

	/*  if this is only command on the queue  */
	if ( !pcmd->next_cmd) {
		ip->b_forw = NULL;
		ip->b_back = NULL;
	}
	else {
		ip->b_forw = (struct buf *)((long)pcmd->next_cmd+
			     (long)fiop_tbl[fnum].me250_mem);
	}

	pcmd->next_cmd = NULL;

	/*  decrement no commands on the active queue */
	ip->NCMDS_QUEUE2--;

#ifdef DEBUG
	printf("After remove b_forw = %x  b_back= %x\n",ip->b_forw, ip->b_back);
	printf("After remove pcmd->next_cmd = %x\n",pcmd->next_cmd);
#endif

}




/*
 *  getcmdbuf returns a command buffer for FIOP fiopnum,
 *  if none are available then it sleeps until one becomes free.
 */


struct me250cmd *getcmdbuf(fnum)
register unsigned short  fnum;
{
	register struct me250cmd *cptr;
	int level;

	level = spl5();

	/*  if no cmd buffers available then sleep  */
	while ((cptr = fiop_tbl[fnum].cptrhead) == NULL) {
		fiop_tbl[fnum].next_cmdwait++;
		sleep(&fiop_tbl[fnum].cptrhead ,PRIBIO);
	}

	/*  we have a free cmd buffer   */
	fiop_tbl[fnum].cptrhead = cptr->next_cmd; 

	cptr->next_cmd = NULL;

	splx(level);

	return(cptr);
}


/*
 *  relcmdbuf releases the command buffer pointed to by cptr for FIOP fiopnum,
 *  anybody sleeping for a command buffer is woken up.
 */

void
relcmdbuf(fiopnum, cptr)
register unsigned short fiopnum;
register struct me250cmd *cptr;
{
	register struct me250cmd *ccptr;
	int level;

	level = spl5();

	ccptr = fiop_tbl[fiopnum].cptrhead;
	fiop_tbl[fiopnum].cptrhead = cptr;
	cptr->next_cmd = ccptr;

	/* if anybody is waiting for a cmd buffer */
	/* wake them up.                           */ 
	if (fiop_tbl[fiopnum].next_cmdwait) {
		fiop_tbl[fiopnum].next_cmdwait = 0;
		wakeup(&fiop_tbl[fiopnum].cptrhead);
	}

	splx(level);
}




/*
 *	init_cmdbufs() is called by ME250 initialisation routine m250init().
 *      It sets up a queue  of NCMDS command buffers in the ME250's command
 *	pool, cptrhead is set to the head of this queue, the last command buffer
 *	points to NULL.
 */

void
init_cmdbufs()
{
	register int i;
	register int id;
	register struct me250cmd *cmd;

	/*
	 * Init command pool for each ME250.
         * Each command pool contains NCMDS command structures.
	 */

	for ( id = 0 ; id < fiopno ; id++) {
		fiop_tbl[id].cptrhead = fiop_tbl[id].cmd_buf;
		cmd = fiop_tbl[id].cmd_buf;
		for(i=0 ; i < NCMDS-1 ; i++) {
			cmd->next_cmd = cmd+1;
			cmd++;
		}

		/* last command in queue points to NULL */
		cmd->next_cmd = NULL;
	}
}



/*
 *	init_databufs() is called by ME250 initialisation routine m250init().
 *      It sets up a memory map structure, this map will be used by the routines
 *	me250malloc() and me250mfree() to allocate and deallocate memory for 
 *	data needed or received from the ME250.
 */

void
init_databufs()
{
	register int id;
	struct map   *mp;

	/* set up memory map for each FIOP */

	for ( id = 0 ; id < fiopno ; id++) {

		/*  500 is the number of array elements in datamap  */
		mapinit(fiop_tbl[id].datamap, 1000);

		mp = mapstart(fiop_tbl[id].datamap);
		mp->m_addr = (unsigned int )fiop_tbl[id].data_buf;
		mp->m_size = DATSIZE;
	}
}





void m250close (ctl,mdev,flag)
int ctl;
register dev_t mdev;
int flag;
{
	register struct me250cmd *cmd;
	struct me250cmd *getcmdbuf();
	unsigned short fnum;
	char id;

#ifdef DEBUG
      	printf("Inside m250close ctl %x  mdev %x flag %x\n", ctl, mdev, flag);
	printf("rewind   = %x\n",rewind(ctl,mdev) );
	printf("writing  = %x\n",writing(ctl,mdev) );
	printf("filemark = %x\n",filemark(ctl,mdev) );
	printf("noaction = %x\n",noaction(ctl,mdev) );
#endif

	/*  get FIOP number */
	fnum = getfnum(ctl, mdev);

	/*  check this fiop is available  */
	if ( fiop_tbl[fnum].state == NOTREADY ) {
		u.u_error = ENXIO;
		return;
	}

	/*  check this device is available  */
	if ( devstat(ctl, mdev) == UNAVAILABLE ) {
		u.u_error = ENXIO;
		return;
	}

	/*        WINCHESTER      */
	if ( devtype(ctl, mdev) == WINCH) {
		resetopflag(fnum, mdev);
		u.u_error = 0;
		return;
	}

	/*         STREAMER        */
	if ( devtype(ctl, mdev) == STREAMER) {
		if ( dev_info[fnum][devnum(mdev)].option & BIT_ONE ) {
			/*  do nothing on close  */
			u.u_error = 0;
			resetopflag(fnum, mdev);
			return;
		}

		if ( (flag & FWRITE) && (writing(ctl,mdev)) ) {
			/* Write 2 file-marks */
			cmd = getcmdbuf(fnum);

			id = devlog(mdev);
			id |= (devcont(mdev) <<4);

		        create_command(WRFMKS_CMD,id,(short)INTERR|DRIV_CMD,
					0,1,0,cmd);

			send_command(cmd, fnum);

			if ( set_error(cmd->error) ) {
				u.u_error = set_error(cmd->error);
				resetopflag(fnum, mdev);
				relcmdbuf(fnum,cmd);
				return;
			}
			
			u.u_error = 0;
			relcmdbuf(fnum, cmd);
		}
	
		if ( !rewind(ctl, mdev) ) {	/* Don't rewind */
			/* if we have been reading */
			if ( reading(ctl,mdev) ) {
				/* space forward a filemark */
				cmd = getcmdbuf(fnum);

				id = devlog(mdev);
				id |= (devcont(mdev) <<4);

				create_command(SPFMKS_CMD, id,
				(short)INTERR|DRIV_CMD, 0, 1,0, cmd);

				send_command(cmd,fnum);

				if (set_error(cmd->error) ) {
					u.u_error = set_error(cmd->error);
					resetopflag(fnum, mdev);
					relcmdbuf(fnum, cmd);
					return;
				}

				u.u_error = 0;
				relcmdbuf(fnum, cmd);
			}
		}
		else {	
			/*  rewind the tape */
			cmd = getcmdbuf(fnum);
	
			id = devlog(mdev);
			id |= (devcont(mdev) <<4);
	
			create_command(REWIND_CMD,id,(short)INTERR|DRIV_CMD,0,0,0,cmd);

			send_command(cmd, fnum);
	
			if ( set_error(cmd->error) ) {
				u.u_error = set_error(cmd->error);
				resetopflag(fnum, mdev);
				relcmdbuf(fnum,cmd);
				return;
			}

			u.u_error = 0;
			relcmdbuf(fnum,cmd);
			wait_device_free(fnum,mdev);
		}

		resetopflag(fnum , mdev);        	/* mark dev closed */
	}
}


void m250open (ctl,mdev,flag)
register ctl;
register dev_t mdev;		/* minor device number */
int flag;
{
	register struct me250cmd *cmd;
	unsigned short fnum;
	char id;
	int level;

#ifdef DEBUG
	printf("Inside m250open ctl %d  mdev %d flag %d\n", ctl, mdev, flag);
#endif

	/*    get FIOP number    */
	fnum = getfnum(ctl,mdev);

	/*  check this fiop is available  */
	if ( fiop_tbl[fnum].state == NOTREADY ) {
		u.u_error = ENXIO;
		return;
	}


	/*  check device is available  */
	if ( devstat(ctl, mdev) == UNAVAILABLE ) {
		u.u_error = ENXIO;
		return;
	}

	/*      if streamer and open    */
	if ((devtype(ctl,mdev) == STREAMER) && opflag(fnum, mdev) )
	{
		u.u_error = EBUSY;	/* tape device already in use */
		return;
	}

	/* If unopened device */
	if ( !opflag(fnum, mdev) ) {

		/* 
		 * Get an me250cmd buffer to load command into.
		 */

		cmd = getcmdbuf(fnum);

		id = devlog(mdev);
		id |= (devcont(mdev) <<4);
		create_command(TEST_CMD, id,(short)INTERR|DRIV_CMD,0 , 0, 0, cmd);

		send_command(cmd, fnum);

		if ( set_error(cmd->error) != 0) {
			u.u_error = set_error(cmd->error);
			relcmdbuf(fnum,cmd);
			return;
		}

		/*  all is well  so open device */
		setopflag(fnum, mdev);   
		relcmdbuf(fnum, cmd);

		/* As device is open in "opflag[fnum][mdev]", u.u_error 
		 * MUST = 0 to allow system/user to close the device. 
		 */
 
		u.u_error = 0;
	}


	/* If streamer tape */
	if ( (devtype(ctl,mdev) == STREAMER) ) {

		/*  set options from mdev number */
		if ( devslice(mdev) & BIT_ZERO ) {
			dev_info[fnum][devnum(mdev)].option |= BIT_ZERO; 
		}
		if ( devslice(mdev) & BIT_ONE ) {
			dev_info[fnum][devnum(mdev)].option |= BIT_ONE; 
		}
		if ( devslice(mdev) & BIT_TWO ) {
			dev_info[fnum][devnum(mdev)].option |= BIT_TWO; 
		}

		/* By default the streamer will have its parameters set  *
		 * to be in streaming mode.(Parameter number 1)  	 */

		cmd = getcmdbuf(fnum);

		id = devlog(mdev);
		id |= (devcont(mdev) <<4);

		create_command(STPARAM_CMD,id,(short)INTERR|DRIV_CMD,0,0,1,cmd);

		send_command(cmd, fnum);

		/* check for errors */
		if ( set_error(cmd->error) != 0) {
			u.u_error = set_error(cmd->error);
			resetopflag(fnum, mdev);   
			relcmdbuf(fnum,cmd);
			return;
		}
		relcmdbuf(fnum,cmd);

		/* retension the tape if bit two is set and tape is
		   opened for writing.					*/

		if ( (flag & (FWRITE)) && (devslice(mdev) & BIT_TWO) ) {
			cmd = getcmdbuf(fnum);

			id = devlog(mdev);
			id |= (devcont(mdev) <<4);

			create_command(RETEN_CMD, id,(short)INTERR|DRIV_CMD,0 , 0, 0, cmd);

			send_command(cmd, fnum);

			/* check for errors */
			if ( set_error(cmd->error) != 0) {
				u.u_error = set_error(cmd->error);
				resetopflag(fnum, mdev);   
				relcmdbuf(fnum,cmd);
				return;
			}
			relcmdbuf(fnum,cmd);

			/* we must not return to the user until device
			   is no longer busy 		      		*/

			wait_device_free(fnum,mdev);
		}

	}
}


/*
 *   opflag() returns the number of times the device has been opened.
 *   openflag[i] represents  slice i of the device.
 */

int
opflag(fnum, mdev)
	unsigned short fnum;
	register dev_t mdev;
{

	return ( dev_info[fnum][devnum(mdev)].openflag[devslice(mdev)] );
}

/*
 *   setopflag() sets the openflag field for a particular device.
 *   openflag[0] represents slice 0, ... , openflag[7] represents slice 7.
 *   The action it takes depends upon whether the device can be partitioned.
 */

void
setopflag(fnum, mdev)
	unsigned short fnum;
	register dev_t mdev;
{
	register int i;

	switch ( sd_conf[fnum][devnum(mdev)].dev_type) {

	case WINCH :
		dev_info[fnum][devnum(mdev)].openflag[devslice(mdev)]++;
		break;

	case STREAMER :
		/*  can not be partitioned                        */
      		/*  so set openflag such that each slice is open  */
		for (i = 0; i < 8 ; i++)
			dev_info[fnum][devnum(mdev)].openflag[i]++; 
		break;
	}
}


/*
 *   resetopflag() resets the openflag field for a particular device.
 *   openflag[0] represents slice 0, ... , openflag[7] represents slice 7.
 *   The action it takes depends upon whether the device can be partitioned.
 */

void
resetopflag(fnum, mdev)
	unsigned short fnum;
	register dev_t mdev;
{
	register int i;

	switch ( sd_conf[fnum][devnum(mdev)].dev_type) {

	case WINCH :
		dev_info[fnum][devnum(mdev)].openflag[devslice(mdev)] = 0;
		dev_info[fnum][devnum(mdev)].option = 0;
		break;

	case STREAMER :
		/*  can not be partitioned                           */
      		/*  so reset openflag such that each slice is closed */
		for (i = 0; i < 8 ; i++)
			dev_info[fnum][devnum(mdev)].openflag[i] = 0; 
		dev_info[fnum][devnum(mdev)].option = 0;
		break;
	}
}

/*
 *	me250start is used to send a queue of commands to an ME250 coprocessor.
 *
 *      To send a queue of commands, the driver must poll the location monitor
 *	of the ME250 where the commands are to be sent. The location monitor
 *	will indicate when the ME250 is free. When the ME250 is free, the driver
 *	must write to the ME250's location monitor. This will generate an 
 *	interrupt in the ME250. Then the start address of the queue must be
 *	written to a fixed location in the address space of the ME250.
 *
 *	The Me250 acknowledges the interrupt by writing to its location monitor.
 *	The Me250 will clear the location containing the address of the queue.
 */

void
me250start(fnum)
	unsigned short  fnum;
{
	long i;
	short location_mon;
	struct me250cmd *cmd;
	int x;

#ifdef DEBUG
	printf("active %d\n", bdctab[fnum].NCMDS_QUEUE1);
	printf("non-active %d\n", bdctab[fnum].NCMDS_QUEUE2);
	printf("busy_flag %d\n", bdctab[fnum].b_active);
#endif

	x = spl5();
	/*  if ME250/2 is busy  */
	if ( bdctab[fnum].b_active ) {
		splx(x);
		return;
	}

	/*  set as active  */
	bdctab[fnum].b_active++;


	/*  check there is something on the queue  */
	if ( !bdctab[fnum].b_actf ) {
		/*  queue is empty */
		bdctab[fnum].b_active = 0;
		splx(x);
		return;
	};

	/* get value in location monitor of ME250/2 */
	location_mon = *fiop_tbl[fnum].me250_loc;

#ifdef DEBUG
	printf("Check for FVME250 busy\n");
#endif

	/* Check for ME250 busy */
	if (location_mon & BIT_ONE) {
		/*   ME250 was busy   */
		bdctab[fnum].b_active = 0;
		printf("Me250 was busy.......\n");
		splx(x);
		return;
	}

	/* make the inactive queue the active queue */
	bdctab[fnum].b_forw = bdctab[fnum].b_actf;
	bdctab[fnum].b_back = bdctab[fnum].b_actl;
	bdctab[fnum].b_actf = bdctab[fnum].b_actl = NULL;
	
	bdctab[fnum].NCMDS_QUEUE2 = bdctab[fnum].NCMDS_QUEUE1;	
	bdctab[fnum].NCMDS_QUEUE1 = 0;

	/* Write to ME250 location monitor */
	*fiop_tbl[fnum].me250_loc |= BIT_ZERO; 	

	/* Write address of command in ME250 queue register */
	/* At the moment give a local address for ME250/2   */

#ifdef DEBUG
	printf("Giving address %d\n",((long)((long)bdctab[fnum].b_forw - 
				     (long)fiop_tbl[fnum].me250_mem)) );
	printf("NCMDS queue = %d", bdctab[fnum].NCMDS_QUEUE2);
#endif


	*(fiop_tbl[fnum].me250_queue) = ((long )((long)bdctab[fnum].b_forw -
					(long)fiop_tbl[fnum].me250_mem));

	/* get acknowledge from ME250 */
	for (i = QUEUE_TIMEOUT; i && *fiop_tbl[fnum].me250_queue ;--i)
		;	

	if (!i) {
		/*   ME250 didn't acknowledge   */
		printf("FVME250 no %d did not acknowledge", fnum);
		bdctab[fnum].b_active = 0;
		return;
	}

	/*  We are not going to wait for ever, for an interrupt back  */
	/*  from the ME250.  					      */
	/*  So set a timeout.					      */

	settimeout(fnum);

#ifdef DEBUG
	printf("Leaving me250start\n");
#endif
	splx(x);
}


/*
 *  	initstart is used to send an initialization command at startup
 *	time to an ME250/2.
 */

void
initstart(fnum)
	unsigned short  fnum;
{
	long i;
	short location_mon;


	/*  set as active  */
	bdctab[fnum].b_active++;


	/* get value in location monitor of ME250/2 */
	location_mon = *fiop_tbl[fnum].me250_loc;

	/* Check for ME250 busy */
	if (location_mon & BIT_ONE) {
		/*   ME250 was busy   */
		return;
	}


	/* Write to ME250 location monitor */
	*fiop_tbl[fnum].me250_loc |= BIT_ZERO; 	

	/* Write address of command in ME250 queue register */
	/* At the moment give a local address for ME250/2   */

	/* make the inactive queue the active queue */
	bdctab[fnum].b_forw = bdctab[fnum].b_actf;
	bdctab[fnum].b_back = bdctab[fnum].b_actl;
	bdctab[fnum].b_actf = bdctab[fnum].b_actl = NULL;
	
	bdctab[fnum].NCMDS_QUEUE2 = bdctab[fnum].NCMDS_QUEUE1;	
	bdctab[fnum].NCMDS_QUEUE1 = 0;

#ifdef DEBUG
	printf("Giving address %x\n",((long)((long)bdctab[fnum].b_forw - 
				     (long)fiop_tbl[fnum].me250_mem)) );
#endif

	*(fiop_tbl[fnum].me250_queue) = ((long )((long)bdctab[fnum].b_forw - 
					(long)fiop_tbl[fnum].me250_mem));

	/* get acknowledge from ME250 */
	for (i = QUEUE_TIMEOUT; i && *fiop_tbl[fnum].me250_queue ;--i)
		;	

	if (!i) {
		/*   ME250 didn't acknowledge   */
		printf("FVME250 no %d did not acknowledge", fnum);
		return;
	}
}




/*
 *	timedout is called by the kernel routine timeout(). It is called
 *	when a sufficient amount of time has elapsed without receiving an
 *      interrupt back from the ME250. It marks this ME250 as unavailable
 *	and prints an error message.
 */

timedout(fnum)
	unsigned short fnum;
{
	struct me250cmd *cmd;
	struct buf      *bp;


	/*  To arrive at this routine, we can not have received an interrupt
	 *  from the ME250 coprocessor.
	 */


	/*   mark this ME250 as closed  */
	fiop_tbl[fnum].state = NOTREADY;
	
	/*   scream blue murder         */
	cmn_err(CE_WARN,"\007\007FVME250 %d did not return an interrupt\n", fnum);
	cmn_err(CE_WARN,"FVME250 %d is now unavailable\n", fnum);

#ifdef DEBUG
	/* for testing purposes */
	printf("b_actf = 0x%x\n", bdctab[fnum].b_actf);
	printf("b_actl = 0x%x\n", bdctab[fnum].b_actl);
	printf("b_forw = 0x%x\n", bdctab[fnum].b_forw);
	printf("b_back = 0x%x\n", bdctab[fnum].b_back);
	printf("b_active = 0x%x\n", bdctab[fnum].b_active);
	printf("QUEUE1 = 0x%x\n", bdctab[fnum].NCMDS_QUEUE1);
	printf("QUEUE2 = 0x%x\n", bdctab[fnum].NCMDS_QUEUE2);
#endif

	/*   get first command in the queue  */
	cmd = (struct me250cmd *)bdctab[fnum].b_forw;

	if (cmd->intsreg & DRIV_CMD) {
		cmd->error = TIMEOUT;
		wakeup(cmd);
		return;
	}

	bp = cmd->bufp;
	bp->b_flags |= B_ERROR;
	bp->b_error = EIO;

	iodone(bp);
}





/*  
 *	m250intr() is called by the kernel in response to an interrupt 
 *	received from an ME250. It is passed a parameter which tells us,
 *	which ME250 has interrupted us.
 */




void
m250intr( param )
register param;
{
	register unsigned short fnum;
	register int mdev;
	register struct me250cmd *cmd;
	register struct buf      *bp;
	register struct iotime   *stp;
	register struct iobuf    *ip;
	register char dev_num;
	register int error_flag = 0;
	register long transfer;
	int i;


#ifdef DINTR
	printf("Inside m250intr param %d\n", param);
#endif

	/*  get fnum from parameter passed by low level interrupt routine  */
	fnum = param/14;

	/*
	 *   Commands issued by open, close, or ioctl procedures cause the
	 *   interrupt procedure to wake up the driver code which is asleep.
	 */


	/*   get first command in the queue  */
	cmd = (struct me250cmd *)bdctab[fnum].b_forw;

#ifdef DINTR
	printf("cmd = %d fnum = %d\n",cmd->cmd, fnum);
#endif

	/*  do we have a spurious interrupt  */
	if (cmd == NULL) {
		logstray(fiop_tbl[fnum].me250_mem);
		printf("M250intr: Spurious interrupt on FVME250 %d \n", fnum);
		return;
	}

	/* if we have been called because of an initialization command */
	if (cmd->cmd == INIT_CMD ) {
		remove_command(fnum);
		bdctab[fnum].b_active = 0;
		clear_intflag(fnum);
		return;
	}
		
	
	/*   turn off timeout() if not called by init routine */
	untimeout(fiop_tbl[fnum].timeout_id);

	/*  check that device was available              */
	/*  if not mark it as closed                     */
	/*  do not do this for copy or restore  commands */
	if ( (cmd->error == NOTAVAIL)  && (cmd->cmd < 0xe) ) {
		/*  get device number  */
		dev_num = cmd->cntrlid >> 3;
		dev_num |= (cmd->cntrlid & BIT_ZERO);
		dev_info[fnum][dev_num].status = UNAVAILABLE;
	}
	

	if (cmd->intsreg & DRIV_CMD) {
		remove_command(fnum);
		wakeup(cmd);
		if ( bdctab[fnum].NCMDS_QUEUE2 == 0) {
			bdctab[fnum].b_active = 0;
			me250start(fnum);
		}	
		/* else set a timeout for the next command */
		else {
			settimeout(fnum);
		}
		/* allow the ME250 to interrupt CPU */
		clear_intflag(fnum);
		return;
	}


	/*  else we have been called by read/write commands   */
	
	/*  get hold of the original buf structure            */
	bp = cmd->bufp;
	if (bp == 0) {
		cmn_err(CE_PANIC,"Inside M250intr:BP == 0 !!!");
	}

	/* remove old bufp */
	cmd->bufp = NULL;

#ifdef DINTR
	printf("Intr : cmd  = 0x%x\n",cmd);
	printf("Intr : 0xbp = %x\n",bp);
	printf("bp->b_bcount = %x\n",bp->b_bcount);
	printf("paddr(bp)    = %x\n",paddr(bp));
#endif

	mdev = minor(bp->b_dev);
	ip = &bdtab[fnum][devnum(mdev)];


	/*    check for serious errors which need to be logged */
	/*    message also printed on the console              */

	switch (cmd->error ) {
		case 0:
			break;
		case DATAERR:
			cmn_err(CE_WARN,"Bad block no %d (decimal) found on mdev %d me250 %d",cmd->addr, mdev, fnum);
			break;
		case TIMEOUT:
		case DMA:
		case SCSIERR:
		case POWRESET:
		case SOFTBUG:
		case DEVERROR:
		cmn_err(CE_WARN,"FVME250intr:error no 0x%x mdev %d on FVME250 %d\n", cmd->error,mdev, fnum);
		fill_eblock(fnum,mdev,cmd);
		logberr(ip, bp->b_flags&B_ERROR);   
		error_flag++;
	}

	/* convert local address to ME250/2 to a VME address */
	cmd->buff_addr = (char *)((long)cmd->buff_addr + (long)fiop_tbl[fnum].me250_mem);


	if ((bp->b_flags & B_READ) && (!error_flag) ) {
		transfer = bp->b_bcount;
		if ( (cmd->error == FILEMARK) || (cmd->error == EOT) )  {
			bp->b_resid += BSIZE*(cmd->count);
			transfer = bp->b_bcount - BSIZE*(cmd->count); 
		}
#ifdef DINTR
	printf("Int read b_resid = 0x%x trans = 0x%x\n",bp->b_resid,transfer);
	printf("bp->b_bcount = 0x%x\n",bp->b_bcount);
#endif
			
		/*  copy data to requested address  */
		my_blkcopy(cmd->buff_addr,paddr(bp),transfer);
	}


	/*  release data space reserved in data pool  */

#ifdef DINTR
	printf("freeing memory num = %x addr = %x\n",BSIZE*btod(bp->b_bcount),
		cmd->buff_addr);
#endif

	me250mfree(fiop_tbl[fnum].datamap, BSIZE*btod(bp->b_bcount),
	   	   cmd->buff_addr);

	/*  make sure b_bcount field is correct */
	if ( (cmd->error == FILEMARK) || (cmd->error == EOT) )  {
		bp->b_bcount = transfer;
	}


	if ( (devtype((fnum/2),mdev) == STREAMER) && (bp->b_flags & B_READ) )  {
		/* indicate we have just done a read */
		dev_info[fnum][devnum(mdev)].option &= ~BIT_FOUR;
		dev_info[fnum][devnum(mdev)].option |= BIT_SIX;
		if ((cmd->error == FILEMARK) && (transfer))
			/*  indicate that we have read a filemark  */
			dev_info[fnum][devnum(mdev)].option |= BIT_SEVEN;
		else
			dev_info[fnum][devnum(mdev)].option &= ~BIT_SEVEN;
	}

	if ((devtype((fnum/2),mdev) == STREAMER) && !(bp->b_flags & B_READ) )  {
		/* indicate we have written some data */
		dev_info[fnum][devnum(mdev)].option |= BIT_FOUR;
	}


	if ( set_error(cmd->error) ) {
		bp->b_flags |= B_ERROR;
		if (bp->b_error == 0)
			bp->b_error = set_error(cmd->error);
	}


	remove_command(fnum);
	relcmdbuf(fnum, cmd);

	/*  statistical update  */
	stp = &m250stat[fnum][devnum(mdev)];
	stp->io_resp += lbolt - bp->b_start; 
	stp->io_act  += lbolt - ip->io_start;

	iodone(bp);

#ifdef DINTR
	printf("intr : done iodone\n");
#endif

	ip = &bdctab[fnum];
	if ( ip->NCMDS_QUEUE2 == 0) {
		bdctab[fnum].b_active = 0;
		me250start(fnum);
	}	

	/* else set a timeout for the next command */
	else {
		settimeout(fnum);
	}
	clear_intflag(fnum);
}

void
m250strategy(ctl, bp)
register ctl;
register struct buf *bp;
{
	register int mdev;
	register struct confg *confgp;
	register struct iotime *stp;
	register unsigned short fnum;
	register daddr_t nblks;
	register struct me250cmd *cmd;
	register char id;
	register long num;
	register char *buff;
	register int level;
	register long off;

#ifdef DSTRATEGY
	printf("Inside m250strategy\n");
	printf("strategy: bp = 0x%x\n", bp);
#endif
	if (bp == 0) {
		cmn_err(CE_PANIC,"M250strategy:BP == 0 !!!");
	}

	/*  get minor device number  */
	mdev = minor(bp->b_dev);

	/*  get  FIOP number */
	fnum = getfnum(ctl,mdev);

	/*  check that this FIOP is available  */
	if ( fiop_tbl[fnum].state == NOTREADY ) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		bp->b_resid = bp->b_bcount;
		iodone(bp);
		return;
	}

	/*  check this device is available  */
	if ( devstat(ctl, mdev) == UNAVAILABLE ) {
		bp->b_flags |= B_ERROR;
		bp->b_error = ENXIO;
		bp->b_resid = bp->b_bcount;
		iodone(bp);
		return;
	}


	/*   get configuration details about device  */
	confgp = &sd_conf[fnum][devnum(mdev)];

	/* ensure error flag is cleared */
	bp->b_error = 0;      
	bp->av_forw = NULL;   
	bp->av_back = NULL;    
	bp->b_start = lbolt;
	bdtab[fnum][devnum(mdev)].io_start = lbolt;

	off = bp->b_blkno + confgp->slice[devslice(mdev)].blkoff;

	bp->b_resid = bp->b_bcount;

#ifdef DSTRATEGY
	printf("Update info statistics\n");
#endif

	/*  statistical info update  */
	stp = &m250stat[fnum][devnum(mdev)];
	stp->io_cnt++;
	stp->io_bcnt += (bp->b_bcount +secsize(ctl,mdev)-1) / secsize(ctl,mdev);       
	nblks = confgp->slice[devslice(mdev)].nblocks;

#ifdef DSTRATEGY
	printf("off = 0x%x  bp->b_count = 0x%x\n",off, bp->b_bcount);
	printf("bp->b_blkno = 0x%x\n",bp->b_blkno);
	printf("bp->b_flags = 0x%x\n",bp->b_flags);
	printf("nblks = 0x%x\n",nblks);
#endif

	/* fix for streamer tape */
	if ( (devtype(ctl,mdev) == STREAMER) && (filemark(ctl,mdev)) )  {
		/* A zero read */
		dev_info[fnum][devnum(mdev)].option &= ~BIT_SEVEN;
		bp->b_resid = bp->b_bcount;
		bp->b_bcount = 0;
		iodone(bp);
		return;
	}

	

	if (bp->b_blkno + btod(bp->b_bcount) > (unsigned long)nblks ) {
		/*  an illegal or partial read/write  */
		register int newcount;

		newcount = dtob(nblks - bp->b_blkno);

#ifdef DSTRATEGY
	printf("Given a partial read/write\n");
	printf("newcount = %x\n",newcount);
#endif

		if ((newcount < 0) || (newcount == 0 && 
	           !(bp->b_flags & B_READ))) {
			/* write at end, or any illegal request */
			bp->b_flags |= B_ERROR;
			bp->b_error |= ENXIO;
			iodone(bp);
			return;
		}

		if (newcount == 0) {
#ifdef DSTRATEGY
	printf("Given a read at the end\n");
#endif
			/*  a read at the end  */
			bp->b_resid = bp->b_bcount;	/* zero read */
		        iodone(bp);
			return;
		}

		/*  this is a partial request , read or write!  */
		bp->b_resid = bp->b_bcount - newcount;
		bp->b_bcount = newcount;
	}
	else
		bp->b_resid = 0;

	/*  Everything is OK, so get a command, fill it in, */
	/*  and then call me250start			    */

	/* raise the priority level */
	spl5();

	cmd = getcmdbuf(fnum);

	/*  keep hold of buf structure  */
	cmd->bufp = bp;

	num = btod(bp->b_bcount);
	buff = NULL;
	while (!buff) {
#ifdef DSTRATEGY
	printf("Allocating memory: num = 0x%x\n",num);
#endif
		buff = (char *)me250malloc(fiop_tbl[fnum].datamap, num*BSIZE);
	}

	id = devlog(mdev);
	id |= (devcont(mdev) <<4);
	if (bp->b_flags & B_READ) {
#ifdef DSTRATEGY
	printf("Creating command for read\n");
	printf("num of blocks = %x\n",num);
	printf("off = 0x%x\n",off);
#endif
		create_command(READ_CMD,id, INTERR, off,
		               num, buff-(char *)fiop_tbl[fnum].me250_mem,cmd);
	}
	else {
		/*   copy data into pool  */
		my_blkcopy( (short *)paddr(bp), (short *)buff, (long)num*512 );
#ifdef DSTRATEGY
	printf("Creating command for write\n");
	printf("num of blocks = %x\n",num);
	printf("off = 0x%x\n",off);
#endif
		create_command(WRITE_CMD,id, INTERR, off,
                               num, buff-(char *)fiop_tbl[fnum].me250_mem, cmd);
	}

	add_command(fnum, cmd);

	if (bdctab[fnum].b_active == 0 );
		me250start(fnum);

	/* return priority level */
	spl0();
}


void
m250read(ctl,mdev)
int ctl;
dev_t mdev;
{
	me250check(ctl, mdev , B_READ);
}


void
m250write(ctl,mdev)
int ctl;
dev_t mdev;
{
	me250check(ctl, mdev , B_WRITE);
}


void 
m250print(ctl, mdev, msg) 
int ctl;
register dev_t mdev;
char *msg;
{
	printf("mdev %x, %s on FVME250 %d\n", mdev,msg,ctl);
	printf("controller %d\n", devcont(mdev) );
	printf("drive %d  slice %d\n", devlog(mdev), devslice(mdev) );
}


void
me250check(ctl, mdev, r_w)
int ctl;
register dev_t;
int r_w;
{
	unsigned short fnum;

	fnum = getfnum(ctl, mdev);

#ifdef DEBUG
	printf("Inside me250check ctl=%d  mdev=%d r_w = %d\n",ctl,mdev,r_w);
	printf("u.u_offset = %d  u.u_count = %d\n",u.u_offset,u.u_count);
	printf("bytes per sector = %d\n",sd_conf[fnum][devnum(mdev)].bytes_per_sector);
#endif


	/*  check u_count and u_base are multiples of block size */
	if (devtype(ctl,mdev) == WINCH) {
		if ( (u.u_offset | u.u_count) & 
	     	(sd_conf[fnum][devnum(mdev)].bytes_per_sector-1) ) {
			u.u_error = EINVAL;
			return;
		}
	}


	/*  strategy needs to know ctl number, so we call a pseudo strategy
         *  routine which passes this ctl number as well the buf structure
         *  to m250strategy().
	 */
#ifdef DEBUG
	printf("About to call strategy\n");
#endif

	if ( (devtype(ctl,mdev) == STREAMER) && (ctl == 0) ) {
		physio(me250breakup, 0 , mdev , r_w);
	}
	

	if ( (devtype(ctl,mdev) == WINCH) && (ctl == 0) ) {
		if (physck (sd_conf[fnum][devnum(mdev)].slice[devslice(mdev)].nblocks, r_w) )
			physio(me250breakup, 0 , mdev , r_w);
	}

	/*  for future, check for other ctl numbers          */
	/*  each ctl number can access 2 different ME250/2s  */
}

void
me250breakup(bp)
	register struct buf *bp;
{
	me250_dma_breakup(ps_0strategy, bp , BSIZE);
}



/*
 *	set_error() converts the error code returned by the firmware
 *	to existing UNIX error codes.
 */

unsigned short
set_error(err)
	unsigned short err;
{
switch (err) {
	case 0          :
		return(0);
	case SEEKERR    :
	case WRFAULT    :
	case CRCERR     :
	case IDERR      :
	case TIMEOUT    :
	case DATAERR    :
	case ILLCOMMD   :
	case ILLCNTRL   :
	case BLANKTAPE  :
	case INVALPARAM :
	case DMA        :
	case SCSIERR    :
	case POWRESET   :
	case BADSCSINUM :
	case BADDEVTYPE :
	case SOFTBUG    :
	case ILLLENGTH  :
	case HEADERERR  :
	case RDFAULT    :
	case NOTAPE     :
	case EOT        :
		return(EIO);
	case NOTAVAIL   :
	case INVALCONF  :
		return(ENXIO);
	case DEVBUSY    :
		return(EBUSY);
	case DEVERROR   :
		return(EFAULT);
	case FILEMARK   :
	case BOT	:
		return(0);
	default		:
		return(EIO);
	}
}	


/*
 *	m250ioctl() is called directly from the ioctl system call.
 *	The parameters to the ioctl procedure are minor device number,
 *	command number, argument and the mode (as in open and close routines).
 *	The argument parameter will depend on the command being sent.
 *
 *	Data which needs to be sent to the FVME250 will have to be copied from
 *	user space to the data pool.
 */

void
m250ioctl(ctl, mdev , cmd, arg, flag)
	register dev_t mdev;
	int ctl, cmd, arg, flag;
{
	short legality;
	short dev_type;
	struct me250cmd *ioctl_cmd;
	unsigned short fnum;
	char id;
	unsigned long addr  = 0;
	unsigned long count = 0;
	char *buff_addr     = NULL;
	register struct file *fp;
	register struct inode *ip;
	dev_t dev, m_dev;
	long value;
	struct cp_struct cp;
	int level;

	/*  get FIOP number  */
	fnum = getfnum(ctl, mdev);
	
	if (fiop_tbl[fnum].state == NOTREADY ) {
		u.u_error = ENXIO;
		return;
	}

	/*  check device is available  */
	if ( devstat(ctl, mdev) == UNAVAILABLE ) {
		u.u_error = ENXIO;
		return;
	}

	/* check for a MVME350 tape standard ioctl call */
	if (cmd & M350IOCTL ) {
		switch (cmd) {
			case M350REWIND:
				cmd = REWIND_CMD;
				break;
			case M350RETENSION:
				cmd = RETEN_CMD;
				break;
			case M350WRTFM:
				cmd = WRFMKS_CMD;
				break;
			case M350RDFM:
				cmd = SPFMKS_CMD;
				break;
			case M350SETDMA:
			case M350GETDMA:
			case M350ERASE:
			case M350BYTESWAP:
				return;
		}
	}

	/* check for a MOTOROLA disk standard ioctl call */
	if (cmd & DKIOCTL ) {
		dkioctl(ctl, mdev, cmd, arg, flag);
		return;
	}


 	/*  check range of command     */
	if ( (cmd < 0) || (cmd >= ME250CMDS) ) {
		u.u_error = EINVAL;
		return;
	}

	legality = legal[cmd];

	/*  is this command allowable  */
	if ( !legality) {
		u.u_error = EINVAL;
		return;
	}
	
	/*  do you need to be super-user  */
	if ( (legality & L_SUSER) && !suser() ) {
		u.u_error = EPERM;
		return;
	}

	/*  check that command is legal for this device */	
	dev_type = devtype(ctl, mdev);

	if ( (dev_type == STREAMER) && !(legality & L_STAPE) ) {
		u.u_error = EINVAL;
		return;
	}
	
	if ( (dev_type == WINCH) && !(legality & L_WINCH) ) {
		u.u_error = EINVAL;
		return;
	}
	

	/* SIZE_CMD does not need to access an ME250/2.
 	   It returns the size of the partition for a device.  */

	if (cmd == SIZE_CMD) {
		struct me250slice sl;
		unsigned long nbl;
	
		sl = sd_conf[fnum][devnum(mdev)].slice[devslice(mdev)];
		nbl = sl.nblocks ;
		copyout(&nbl, arg, sizeof(long) );
		return;
	}


	ioctl_cmd = getcmdbuf(fnum);
	id = devlog(mdev);
	id |= (devcont(mdev) <<4);

	switch (cmd) {

	case FORMAT_CMD :
	case RETEN_CMD :
	case REWIND_CMD :
	case GTPARAM_CMD :
		break;
	case REASSIGN_CMD:
		addr = arg;
		break;
	case WRITEH_CMD :
		
		buff_addr = (char *)me250malloc(fiop_tbl[fnum].datamap, 512);

		if ( copyin( arg, buff_addr, 512) < 0 ) {
			u.u_error = EINVAL;
			relcmdbuf(fnum, ioctl_cmd);
			return;
		}
		
		/* convert VME address to local address for ME250/2 */
		buff_addr = (char *)((long)buff_addr - (long)fiop_tbl[fnum].me250_mem);
		break;

	case READH_CMD :
		buff_addr = (char *)me250malloc(fiop_tbl[fnum].datamap, 512);
		/* convert VME address to local address for ME250/2 */
		buff_addr = (char *)((long)buff_addr - (long)fiop_tbl[fnum].me250_mem);
		break;

	case WRFMKS_CMD :
	case SPFMKS_CMD :
		count = arg;
		break;
		
	case STPARAM_CMD :
		buff_addr = (char*)arg;
		break;
	case COPY_CMD :
	case RESTORE_CMD : {
		struct me250slice sl;
		unsigned long off;
	
		sl = sd_conf[fnum][devnum(mdev)].slice[devslice(mdev)];
		off = sl.blkoff ;

		copyin( arg, &cp, sizeof(struct cp_struct) );

		buff_addr = (char *)(id<<16);

		addr  = cp.offset + off;
		count = cp.nblocks; 

		/*  we need to find mdev number from a file descriptor */
		fp = u.u_ofile[cp.fd];
		ip = fp->f_inode;
		dev = (dev_t)ip->i_rdev;
		m_dev = minor(dev);        /*  at last! */

		id = devlog(m_dev);
		id |= (devcont(m_dev) << 4);

		buff_addr = (char *)( (unsigned long)buff_addr |(id) );

		id = 0;

		break;
		}
	default :
		u.u_error = EINVAL;
		relcmdbuf(fnum, ioctl_cmd);
		return;
	}

	create_command(cmd, id,(short)INTERR|DRIV_CMD, addr,count,buff_addr,ioctl_cmd); 
	send_command(ioctl_cmd, fnum);

	if ( ioctl_cmd->error) {
		u.u_error = set_error(ioctl_cmd->error);
		relcmdbuf(fnum, ioctl_cmd);
		return;
	}

	/*  action now depends upon what command we have just serviced  */
	switch (cmd) {
	case RETEN_CMD :
	case REWIND_CMD :
		wait_device_free(fnum,mdev);
		break;
	case GTPARAM_CMD :
		value = (long)(ioctl_cmd->buff_addr);
	
		copyout(&value, arg, sizeof(long) );
		break;

	case WRITEH_CMD :
		/* convert local address to a VME address */
		buff_addr = (char *)((long)buff_addr + (long)fiop_tbl[fnum].me250_mem);
		me250mfree(fiop_tbl[fnum].datamap, 512, buff_addr);
		break;

	case READH_CMD :
		/* convert local address to a VME address */
		buff_addr = (char *)((long)buff_addr + (long)fiop_tbl[fnum].me250_mem);
		if ( copyout( buff_addr, arg , 512) < 0 ) {
			u.u_error = EINVAL;
			me250mfree(fiop_tbl[fnum].datamap, 512, buff_addr);
			relcmdbuf(fnum, ioctl_cmd);
			return;
		}
		me250mfree(fiop_tbl[fnum].datamap, 512, buff_addr);
		break;
	}
	
	relcmdbuf(fnum, ioctl_cmd);
	u.u_error = 0;
	if ( set_error(ioctl_cmd->error) ) {
		u.u_error = set_error(ioctl_cmd->error);
	}

	/* Change option setting for tape device */
	if ( (devtype((fnum/2),mdev) == STREAMER) )  {
		dev_info[fnum][devnum(mdev)].option &= ~BIT_SEVEN;
		dev_info[fnum][devnum(mdev)].option &= ~BIT_SIX;
	}

	return;
}


/* 	
 *	dkioctl is used to implement MOTOROLA disk standard ioctl calls
 */

dkioctl(ctl, mdev, cmd, arg, flag)
	dev_t mdev;
	int ctl, cmd, arg, flag;
{
	struct me250cmd *ioctl_cmd;
	int fnum,id;

	/*  get FIOP number  */
	fnum = getfnum(ctl, mdev);

	/* get a command buffer */
	ioctl_cmd = getcmdbuf(fnum);


	if (cmd == DKGETSLC) {
		copyout(sd_conf[fnum][devnum(mdev)].slice, arg, 
		        8*(sizeof(struct me250slice)) );
		return;
	}

	if (cmd == DKSETSLC) {

		/*  is user the super-user  */
		if ( !suser() ) {
			u.u_error = EPERM;
			return;
		}

		copyin(arg, sd_conf[fnum][devnum(mdev)].slice, 
	               8*(sizeof(struct me250slice)) );
		return;
	}

	/*  check for a format command */
	if (cmd == DKFORMAT) {

		/*  is user the super-user  */
		if ( !suser() ) {
			u.u_error = EPERM;
			return;
		}

		/* work out id no */
		id = devlog(mdev);
		id |= (devcont(mdev) << 4);

		create_command(FORMAT_CMD, id,INTERR|DRIV_CMD, (1<<16),0,0,ioctl_cmd); 
		send_command(ioctl_cmd, fnum);

		/* check for an error */
		if ( ioctl_cmd->error) {
			u.u_error = set_error(ioctl_cmd->error);
		}

		relcmdbuf(fnum, ioctl_cmd);
		return;
	}

	/* 
	 *  At the present time ignore all other MOTOROLA
	 *  standard ioctl requests. This may be changed at a 
	 *  future date.
	 *
         *  Do not return an error for these ioctl calls.
	 */
	
}

/*
 *	Dummy dump routine
 */

m250dump()
{
	/* do nothing */
}


int
my_blkcopy(from, to, count)
register short *from, *to;
register long count;
{
	for (; count > 0; count = count - 2)
		*to++ = *from++;
}

print_command(cmd)
struct me250cmd *cmd;
{
	static long num_times;
	int delay_time;

	for (delay_time = 0x400000; delay_time; delay_time--)
		;
	num_times++;

	printf("no_times = 0x%x\n",num_times);
	printf("cmd      = 0x%x\n",cmd);
	printf("cmd->cmd = 0x%x\n",cmd->cmd);
	printf("cmd->cntrlid = 0x%x\n",cmd->cntrlid);
	printf("cmd->instreg = 0x%x\n",cmd->intsreg);
	printf("cmd->addr = 0x%x\n",cmd->addr);
	printf("cmd->count = 0x%x\n",cmd->count);
	printf("cmd->buff_addr = 0x%x\n",cmd->buff_addr);
	printf("cmd->error = 0x%x\n",cmd->error);
	printf("cmd->next_cmd = 0x%x\n",cmd->next_cmd);
	printf("cmd->bufp = 0x%x\n",cmd->bufp);              

}

/*
 *	settimeout - sets a timeout for the next command to be sent to an 
 *	ME250.
 */

settimeout(fnum)
	int fnum;
{
	struct me250cmd *cmd;

	/* timeouts for copy/restore/format commands need to be a lot larger 
	   than for normal commands. */

	/*   look at next command on the queue  */
	cmd = (struct me250cmd *)bdctab[fnum].b_forw;
	if ( (cmd->cmd == COPY_CMD) || (cmd->cmd == RESTORE_CMD) ||
	     (cmd->cmd == FORMAT_CMD) ) {
       		fiop_tbl[fnum].timeout_id = timeout(timedout, fnum , 30*WAIT_TIMEOUT);
		return;
	}
	if ( cmd->cmd == SPFMKS_CMD) {
       		fiop_tbl[fnum].timeout_id = timeout(timedout, fnum , 30*WAIT_TIMEOUT);
	}
	else
       		fiop_tbl[fnum].timeout_id = timeout(timedout, fnum , WAIT_TIMEOUT);
}


/* 
	send_command() does the following:
	It raises the interrupt level.
        It adds the command onto the queue.			
	If the appropiate ME250/2 is not busy then it calls me250start().
        The procedure sleeps.
	It lowers the interrupt level.	
	It prevents the problem of being woken up before we fall asleep.
*/
	
send_command(cmd, fnum)
	struct me250cmd *cmd;
	unsigned short fnum;
{
	int level;

	level = spl5();
	add_command(fnum, cmd);

	if (bdctab[fnum].b_active == 0) 
		me250start(fnum);
	sleep(cmd, PRIBIO);
	splx(level);
}



/* 
	clear_intflag() clears a location in an ME250/2's memory.
	This allows the ME250/2 to now interrupt us.
*/

int
clear_intflag(fnum)
	int fnum;
{
	short *location;

	/* clear a flag in an ME250s memory indicating that we are now able
	   to receive another interrupt */

	location = (short *)( (long)fiop_tbl[fnum].me250_mem + SOFTREADY );
	*location = (short)0;
}


/*
	send_test_cmd() sends a TEST_CMD to an ME250/2.
*/

int send_test_cmd(fnum,mdev)
	unsigned short fnum;
	dev_t mdev;
{
	struct me250cmd *cmd;
	int id;

#ifdef DEBUG
	printf("Inside send_test_cmd() mdev 0x%x fnum 0x%x\n", mdev,fnum);
#endif

	/* Get an me250cmd buffer to load command into. */

	cmd = getcmdbuf(fnum);

	id = devlog(mdev);
	id |= (devcont(mdev) <<4);
	create_command(TEST_CMD, id,(short)INTERR|DRIV_CMD,0 , 0, 0, cmd);

	send_command(cmd, fnum);

	if ( set_error(cmd->error) != 0) {
		if (cmd->error == DEVBUSY) {
			relcmdbuf(fnum,cmd);
			return(1);
		}
		else {	
			u.u_error = set_error(cmd->error);
			relcmdbuf(fnum,cmd);
			return(-1);
		}
	}		
	relcmdbuf(fnum,cmd);
	return(0);
}

/*
 *  	wait_device_free does not return until an error is detected,
 *	or the device becomes "free" ie non-busy.
 */

int wait_device_free(fnum,mdev) 
	unsigned short fnum;
	dev_t mdev;
{
	int flag;

	while (1) {	
		/* sleep for five seconds */
		delay(60*5);

		flag = send_test_cmd(fnum,mdev);

		/* if device is not busy */
		if (flag == 0)
			break;

		/* if device is busy */
		if (flag == 1)
			continue;

		break;
	}
}

int fill_eblock(fnum,mdev,cmd) 
	unsigned short fnum;
	dev_t mdev;
	struct me250cmd *cmd;
{
	struct eblock *ep;
	struct buf *bp;
	int n;
	struct iostat *iosp;
	struct iobuf *ip;

	ip = &bdtab[fnum][devnum(mdev)];
	iosp = &m250stat[fnum][devnum(mdev)].ios;

	ep = (struct eblock *)geteslot( sizeof(struct eblock) );
	if ( ep == NULL) {
		iosp->io_unlog++;
		return;
	}

	bp = cmd->bufp;

	n = major(bp->b_dev);

	ep->e_dev = makedev(n, (bp == NULL)?minor(bp->b_dev):minor(bp->b_dev));
	ep->e_regloc = (physadr)fiop_tbl[fnum].me250_mem;
	ep->e_bacty = blkacty;
	ep->e_stats.io_ops = iosp->io_ops;
	ep->e_stats.io_misc = iosp->io_misc;
	ep->e_stats.io_unlog = iosp->io_unlog;

	if (bp != NULL) {
		ep->e_bflags = (bp->b_flags&B_READ) ? E_READ : E_WRITE;
		if (bp->b_flags & B_PHYS)
			ep->e_bflags |= E_PHYS;
		if (bp->b_flags & B_MAP)
			ep->e_bflags |= E_FORMAT;
		ep->e_bnum = bp->b_blkno;
		ep->e_bytes = bp->b_bcount;
		ep->e_memadd  = paddr(bp);
	}
	else
		ep->e_bflags = E_NOIO;
	ep->e_cyloff = 0;
	ep->e_nreg  = 1;

	ip->io_erec = ep;
}

int print_eblock(fnum,mdev) 
	unsigned short fnum;
	dev_t mdev;
{
	struct eblock ep;
	ep = *bdtab[fnum][devnum(mdev)].io_erec;

	printf("fnum = 0x%x mdev=0x%x\n", fnum, mdev);

	printf("e_dev 0x%x      physadr 0x%x\n",ep.e_dev, ep.e_regloc); 
	printf("e_bacty 0x%x\n",ep.e_bacty); 
	printf("e_stats.io_ops 0x%x\n",ep.e_stats.io_ops); 
	printf("e_stats.io_misc 0x%x\n",ep.e_stats.io_misc); 
	printf("e_stats.io_unlog 0x%x\n",ep.e_stats.io_unlog);
	printf("e_bflags 0x%x   e_cyloff 0x%x\n",ep.e_bflags, ep.e_cyloff);
	printf("e_bnum   0x%x   e_bytes 0x%x\n",ep.e_bnum, ep.e_bytes);
	printf("e_memadd 0x%x   e_rtry 0x%x\n",ep.e_memadd, ep.e_rtry);
	printf("e_nreg 0x%x\n",ep.e_nreg);
}

int print_stats(fnum,mdev) 
	unsigned short fnum;
	dev_t mdev;
{
	struct iotime it;

	it = m250stat[fnum][devnum(mdev)];

	printf("Stats mdev = 0x%x fnum = 0x%x\n",mdev,fnum);
	printf("ios.io_ops = 0x%x\n", it.ios.io_ops);
	printf("ios.io_misc = 0x%x\n", it.ios.io_misc);
	printf("ios.io_unlog = 0x%x\n", it.ios.io_unlog);
	printf("io_bcnt = 0x%x\n", it.io_bcnt);
	printf("io_act = 0x%x\n", it.io_act);
	printf("io_resp = 0x%x\n", it.io_resp);
}
