/*
 *	Copyright (c) 1987 Motorola Inc
 *	All Rights Reserved
 *	@(#)gentape.c	35.1
 */

#ident	"@(#)gentape.c	35.1"

/*
 *	Indirect driver for controlling TAPE.CART and /dev/r4[01][a][t][n].
 *
 *	On all SYSTEM V/68 systems, the node /dev/TAPE.CART is defined
 *	to be the system's standard (streaming) tape character device
 *	regardless of which peripheral is used. For example, if the
 *	MVME350 is used, then /dev/TAPE.CART needs to access the first
 *	MVME350. If the MVME327 is used, then /dev/TAPE.CART needs to
 *	access the first tape unit on the MVME327.
 *
 *	To accomplish this function, this driver during its init
 *	attempts to probe a series of addresses in an array
 *	defined in space.h and compiled into conf.c.  This array also
 *	contains the block and character major number for each driver
 *	which might be indirectly referenced.
 *
 *	The first successfully probed address in the list determines
 *	the driver it will indirectly call.
 *
 *	The bprobes may be circumvented if the kernel has been booted
 *	from a tape device.  This is determined by checking the value
 *	of bootctlr to see if it corresponds to any of the values for
 *	the possible tape controllers.  If it does then this determines
 *	the driver it will indirectly call.
 *
 *	The results of the above two cases may be overridden by the
 *	values stored in the variables, btape_maj and ctape_maj.
 *	If these variables are initialized with values of NULL then
 *	they have no effect and thus only the bprobes are used.
 *
 */

/* 	Additions for FVME250. Author: M. Woods  Date: 31/1/89
 *
 * 	If the normal probes should fail to find a tape device, then a search
 *	is conducted for FVME250s.
 *
 *	The first FVME250 present on the VMEbus must contain a streamer
 *	tape drive and the drive is operational. If it is not available
 *	then /dev/TAPE.CART, etc is not available to the system. 
 *
 *	Finding a FVME250 is done by searching for the strings 'LAmB' and 'FIOP'
 *	at addresses 0x404 and 0x408 respectively, in the FVME250's RAM.
 *
 *	Having found a FVME250, to determine a streamer tape drive is available
 *	the FVME250 firmware as to be probed.
 *
 *	The FVME250 firmware contains a record of information on each device
 *	connected to the SCSI bus. This information is kept in an array called
 *	tsk_rec[], at RAM address F250TSKADR. The streamer tape drive entry
 *	is found by probing the array control[] in the FVME250, at address
 *	F250CTLADR. The fourth entry in control[] contains the offset into 
 *	tsk_rec for the streamer tape drive.
 *
 *	Each entry in tsk_rec contains a status byte. This byte when
 *		1 = Device CLOSED, not available
 *		2 = Device OPEN, device available
*/

#include "sys/types.h"
#include "sys/param.h"
#include "sys/immu.h"
#include "sys/sysmacros.h"
#include "sys/buf.h"
#include "sys/fs/s5dir.h"
#include "sys/errno.h"
#include "sys/signal.h"
#include "sys/region.h"
#include "sys/proc.h"
#include "sys/psw.h"
#include "sys/user.h"
#include "sys/conf.h"
#include "sys/tty.h"
#include "sys/stream.h"
#include "sys/mvmecpu.h"
#include "sys/gentape.h"

#include "sys/fvme250/f250value.h"
#include "sys/fvme250/f250fix.h"

extern int tape_cmaj, tape_bmaj;	/* majors from conf.c */

/* from the sysgen parameters for defining the major numbers */
/* otherwise tapeinit() does probes to determine the major numbers */
/* declared in space.h */
extern int *btape_maj, *ctape_maj;
/* declared in space.h */
extern int *b350_maj, *c350_maj, *b147_maj, *c147_maj, *b327_maj, *c327_maj;

	/* Probes addresses in the FVME250's RAM */
extern me250probe();

extern struct gentape f250_list;
extern short f250_minor[];  /* Table of fvme250 driver minor device numbers */
extern int f250_ctladdr;
extern int f250_tskaddr;

/*
 *	tapeinit: generic tape init
 */

tapeinit()
{
	extern int bootctlr;
	extern int bprobe();
	extern int cpuboard;
	extern struct gentape gt_list[];
	register struct gentape *gen_ptr;
	register char **prb_addr;
	static int bbootdev, cbootdev;

	if ((char *)btape_maj == NULL) {
		switch (bootctlr) {
		case BOOT350_00:
		case BOOT350_10:
			btape_maj = b350_maj;
			break;
		case BOOT147_00:
			bbootdev = *b147_maj;
			btape_maj = &bbootdev;
			break;
		case BOOT147_01:
			bbootdev = *b147_maj + 1;
			btape_maj = &bbootdev;
			break;
		case BOOT327_00:
			bbootdev = *b327_maj;
			btape_maj = &bbootdev;
			break;
		case BOOT327_01:
			bbootdev = *b327_maj + 1;
			btape_maj = &bbootdev;
			break;
		case BOOT327_10:
			bbootdev = *b327_maj + 8;
			btape_maj = &bbootdev;
			break;
		case BOOT327_11:
			bbootdev = *b327_maj + 9;
			btape_maj = &bbootdev;
			break;
		}
	}
	if ((char *)ctape_maj == NULL) {
		switch (bootctlr) {
		case BOOT350_00:
		case BOOT350_10:
			ctape_maj = c350_maj;
			break;
		case BOOT147_00:
			cbootdev = *c147_maj;
			ctape_maj = &cbootdev;
			break;
		case BOOT147_01:
			cbootdev = *c147_maj + 1;
			ctape_maj = &cbootdev;
			break;
		case BOOT327_00:
			cbootdev = *c327_maj;
			ctape_maj = &cbootdev;
			break;
		case BOOT327_01:
			cbootdev = *c327_maj + 1;
			ctape_maj = &cbootdev;
			break;
		case BOOT327_10:
			cbootdev = *c327_maj + 8;
			ctape_maj = &cbootdev;
			break;
		case BOOT327_11:
			cbootdev = *c327_maj + 9;
			ctape_maj = &cbootdev;
			break;
		}
	}

	if (btape_maj != NULL && ctape_maj != NULL)
		return;

	for (gen_ptr = gt_list;
	    (prb_addr = (char **)(gen_ptr->gt_addr)) != NULL; gen_ptr++) {
		register int cpu_brd;

		if (*prb_addr == NULL || (cpu_brd = gen_ptr->gt_cpubrd) && 
		    cpu_brd != cpuboard)
			continue;
		if (!bprobe(*prb_addr, -1)) {
			if ((char *)btape_maj == NULL) {
				btape_maj = gen_ptr->gt_bmaj;
			}
			if ((char *)ctape_maj == NULL) {
				ctape_maj = gen_ptr->gt_cmaj;
			}
			break;
		}
	}
	if (btape_maj != NULL && ctape_maj != NULL)
		return;
		/* Probe for a possible FVME250 streamer drive */	
	{
		register unsigned long base;
		register unsigned long end;
		long me250_lamb;
		long me250_fiop;
		short control_ent;
		short *tsk_ent;

		switch (cpuboard) {
			case FVME252:
				base = PROBEST_252;
				end = PROBEEND_252;
				break;
			case VME141:
				base = PROBEST_141;
				end = PROBEEND_141;
				break;
			default:
				base = PROBEST_132;
				end = PROBEEND_132;
				break;
		}
		for (; base < end; base += PROBEINC){
			if (me250bprobe( (base + BRDID), -1))
				continue;
			me250_lamb = *((long *) (base + BRDID));
			me250_fiop = *((long *) (base + BRDTYPE));
			if (strncmp(LAMBSTR, (char *)&me250_lamb, IDLEN))
				continue;
			if (strncmp(FIOPSTR, (char *)&me250_fiop, TYLEN))
				continue;
			control_ent = *(short *)((unsigned long)f250_ctladdr + base + 8);
			control_ent &= 0xff;
			if (control_ent == -1)
				break;
			tsk_ent = (short *)(f250_tskaddr + ((long)control_ent * 22) + 4 + base);
			if ( (*tsk_ent & 0xff) == 1)
				break;
			btape_maj = f250_list.gt_bmaj;
			ctape_maj = f250_list.gt_cmaj;
		}
	}
}

/*
 *	do_maj: find major number
 */

short
do_maj(dev, tape_xmaj, xtape_maj, x350_maj)
register int dev;
int tape_xmaj;
register int *xtape_maj, *x350_maj;
{
	register short i, x_maj;

	i = major(dev) - tape_xmaj;
	if (i < 0 || i > 1 || (char *)xtape_maj == NULL) {
		u.u_error = ENXIO;
		return(NOTAPE);
	}
	x_maj = *xtape_maj;
	if (xtape_maj != x350_maj) {
		x_maj += i;
	}
	return(x_maj);
}

/*
 *	do_min: find minor number
 */

short
do_min(dev, x_maj, x350_maj)
register int dev;
short x_maj;
register int *x350_maj;
{
	register short x_min;

	x_min = minor(dev);
	if ((char *)x350_maj != NULL && x_maj == (short)(*x350_maj)) {
		x_min = (x_min & 0x0003)|
		      ((x_min & 0x0004)<<5)|
		      ((x_min & 0x00e0)>>3);
	}
	else 
		if ( x_maj == (short)(*f250_list.gt_bmaj) || 
		     x_maj == (short)(*f250_list.gt_cmaj)) 
			x_min = f250_minor[x_min];
	return(x_min);
}

/*
 *	tapebopen: block device open
 */

tapebopen(dev, flag)
register int dev;
{
	register short b_maj, b_min;
	short do_maj(), do_min();

	if ((b_maj = do_maj(dev, tape_bmaj, btape_maj, b350_maj)) == NOTAPE) 
		return;

	b_min = do_min(dev, b_maj, b350_maj);
	dev = makedev(b_maj, b_min);
	(*bdevsw[b_maj].d_open)(dev, flag);
}

/*
 *	tapecopen: character device open
 */

tapecopen(dev, flag)
register int dev;
{
	register short c_maj, c_min;
	short do_maj(), do_min();

	if ((c_maj = do_maj(dev, tape_cmaj, ctape_maj, c350_maj)) == NOTAPE) 
		return;

	c_min = do_min(dev, c_maj, c350_maj);
	dev = makedev(c_maj, c_min);
	(*cdevsw[c_maj].d_open)(dev, flag);
}

/*
 *	tapebclose: block device close
 */

tapebclose(dev, flag)
register int dev;
{
	register short b_maj, b_min;
	short do_maj(), do_min();

	if ((b_maj = do_maj(dev, tape_bmaj, btape_maj, b350_maj)) == NOTAPE) 
		return;

	b_min = do_min(dev, b_maj, b350_maj);
	dev = makedev(b_maj, b_min);
	(*bdevsw[b_maj].d_close)(dev, flag);
}

/*
 *	tapecclose: char. device close
 */

tapecclose(dev, flag)
register int dev;
{
	register short c_maj, c_min;
	short do_maj(), do_min();

	if ((c_maj = do_maj(dev, tape_cmaj, ctape_maj, c350_maj)) == NOTAPE) 
		return;

	c_min = do_min(dev, c_maj, c350_maj);
	dev = makedev(c_maj, c_min);
	(*cdevsw[c_maj].d_close)(dev, flag);
}

/*
 *	tapestrategy: strategy routine - should only be called with
 *			a block device since it should only be called
 *			directly from the OS! That is, no calls to
 *			physio() or other routine should be made using
 *			the address of tapestrategy as an argument.
 */

tapestrategy(bp)
register struct buf *bp;
{
	register short b_maj;
	short do_maj();

	if ((b_maj = do_maj(bp->b_dev, tape_bmaj, btape_maj, b350_maj))
	    == NOTAPE) 
		return;

	(*bdevsw[b_maj].d_strategy)(bp);
}

/*
 *	taperead: read the tape
 */

taperead(dev)
register int dev;
{
	register short c_maj, c_min;
	short do_maj(), do_min();

	if ((c_maj = do_maj(dev, tape_cmaj, ctape_maj, c350_maj)) == NOTAPE) 
		return;

	c_min = do_min(dev, c_maj, c350_maj);
	dev = makedev(c_maj, c_min);
	(*cdevsw[c_maj].d_read)(dev);
}

/*
 *	tapewrite: write the tape
 */

tapewrite(dev)
register int dev;
{
	register short c_maj, c_min;
	short do_maj(), do_min();

	if ((c_maj = do_maj(dev, tape_cmaj, ctape_maj, c350_maj)) == NOTAPE) 
		return;

	c_min = do_min(dev, c_maj, c350_maj);
	dev = makedev(c_maj, c_min);
	(*cdevsw[c_maj].d_write)(dev);
}

/*
 *	tapeioctl: perform ioctl on tape -
 *			requires common ioctl command set!
 */

tapeioctl(dev, cmd, arg, mode)
register int dev;
{
	register short c_maj, c_min;
	short do_maj(), do_min();

	if ((c_maj = do_maj(dev, tape_cmaj, ctape_maj, c350_maj)) == NOTAPE) 
		return;

	c_min = do_min(dev, c_maj, c350_maj);
	dev = makedev(c_maj, c_min);
	(*cdevsw[c_maj].d_ioctl)(dev, cmd, arg, mode);
}

tapeprint (dev, str)
register int dev;
{
	register short b_maj, b_min;
	short do_maj(), do_min();

	if ((b_maj = do_maj(dev, tape_bmaj, btape_maj, b350_maj)) == NOTAPE) 
		return;

	b_min = do_min(dev, b_maj, b350_maj);
	dev = makedev(b_maj, b_min);
	(*bdevsw[b_maj].d_print)(dev, str);
}
