#ident  "@(#)m146818tdc.c	35.1"

/*
 *  Driving routines for the Motorola
 *  MC146818 time of day clock.
 *  Copyright, 1988,  Motorola, Inc.
 */


#include "sys/types.h"
#include "sys/todc.h"


/*
 *  Defines used in conjunction with the clock registers
 */
#define TODC_ADDR	tdcaddr
#define  C_BSET		0x80		/* enable set flag in register B     */
#define  C_AINIT	0		/* initial control values in reg A   */
#define  C_BINIT	0x06		/* initial control values in reg B;
					   binary data, 24 hour clock        */

/*
 *  Memory map of the MC146818 Clock Registers
 */
struct clk_map
{
   char fill0,  c_secs;			/* seconds elapsed in current minute */
   char fill2,  c_sec_alarm;		/* when to generate alarm interrupt  */
   char fill4,  c_mins;			/* minutes elapsed in current hour   */
   char fill6,  c_min_alarm;		/* when to generate alarm interrupt  */
   char fill8,  c_hrs;			/* hours elapsed in current day	     */
   char fill10, c_hrs_alarm;		/* when to generate alarm interrupt  */
   char fill12, c_day_wk;		/* day of week (1-7); sunday=1       */
   char fill14, c_day_mon;		/* day of month (1-31);              */
   char fill16, c_mon;			/* month of year (1-12)	             */
   char fill18, c_yr;			/* year of century (0-99)            */
   char fill20, c_a;			/* register A		             */
   char fill22, c_b;			/* register B		             */
   char fill24, c_c;			/* register C		             */
   char fill26, c_d;			/* register D		             */
};

extern unsigned char *tdcaddr;

static char dmsize[12] = {31,28,31,30,31,30,31,31,30,31,30,31};


/***************
 **  m6818tdc **
 ***************
 **
 **  FUNCTION:  Entry point for all work done with the MC146818
 **		Time of Day clock.  
 **
 **  CALLED:	From within the kernel, specifically by todc.
 **
 **  RETURNS:	m6818tdc returns an integer value of
 **			   0 on success 
 **			or
 **			   -1 on failure of action (for any reason)
 **
 **  NOTES:	[1] No use is made of the VRT (valid RAM/TODC) bit (register D)
 **		    in these routines.  A higher level, "generic" validation of
 **		    the TODC contents is made in routine todc.
 **		[2] The daylight savings enable (DSE, register B) is never set
 **		    in this implementation as the beginning of daylight savings
 **		    time has changed since the MVME050 was designed and not
 **		    all customer sites abide by daylight savings time.
 **
 *****************/
m6818tdc(action, tdctime)
    int action;
    time_t *tdctime;
{
	register struct clk_map *clk = (struct clk_map *) TODC_ADDR;

	extern void	w_6818tdc();
	extern time_t	r_6818tdc();


	ENTRY("m6818tdc");

	switch (action) {
	case INIT:
		/*
		 * Initialize read/write control registers.
		 * Fall thru to do the initial read of todc.
		 */
		clk->c_a = C_AINIT;
		clk->c_b = C_BINIT;
	case GETTIME:
		if ((*tdctime = r_6818tdc()) < 0) {
			EXIT("m6818tdc - read failed");
			return(-1);
		}
		break;
	case SETTIME:
		w_6818tdc(*tdctime);
		break;			/* assumes no error on set */
	default:
		EXIT("m6818tdc - unknown action");
		return(-1);
	}
	EXIT("m6818tdc");
	return(0);

}  /* end m6818tdc */


/****************
 **  w_6818tdc **
 ****************
 **
 **  FUNCTION:  Set time for the MC146818 Time of Day clock.
 **
 **  CALLED:	From m6818tdc ONLY.
 **
 **  RETURNS:	none.
 ****************/
void
w_6818tdc(oldtime)
    time_t oldtime;
{
   	register short	scratch;
	register struct clk_map *clk;
	int seconds, minutes, hours, dayofweek, day, month, year;

	ENTRY("w_6818tdc");
	clk = (struct clk_map *) TODC_ADDR;
	seconds = oldtime % 60;
	oldtime /= 60;
	minutes = oldtime % 60;
	oldtime /= 60;
	hours = oldtime % 24;
	oldtime /= 24;
	dayofweek = ((oldtime + 7340036) %7) + 1;
	for(year=70; oldtime >= (year%4 ? 365:366); year++)
		oldtime -= (year%4 ? 365:366);
	if(!(year%4)) dmsize[1] = 29;
	for(month=0; oldtime >dmsize[month]; month++)
		oldtime -= dmsize[month];
	dmsize[1]=28;
	month++;
	day = oldtime + 1;

	clk->c_b |= C_BSET;		/* Inhibit updates. */

	clk->c_sec_alarm = 0 ;   	/* Clear the alarm registers. */
	clk->c_min_alarm = 0 ;
	clk->c_hrs_alarm = 0 ;

	clk->c_secs = seconds ;   	/* Write the time/date registers. */
	clk->c_mins = minutes;
	clk->c_hrs = hours;
	clk->c_day_wk = dayofweek;
	clk->c_day_mon = day;
	clk->c_mon = month;
	clk->c_yr = year;

	clk->c_b &= ~C_BSET;    	/* Re-enable updates. */
	EXIT("w_6818tdc");
}  /* end w_6818tdc */


/****************
 **  r_6818tdc **
 ****************
 **
 **  FUNCTION:  Get time from the MC146818 Time of Day clock.
 **
 **  CALLED:	From m6818tdc ONLY.
 **
 **  RETURNS:	r_mk48tdc returns a "time_t" (long) value of
 **			   <time> on success (internal UNIX time format)
 **			or
 **			   -1 on failure of action (for any reason)
 **
 **  NOTES:	[1] The weekday value in the time of day clock is not used.
 **		[2] The read algorithm in this routine uses the SET bit of
 **		    register B to guarentee a lucid read of TODC.  This is
 **		    fine for occassional reads of the time of day clock, but
 **		    can cause problems in a tight loop which continually
 **		    reads TODC because the SET bit aborts/inhibits update
 **		    cycles for TODC registers.  If too many of these update
 **		    cycles are lost, an appreciable loss of time can occur
 **		    in the TODC.  If a continual, accurrate read capability
 **		    is required in the future, use of the "update-ended
 **		    interrupt" or the "update in progress" flag (register A)
 **		    should be investigated.
 **
 ****************/
time_t
r_6818tdc()
{
	register int tmptim;
	register struct clk_map *clk;
	int i,year,month, day, hour, minute, second;


	ENTRY("r_6818tdc");
	clk = (struct clk_map *) TODC_ADDR;

	clk->c_b |= C_BSET;     	/* Inhibit updates. */

	second = clk->c_secs;   	/* Read the registers. */
	minute = clk->c_mins;
	hour = clk->c_hrs;
	day = clk->c_day_mon;
	month = clk->c_mon;
	year = clk->c_yr;

	clk->c_b &= ~C_BSET;    	/* Re-enable updates. */

	/* 
	 * Validate time values read and convert to seconds since
	 * 1 Jan 00:00:00 1970 GMT.
	 */
	if ((year < 70) || (month > 12) || (month < 1) || (day > 31) ||
	    (day < 1) || (hour > 23) || (minute > 59) || (second > 59)) {
		EXIT("r_6818tdc - bad time values read");
		return(-1);
	}

	tmptim=0;
	year += 1900;
	for(i= 1970; i < year; i++)
		tmptim += i%4 ? 365 : 366;
	if(!(year % 4) && month >= 3)
		tmptim += 1;
	while(--month)
		tmptim += dmsize[month -1];
	tmptim += (day - 1);
	tmptim *= 24;
	tmptim += hour;
	tmptim *= 60;
	tmptim += minute;
	tmptim *= 60;
	tmptim += second;

	EXIT("r_6818tdc");
	return(tmptim);
}  /* end r_6818tdc */
