/*		Copyright (c) 1985 Motorola,1987 Inc.		*/
/*		All Rights Reserved				*/
/*	@(#)mfp.c	35.7		*/
/*
 *      Driver for MVME120 On-Board RS-232C Serial Port
 *      Uses Serial Port of MC68901 Multi-Function Peripheral   
 *
 *	NOTE: This driver does not support any type of modem
 *	      control - modem control is not available on the
 *	      MVME120 serial port.
 *
 *	NOTE: DTR is available on the MVME120 but is ignored in
 *            this driver.
 *
 *	NOTE: The break detect status in the mfp status register
 *	      does not work as advertised
 *	      in the book. You get 2 break detect interrupts, sometimes
 *	      only 1 and you do not get an interrupt at the end of 
 *	      break. I had to simulate the end of break with a
 *	      timeout (stinks, eh ?) and call mfpinputc.
 *
 */


#ident	"@(#)mfp.c	35.7"

#include "sys/param.h"
#include "sys/types.h"
#include "sys/sysmacros.h"
#include "sys/dir.h"
#include "sys/file.h"
#include "sys/signal.h"
#include "sys/tty.h"
#include "sys/user.h"
#include "sys/errno.h"
#include "sys/termio.h"
#include "sys/conf.h"
#include "sys/sysinfo.h"
#include "sys/mfp.h"
#include "sys/gentty.h"


/* MC68901 USART control register (UCR) masks */
#define WSZMSK          (~(WORDLEN8|WORDLEN7|WORDLEN6|WORDLEN5)&0xff)
#define STPMSK          (~(ASYNST1|ASYNST15|ASYNST2)&0xff)

/* Interrupt definitions */
#define RXTX_INT        (IRXRDY | ITXRDY | IRXERR | ITXERR)
#define RDY_INT         (IRXRDY | ITXRDY)
#define ERR_INT         (IRXERR | ITXERR)
#define RX_INT          (IRXRDY | IRXERR)
#define TX_INT          (ITXRDY | ITXERR)

/* Initialization constants (defaults) */
#define UCR_INIT        (CLOCK16 | WORDLEN8 | ASYNST2)  /* USART control     */
#define DEFBAUD         B9600		/* baud rate         */
#define DISABLE         0x00		/* rx/tx disable     */

/* Miscellaneous Definitions */
#define C_DATA  0xff
#define CR	0x0d
#define LF	0x0a
#define FALSE   0
#define TRUE    1
#define OFF     0
#define ON      1
#define CNTLO   0x0f
#define CONS    0
#define HI_MARK 18		/* clist high-water-mark               */

#define MFPCSCAN 	2	/* clock ticks for character scan      */
#define MFP_CHAR 	01	/* character arrived flag (mfp_work)   */
#define MFP_ACT 	02	/* character in buffer flag (mfp_work) */



/* Baud rate tables (MC68901 timer C)  */

/* Using 12.5 Mhz clock */
unsigned char mfp_ps12[] = {            
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        CDIV16,  /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        CDIV4,   /*  300 */
        CDIV4,   /*  600 */
        CDIV4,   /* 1200 */
        0,       /* 1800 */
        CDIV4,   /* 2400 */
        CDIV4,   /* 4800 */
        CDIV10,  /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};

unsigned char mfp_ct12[] = {    
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        55,      /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        81,      /*  300 */
        41,      /*  600 */
        20,      /* 1200 */
        0,       /* 1800 */
        10,      /* 2400 */
        5,       /* 4800 */
        1,       /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};

/* Using 10 Mhz clock */
unsigned char mfp_ps10[] = {            
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        CDIV10,  /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        CDIV4,   /*  300 */
        CDIV4,   /*  600 */
        CDIV4,   /* 1200 */
        0,       /* 1800 */
        CDIV4,   /* 2400 */
        CDIV4,   /* 4800 */
        CDIV4,   /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};

unsigned char mfp_ct10[] = {    
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        71,      /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        65,      /*  300 */
        33,      /*  600 */
        16,      /* 1200 */
        0,       /* 1800 */
        8,       /* 2400 */
        4,       /* 4800 */
        2,       /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};


/* Using 1.230679 Mhz clock */
unsigned char mfp_ps1[] = {            
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        CDIV4,   /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        CDIV4,   /*  300 */
        CDIV4,   /*  600 */
        CDIV4,   /* 1200 */
        0,       /* 1800 */
        CDIV4,   /* 2400 */
        CDIV4,   /* 4800 */
        CDIV4,   /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};

unsigned char mfp_ct1[] = {    
        0,       /*    0 */
        0,       /*   50 */
        0,       /*   75 */
        0x57,    /*  110 */
        0,       /*  134 */
        0,       /*  150 */
        0,       /*  200 */
        0x20,    /*  300 */
        0x10,    /*  600 */
        0x08,    /* 1200 */
        0,       /* 1800 */
        0x04,    /* 2400 */
        0x02,    /* 4800 */
        0x01,    /* 9600 */
        0,       /* EXTA */
        0        /* EXTB */
};


extern int		m120_addr;	/* MVME120 module control register   */
					/*	   (see conf.c)		     */
extern int 		*mfp_addr[];    /* MFP base address - in conf.c      */
extern struct tty 	mfp_tty[];     	/* MFP tty structure                 */
short 			mfp_state;	/* additional flags that should be
					   in the tty structure but are not. */
int mfpbrk;				/* break detect flag */
char    		mfp_work;	/* activity info for scanner	     */
char			clk12;		/* flag set if 12.5MHz clk (mfpinit) */
char			mfp_scan;	/* mfpscan on/off flag		     */
extern 			ttrstrt();
extern  int		consdev;	/* console device */
extern  int		dualcons;	/* alternate console device */
extern  int   		Dconcurrent;	/* dual console mode enable */

/* 	NOTE: the following variable initialization will be modified
	      in released version of this driver.
*/
extern int mfp_ivec;		/* interrupt vector for mfp (from config)   */
 
/* #define DEBUG	1 */

#ifdef DEBUG
char	mfp_diag = 0;	/* if 1 print diagnostic messages    */
#endif



/*
 *	open routine
 */
mfpopen(device, flag)
dev_t device;
{
        register struct tty *tp;
	register struct mfp_io *mfp;
	register dev_t  dev;	

	extern mfpproc();

	dev = minor(device);

        /* check for valid dev or no device */
        if ( (dev != CONS) || (mfp_addr[0] == NULL)) {
                u.u_error = ENXIO;
                return;
        }
	mfp = (struct mfp_io *)mfp_addr[0];
        tp = &mfp_tty[0];               /* get tty struct		     */
        tp->t_proc = mfpproc;

	/* Put process to sleep if the port is currently being used as a 
	   dual console. */ 
	if(Dconcurrent)
		/* sleep if port has previously been opened as a dual console */
		while(mfp_state & DCON){
			sleep(( caddr_t)&tp->t_canq,TTIPRI);
		}

	/* wait for close to complete */
	while(mfp_state & TTWCLOSE)
		sleep((caddr_t)&tp->t_canq, TTOPRI);

		/* Does not | WOPEN as there probably will not be a modem on console */
        if (!(tp->t_state & ISOPEN)) {
                ttinit(tp);             /* init tty structure 		     */
		tp->t_state |= CARR_ON;

		if(device == consdev)
			mfp_state |= SCON; 

                mfpparam();
		mfp->iera |= RX_INT;	/* enable Rx interrupts */
		mfp->imra |= RX_INT;	/* unmask Rx interrupts */
        }
        (*linesw[tp->t_line].l_open)(tp);
	if (mfp_scan == 0) {		/* if not on, turn on scanner	   */
		mfpscan();
		mfp_scan++;
#ifdef DEBUG
        	if (mfp_diag)
        		printf("mfpopen: mfpscan started.\n");
#endif
	}
#ifdef DEBUG
        if (mfp_diag)
                printf("mfpopen: opened dev %d\n", dev);
#endif
}



/*
 *	close routine
 */

mfpclose(dev)
dev_t dev;
{
        register struct tty *tp;
	int lvl;

	dev = minor(dev);

        tp = &mfp_tty[0];

	lvl = spltty();

	mfp_state |= TTWCLOSE;

        (*linesw[tp->t_line].l_close)(tp);

	mfp_state &= ~(SCON|TTWCLOSE);

	splx(lvl);
	
	wakeup(&tp->t_canq);

#ifdef DEBUG
        if (mfp_diag)
                printf("mfpclose: closed dev %d\n", dev);
#endif
}



/*
 *	read routine
 */

mfpread(dev)
dev_t dev;
{
        register struct tty *tp;

	dev = minor(dev);

        tp = &mfp_tty[0];
        (*linesw[tp->t_line].l_read)(tp);
#ifdef DEBUG
        if (mfp_diag)
                printf("mfpread: read dev %d\n", dev);
#endif
}

/*
 *	write routine
 */

mfpwrite(dev)
dev_t dev;
{
        register struct tty *tp;

	dev = minor(dev);

#ifdef DEBUG
        if (mfp_diag)
                printf("mfpwrite: write to dev %d\n", dev);
#endif
        tp = &mfp_tty[0];
        (*linesw[tp->t_line].l_write)(tp);
}



/*
 *	ioctl routine
 */
#define	CFLAGMASK	(CBAUD|CSIZE|CSTOPB|PARENB|PARODD|CREAD)

mfpioctl(device, cmd, arg, mode)
dev_t device;
unsigned int cmd, arg, mode;
{

	register struct tty *tp;
	register cflag;
	register dev_t dev;
	register struct mfp_io *mfp;

	dev = minor(device);
	mfp = (struct mfp_io *)mfp_addr[0];

#ifdef DEBUG
        if (mfp_diag)
                printf("mfpioctl:\n");
#endif
	tp= &mfp_tty[0];
	cflag= tp->t_cflag&CFLAGMASK;
	switch(cmd){

		case CONC_PUTC:
			mfpputc (dev, arg);
			break;

		case CONC_IPUTC:
			mfpinputc (dev, (arg&0xff), (arg >> 8));
			break;

		case CONC_BOOT:
			/* The system was booted in dual console mode.
			   The port is already open. Do not disturb the current
			   baud rate and protocol settings, but do enable 
			   interrupts. */
			mfp->iera |= RX_INT;	/* enable Rx interrupts */
			mfp->imra |= RX_INT;	/* unmask Rx interrupts */

			mfp_state |= DCON; 
			break;
			

		case CONC_ON:
			if( mfp_state & DCON){ /* already dual console */
				break;
			}
			if( ! (tp->t_state & (ISOPEN | WOPEN))){
				mfpopen(device, 0);  	

				/* Remember to close when dual console is turned				   off. */
				if(!u.u_error)
					mfp_state |= (DCON | DCCLOSE); 
				break;
			}
			/* Port is already open due to a previous open. Dont
			   disturb the current baud rate and protocol settings
			   on the port. */

			mfp_state |= DCON; 
			break;
		


		case CONC_OFF:
			if(!(mfp_state & DCON)){ /* already off */
				break;
			}

			if(mfp_state & DCCLOSE){
				mfpclose(device);
			}
			mfp_state &= ~(DCON | DCCLOSE);
			wakeup(&tp->t_canq);
			break;

		default:

        		if (ttiocom(&mfp_tty[0], cmd, arg, mode))
				if(cflag != (tp->t_cflag&CFLAGMASK))
                			mfpparam();
			break;

	} /* end of switch */
}



/*
 *	param routine
 */

mfpparam()
{
        register struct tty     *tp;
        register struct mfp_io  *mfp;
        register ushort         flags;
	register char 		ucrtmp;
	int			cur_lvl;

#ifdef DEBUG
        if (mfp_diag)
                printf("mfpparam:\n");
#endif
        mfp = (struct mfp_io *)mfp_addr[0];	/* pointer to MC68901 regs */
        tp = &mfp_tty[0];

        flags = tp->t_cflag;
        if ((flags & CBAUD) == B0) {    /* hang up on zero baud rate */
                tp->t_state &= ~BUSY;
                mfp->imra &= ~TX_INT;
                return;
        }
	ucrtmp = mfp->ucr;		/* get current USART cntl reg */

        /* Set baud rate */
        /* Switch S3, section 2 is read to determine mpu clock
           rate:
                open   - 12.5 MHz
                closed - 10 MHz
        */
        if (mfp_ct10[flags&CBAUD]) {		/* if valid baud rate */
                mfp->tcdcr &= ~TMCMODE;		/* turn off timer C   */
/*                if (clk12 & SWTCH2) { */	/* 12.5 MHz mpu clock */
/*                      mfp->tcdr = mfp_ct12[flags & CBAUD]; 
                        mfp->tcdcr |= mfp_ps12[flags & CBAUD];
                } else {               */         /* 10 MHz mpu clock   */
 /*                     mfp->tcdr = mfp_ct10[flags & CBAUD]; 
                        mfp->tcdcr |= mfp_ps10[flags & CBAUD];
                }
*/
                mfp->tcdr = mfp_ct1[flags & CBAUD]; 
                mfp->tcdcr |= mfp_ps1[flags & CBAUD];
		
        }       /* if not valid baud rate, don't attempt to change it */

        ucrtmp &= STPMSK & WSZMSK;       /* clear stp and sz bits */

        /* set character size */        
        if ((flags & CS8) == CS8)
                ucrtmp |= WORDLEN8;
        else if (flags & CS7)
                ucrtmp |= WORDLEN7;
        else if (flags & CS6)
                ucrtmp |= WORDLEN6;
        else    /* CS5 */
                ucrtmp |= WORDLEN5;

        /* set number of stop bits */
        if (flags & CSTOPB)
                ucrtmp |= ASYNST2;
        else
                ucrtmp |= ASYNST1;


        /* set parity options */
        if (flags & PARENB) {
                ucrtmp |= PARENAB;
                if (flags & PARODD)
                        ucrtmp &= ~PAREVEN;
                else
                        ucrtmp |= PAREVEN;
        } else
                ucrtmp &= ~PARENAB;

	cur_lvl = spl6();

	/* setup USART control register with new values */
	mfp->ucr = ucrtmp;

        /* enable/disable receiver */
        if (flags & CREAD)                      /* enable receiver? */
                mfp->rsr |= RXENAB;
        else
                mfp->rsr &= ~RXENAB;
        splx(cur_lvl);

}



/*
 *	proc routines
 */

mfpproc(tp, cmd)
register struct tty *tp;
unsigned int cmd;
{
        register struct mfp_io  *mfp;
	register int	piplev;
	int c;

	piplev = spl6();
        mfp = (struct mfp_io *)mfp_addr[0];

        switch (cmd) {

        case T_TIME:            /* resuming output after delay */
                tp->t_state &= ~(TIMEOUT | BUSY);
                mfp->tsr &= ~TXBREAK;
                goto start;

        case T_WFLUSH:          /* flush output queue */
                tp->t_tbuf.c_size -= tp->t_tbuf.c_count;
                tp->t_tbuf.c_count = 0;
                /* fall thru */

        case T_RESUME:          /* resume output, XON has been received */
                tp->t_state &= ~TTSTOP;
                mfp->tsr |= TXENAB;            /* re-enable the transmitter */
                if ( !(mfp->tsr & TXRDY))   /* if no chars to transmit break*/
                   break;
                /* fall thru */

        case T_OUTPUT:          /* handle first or next char */
        start: {
                register struct ccblock *tbuf;

                if (tp->t_state & (BUSY|TTSTOP))
                        break;
                tbuf = &tp->t_tbuf;
                if (tbuf->c_ptr == NULL || tbuf->c_count == 0) {
                        if (tbuf->c_ptr)
                                tbuf->c_ptr -= tbuf->c_size - tbuf->c_count;
                        if (!(CPRES & (*linesw[tp->t_line].l_output)(tp))) {
                                break;
                        }
                }
		mfp->udr = c = *tp->t_tbuf.c_ptr++;
		tp->t_tbuf.c_count--;
                tp->t_state |= BUSY;                    /* send char */
                mfp->imra |= TX_INT;		/* unmask tx ints */
		/* Also send character to dual console if enabled. */
		if(Dconcurrent  && (mfp_state & SCON) && c)
			dconsputc(c);
                break;
        }

        case T_SUSPEND:         /* stop output, XOFF has been received */
                tp->t_state |= TTSTOP;
		if( tp->t_state&BUSY )	/* waiting for interrupt? */
			break;
		mfp->tsr &= ~TXENAB;	/* disable the transmitter */
                break;

        case T_BLOCK:           /* send XOFF */
                tp->t_state &= ~TTXON;
                tp->t_state |= TBLOCK | TTXOFF;
		if( tp->t_state&BUSY )	/* waiting for interrupt? */
			break;
                mfp->imra |= TX_INT;
                break;

        case T_RFLUSH:  /* flush input queue.  if input blocked, send XON */
                if (!(tp->t_state & TBLOCK))
                        break;
                /* fall thru */

        case T_UNBLOCK:         /* send XON */
                tp->t_state &= ~(TTXOFF | TBLOCK);
                tp->t_state |= TTXON;
		if( tp->t_state&BUSY )	/* waiting for interrupt? */
			break;
                mfp->imra |= TX_INT;
                break;

        case T_BREAK:           /* send break condition for 1/4 second */
                mfp->tsr |= TXBREAK;
                tp->t_state |= TIMEOUT | BUSY;
                timeout(ttrstrt, tp, HZ/4);
                break;

        case T_PARM:            /* used with sxt driver (R5.2) */
                mfpparam();
                break;
        }
	splx(piplev);
}



/*
 *	init routine
 */

mfpinit()
{
        register struct mfp_io *mfp;
	int cur_lvl;

	cur_lvl=spl6();
        mfp = (struct mfp_io *)mfp_addr[0];
        mfp->rsr = DISABLE;    	/* disable receiver		*/
        mfp->tsr = DISABLE;    	/* disable transmitter		*/
        mfp->imra &= ~RXTX_INT;	/* mask all rx/tx ints		*/
        mfp->iera = TX_INT;	/* enable selected tx ints	*/
				/* NOTE: receive ints are enabled
					 at mfpopen		*/

        /* Set default baud rate */
        mfp->tcdcr &= ~TMCMODE;		/* turn off timer C */

        /* Switch S3, section 2 is read to determine mpu clock
           rate:
                open   - 12.5 MHz
                closed - 10 MHz
        */
/*	clk12 = mfp->gpip & SWTCH2; */  /* read switch S3 sect 2 for mpu clk */
/*      if (clk12 & SWTCH2) {  	    */	          /* 12.5 MHz mpu clock */
/*              mfp->tcdr = mfp_ct12[DEFBAUD]; 
                mfp->tcdcr |= mfp_ps12[DEFBAUD];
        } else {                          */        /* 10 MHz mpu clock   */
/*              mfp->tcdr = mfp_ct10[DEFBAUD]; 
                mfp->tcdcr |= mfp_ps10[DEFBAUD];
        }
*/
        mfp->tcdr = mfp_ct1[DEFBAUD]; 
        mfp->tcdcr |= mfp_ps1[DEFBAUD];

	mfp->vr = (mfp_ivec>>2) & UVECTOR;	/* use automatic EOI    */
        mfp->ucr = UCR_INIT;         	/* USART control	*/
        mfp->rsr = RXENAB;           	/* enable receiver	*/
        mfp->tsr = TXOUTHI;		/* force output high	*/
	mfp->tsr |= TXENAB;		/* enable transmitter	*/
 /*     *(char *)m120_addr &= ~CTSOFF;*/  /* turn on cts		*/
	mfp->gpip = 0;


#ifdef DEBUG
        if (mfp_diag)
        	printf("mfpinit: initialized mfp\n");
#endif
	splx(cur_lvl);
}



/*
 *	interupt handler
 */

mfpintr(dev)
dev_t dev;
{
        register struct tty     *tp;
        register struct mfp_io  *mfp;
        register unsigned char  rxdata, ctmp;
        register unsigned char  rxstat, txstat;
	register unsigned char  count;
        register int            flg;
	register unsigned char  c;
        char                    lbuf[3];
        int                     lcnt, cur_lvl;
	unsigned short		dcstatus;

	extern int mfpbreak();

        mfp = (struct mfp_io *)mfp_addr[0];
	tp = &mfp_tty[0];


/***
 ***
 ***	Get receive and transmit status;
 ***
 ***/
	rxstat = mfp->rsr;
	
/***
 ***
 ***	Receive processing
 ***
 ***/
	if (rxstat & (RXRDY|FRAMERR|PARERR|OVERERR|RXBREAK)) {
		
		rxdata = mfp->udr;		/* read rx data */

                ctmp = rxdata & 0x7F;


		/* If dual console enabled, and character received from 
		   alternate console then send character to main console driver.
	         */
		if(Dconcurrent && (mfp_state & DCON)){
			dcstatus = 0;
			if(rxstat&(FRAMERR|PARERR|OVERERR|RXBREAK)) {
				if(rxstat & PARERR)
					dcstatus |= DC_PARERR;
				if(rxstat & OVERERR)
					dcstatus |= DC_ROVRN;
				if(rxstat & FRAMERR)
					dcstatus |= DC_FRERR;
				if(rxstat & RXBREAK)
					dcstatus |= DC_BREAK;
			}
			consinputc(rxdata, dcstatus); 
			return;
		}


		if (!(tp->t_state&(ISOPEN|WOPEN)))
			goto TX;

/***
 ***
 ***	Check for XON/XOFF
 ***
 ***/
		if ((rxstat&RXRDY) && (tp->t_iflag&IXON)) {
                        if (tp->t_state & TTSTOP) {
                            if (ctmp == CSTART || tp->t_iflag & IXANY)
                                (*tp->t_proc)(tp, T_RESUME);
                        } else {
                            if (ctmp == CSTOP)
                                (*tp->t_proc)(tp, T_SUSPEND);
                        }
                        if (ctmp == CSTART || ctmp == CSTOP) {
				goto TX;
                        }
		}
/***
 ***
 ***	Check for errors or break
 ***
 ***/
		lcnt = 1;
		flg = tp->t_iflag;
		if (rxstat&PARERR && !(flg&INPCK))
			rxstat &= ~PARERR;
		if (rxstat&(FRAMERR|PARERR|OVERERR|RXBREAK)) {
			if (rxstat&RXBREAK) {
				if(!mfpbrk)
				{
					/* set flag to ignore other breaks */
					mfpbrk = 1;

					/* simulate end of break with timeout */
					timeout(mfpbreak,0,34); 
				}
				goto TX;
			} else {
				if (flg&IGNPAR) {
					goto TX;
				}
			}
			if (flg&PARMRK) {
				lbuf[2] = 0xff;
				lbuf[1] = 0x00;
				lcnt = 3;
				sysinfo.rawch += 2;
			} else
				rxdata = 0x00;
/***
 ***
 ***	Process normal character
 ***
 ***/
		} else {
		/*
		 *	If ISTRIP is set, strip input character to 7 bits,
		 *	or if less than CS8, strip possible parity bit,
		 *	instead of letting it pass thru.  This conforms with
		 *	the intent of ATT of not having parity in the data byte.
		 */

			if (flg & ISTRIP || (tp->t_cflag & CSIZE) < CS8)
				rxdata &= 0x7f;
			else {
			/*
			 * following line, not needed.
			 * was in original ATT source
			 * because of original ATT device.
			 */
				/*
				rxdata &= 0xff;
				*/
				if (rxdata == 0xff && flg&PARMRK) {
					lbuf[1] = 0xff;
					lcnt = 2;
				}
			}
		}

/***
 ***	Move character to receive control buffer
 ***
 ***/
		lbuf[0] = rxdata;
		if (tp->t_rbuf.c_ptr == NULL) {
			goto TX;
		}
                while (lcnt) {
                        cur_lvl = spl6(); /*block mfpscan() running at clk pri*/
                        *tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
                        if ( (count = --tp->t_rbuf.c_count) < HI_MARK ) {
                                if( !(tp->t_state & TBLOCK))
                                     mfpproc(tp, T_BLOCK);
                                if( !(mfp->rsr & RXRDY)  || count < 1 ) {
                                    mfp_work = 0;
                                    tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size - count;
                                    (*linesw[tp->t_line].l_input)(tp, L_BUF);
                                    if( tp->t_rawq.c_cc <= TTXOHI)
                                        mfpproc( tp, T_UNBLOCK );
                                }
                        } else    /* map with if ! RXRDY ) ?? */
                                mfp_work = MFP_ACT; /* note MFP_CHAR bit clrd */
                        splx(cur_lvl);
                } 
	} 
/***
 ***
 ***	Transmit processing
 ***
 ***/

TX:
	c = 0;
	txstat = mfp->tsr;
	if ( txstat&TXRDY ) {
                sysinfo.xmtint++;
		if (tp->t_state & TTXON) {
			mfp->udr = c = CSTART;
			tp->t_state &= ~TTXON;
		} else if (tp->t_state & TTXOFF) {
                        mfp->udr = c = CSTOP;
                        tp->t_state &= ~TTXOFF;
		} else if (tp->t_state & BUSY) {
			if (tp->t_state&TTSTOP) {
				mfp->tsr &= ~TXENAB;
				tp->t_state &= ~BUSY;
			} else {
				tp->t_state &= ~BUSY;
				mfp->imra &= ~TX_INT;	/* mask tx ints */
				mfpproc(tp, T_OUTPUT);
			}
		} else
			mfp->imra &= ~TX_INT;	/* mask tx ints */


	}
	if ( txstat&TXEOT ) {
			/* generated END bit */
		mfp->tsr &= ~TXENAB;	/* disable Tx */
		mfp->tsr |= TXENAB;	/* renable Tx to clear END */
	}
}

mfpbreak()
{
        register struct mfp_io  *mfp;
	int rxdata;

        mfp = (struct mfp_io *)mfp_addr[0];

	/* get any residual data, left from break */
	rxdata = mfp->udr;	

	/* do L_BREAK via mfpinputc */
	mfpinputc(0,0,DC_BREAK);

	/* process null character */
	mfpinputc(0,0,0);

	/* allow another break */
	mfpbrk = 0;
}

/***
 ***
 ***	End of mfpintr -- interrupt handler
 ***
 ***/

mfpinputc(dev, c, rxstat)
register unsigned int	dev;
register unsigned char	c;
register unsigned char	rxstat;

{
	register struct tty     *tp;
	register struct mfp_io  *mfp;
	register unsigned char  ctmp;
	register unsigned char  count;
	register int		flg;
	char                    lbuf[3];
	int                     lcnt, cur_lvl;

	mfp = (struct mfp_io *)mfp_addr[0];
	tp = &mfp_tty[0];

	if (!(tp->t_state&(ISOPEN|WOPEN)))
		return;

	flg= tp->t_iflag;
	
	switch (rxstat) {

	case DC_BREAK:
		if ( !(flg & IGNBRK) && (flg & BRKINT) ) {
			(*linesw[tp->t_line].l_input) (tp, L_BREAK);
		}
		return;

	case DC_PARERR:
	case DC_ROVRN:
	case DC_FRERR:
		
		lcnt = 1;
		flg = tp->t_iflag;
		
		if (!(flg & IGNPAR)) {
			if (!(flg & INPCK))
				rxstat &= ~DC_PARERR;
			if (rxstat & (DC_PARERR | DC_ROVRN | DC_FRERR)) {
				/*
		 		 *	If PARMRK is set, a character with a framing
		 		 *	or parity error is read as the three-character
		 		 *	sequence 0xFF, 0, X, where X is the data of the
		 		 *	character received in error. 
		 		 */

				if (flg & PARMRK) {					
					lbuf[2] = 0xFF;						
					lbuf[1] = 0;					
					lcnt = 3;					
					sysinfo.rawch += 2;				
				} 
				/*
		 		 *	If PARMRK (mark parity errors) is not set,
		 		 *	a framing or parity error which is not 
		 		 *	ignored is read as the character NUL (0).
		 		 */

				else c = 0;						
			}
		}
		break;

	default:
		lcnt = 1;

		sysinfo.rcvint++; /* Increment counter of receiver interrupts */

		/*	If start/stop output control is enabled in t_iflag,
		 *	the state of the tty is TTSTOP and the character 
		 *	read from the receive buffer is a CSTART, then call
		 *	the command processing routine to resume output.
		 *	If the character is a CSTOP, call the proc routine
		 * 	to suspend output.
		 */
		
		if (flg & IXON) {						
			ctmp= c & 0x7f;				
			if (tp->t_state & TTSTOP){			
				if ((ctmp == CSTART) || (flg & IXANY))		
   					(*tp->t_proc) (tp, T_RESUME);		
			} else {					
				if (ctmp == CSTOP)			
					(*tp->t_proc) (tp, T_SUSPEND);		
			  }						
			if (ctmp == CSTART || ctmp == CSTOP) {
				return; /* ignore these characters */	
			}
		}							
		/*
		 *	If ISTRIP is set, strip input character to 7 bits,
		 *	or if less than CS8, strip possible parity bit,
		 *	instead of letting it pass thru.  This conforms with
		 *	the intent of ATT of not having parity in the data byte.
		 */

		if (flg & ISTRIP || (tp->t_cflag & CSIZE) < CS8)
			c &= 0x7f;					
		
			/*
			 *	Otherwise, to avoid ambiguity, a valid
			 *	character of 0xFF is read as 0xFF, 0xFF.
			 */

		else if (c == 0xff && flg & PARMRK) {			
				lbuf[1]= 0xff;				
				lcnt= 2;				
		}		   						

		break;
	}

/***
 ***	Move character to receive control buffer
 ***
 ***/
		lbuf[0] = c;
		if (tp->t_rbuf.c_ptr == NULL) 
			return;	

                while (lcnt) {
                        cur_lvl = spl6(); /*block mfpscan() running at clk pri*/
                        *tp->t_rbuf.c_ptr++ = lbuf[--lcnt];
                        if ( (count = --tp->t_rbuf.c_count) < HI_MARK ) {
                                if( !(tp->t_state & TBLOCK))
                                     mfpproc(tp, T_BLOCK);
                                if( !(mfp->rsr & RXRDY)  || count < 1 ) {
                                    mfp_work = 0;
                                    tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size - count;
                                    (*linesw[tp->t_line].l_input)(tp, L_BUF);
                                    if( tp->t_rawq.c_cc <= TTXOHI)
                                        mfpproc( tp, T_UNBLOCK );
                                }
                        } else    /* map with if ! RXRDY ) ?? */
                                mfp_work = MFP_ACT; /* note MFP_CHAR bit clrd */
                        splx(cur_lvl);
                } 
}



/*
 *	scan routine
 */

mfpscan()
{
        register struct tty     *tp;

        if (mfp_work & MFP_CHAR) {
                tp = &mfp_tty[0];
                mfp_work = 0;
                if (tp->t_rbuf.c_ptr) {
                        tp->t_rbuf.c_ptr -= tp->t_rbuf.c_size
                                        - tp->t_rbuf.c_count;
                        (*linesw[tp->t_line].l_input)(tp, L_BUF);
                }
        }
        if (mfp_work & MFP_ACT)
                mfp_work |= MFP_CHAR;
        timeout(mfpscan, (struct tty *)0L, MFPCSCAN);
}

mfpputc(dev, c)
register unsigned int dev;
register char c;
{
        register struct mfp_io 	*mfp;
        register char 		state;
	register int 		cur_lvl;

        mfp = (struct mfp_io *)mfp_addr[0]; 	/* mfp address 		 */
	cur_lvl = spl6();

        state = mfp->imra;		/* save current int mask */
        mfp->imra &= ~RXTX_INT;		/* mask interrupts 	 */

        for (;;)
        {
                while (!(mfp->tsr & TXRDY)); 	/* wait for tx ready 	 */
                mfp->udr = c;                	/* xmit char 	 */
                if (c != LF)
                        break;
                c = CR;                         /* send CR if previous c==LF */
        }
        while (!(mfp->tsr & TXRDY));	  	/* wait for tx ready 	 */
        mfp->imra = state;                   	/* restore previous state */

	splx(cur_lvl);
}

