#ident  "@(#)mk48t02nvr.c	35.2"

/*
 *  Access routines for the Thomson Components - Mostek Corp.
 *  MK48T02/12 nvram chip.
 *  Copyright (c) 1988  Motorola, Inc.
 *  All Rights Reserved
 *  @(#)mk48t02nvr.c	35.2
 */


#include "sys/types.h"
#include "sys/nvram.h"



/*
 * The MK48T02/12 chip provides 2040 bytes of sequential storage.  This 
 * area has been mapped in the following fashion:
 *
 *	ADDRESS RANGE	SIZE 		FUNCTION	SEGMENT/minor dev #
 * 	------------- 	----	-----------------------	-------------------
 *	 0000 - 03ff	1024	Reserved for user  		0
 *	 0400 - 043f	  64	Reserved for networking		1
 *	 0440 - 05f7	 440	Reserved for OS use		3
 *	 05f8 - 07f7	 512	Reserved for BUG use		5
 *	 0000 - 07f7	2040	Total NVRAM available		7
 *
 * The actual address of any segment ("slice") can be determined by adding
 * the above starting address to the value of nvr_addr.
 */

STATIC struct slice {
	int  start;		/* base address of the "slice" */
	int  size;		/* size (in bytes) of the slice */
} nvrtab[] = {
		{ 0x000, 1024 },
		{ 0x400,   64 },
		{ 0x000,    0 },
		{ 0x440,  440 },
		{ 0x000,    0 },
		{ 0x5f8,  512 },
		{ 0x000,    0 },
		{ 0x000, 2040 }
};

extern int  nvr_addr[];		/* base addr of chip from master/dfile */


/***************
 **  mk48nvr  **
 ***************
 **
 **  FUNCTION:  Access point for all work done with the MK48T02/12 NVRAM.
 **
 **  CALLED:	From the kernel routine, nvram.
 **
 **  RETURNS:	mk48nvr returns an integer value of
 **			   0 or greater on success (# of bytes read/written)
 **			   ERR (-1) on "unknown" failure of action
 **			   BNDS_ERR (-2) on segmentation violation
 **			or
 **			   CMD_ERR (-3) on undefined action request
 **		These values are defined in sys/nvram.h.
 **		
 **  NOTES:	This routine views the NVRAM chip as a sequential collection
 **		of bytes (like an array) and numbers them from 0 thru N-1
 **		where N is the total number of bytes available on the NVRAM
 **		chip.  Actual addressing of the NVRAM chip bytes is done in
 **		the read/write routines.
 **
 *****************/

mk48nvr(action, seg, buf, bytes, offset)
    int action, seg, bytes, offset;
    char *buf;
{
	register int start, end, rval;

	
	ENTRY("mk48nvr");
	/*
	 *  After making sure that the NVRAM chip is on the cpuboard,
	 *  calculate the starting point for the action.  Do not allow
	 *  offset values less than 0 as this translates into a
	 *  boundary violation at segment start.
	 */
	if (nvr_addr[0] == 0) {
	    EXIT("mk48nvr - no chip");
	    return(NO_NVRAM);
	}
	if (offset < 0) {
	    EXIT("mk48nvr - negative offset");
	    return(BNDS_ERR);
	}
	start = nvrtab[seg].start + offset;

	switch (action) {
	case CK_BNDS:
		end = nvrtab[seg].start + nvrtab[seg].size - 1;
		if ((start + bytes - 1) > end)
			rval = BNDS_ERR;
		else rval = OK;
		break;
	case READ:
		rval = r_mk48nvr(start, bytes, buf);
		break;
	case WRITE:
		rval = w_mk48nvr(start, bytes, buf);
		break;
	default:
		rval = CMD_ERR;
		break;
	}

	EXIT("mk48nvr");
	return(rval);

}  /* end mk48nvr */


/****************
 **  w_mk48nvr **
 ****************
 **
 **  FUNCTION:  Write the MK48T02 NVRAM.
 **
 **  CALLED:	From mk48nvr ONLY.
 **
 **  RETURNS:	Number of bytes written, on success and ERR (-1)
 **		 otherwise.
 **
 ****************/

w_mk48nvr(where, number, data)
    int where, number;
    char *data;
{
	register int i = 0;
	register char *to, *from;


	/*
	 *  The argument "where" refers to the first byte of the target
	 *  area within the NVRAM chip.  Calculate the actual address
	 *  for writing and proceed to copy "number" bytes from "what"
	 *  into the NVRAM.
 	 */
	ENTRY("w_mk48nvr");
	to = (char *) (nvr_addr[0] + where);

	for (from=data, i=0; i < number; i++, from++)  {
		*to = *from;
		to++;		/* MK48T02 provides sequential byte access */
	}

	EXIT("w_mk48nvr");
	if (i == number)
		return(number);	    /* requested number of bytes written */
	else
		return(ERR);
}  /* end w_mk48nvr */


/****************
 **  r_mk48nvr **
 ****************
 **
 **  FUNCTION:  Read data from the MK48T02 NVRAM chip.
 **
 **  CALLED:	From mk48nvr ONLY.
 **
 **  RETURNS:	Number of bytes read, on success and ERR (-1)
 **		 otherwise.
 **
 ****************/

r_mk48nvr(where, number, buf)
    int where, number;
    char *buf;
{
	register int i;
	register char *to, *from;

	/*
	 *  The argument "where" refers to the first byte within the 
	 *  NVRAM chip of the desired data.  Calculate the actual address
	 *  for reading and proceed to copy "number" bytes from "where"
	 *  into buf.
	 */
	ENTRY("r_mk48nvr");
	from = (char *) (nvr_addr[0] + where);

	for (to=buf, i=0; i < number; i++, to++)  {
		*to = *from;
		from++;		/* MK48T02 provides sequential byte access */
	}

	EXIT("r_mk48nvr");
	if (i == number)
		return(number);	    /* requested number of bytes read */
	else
		return(ERR);

}  /* end r_mk48nvr */

