#ident  "@(#)mk48t02tdc.c	35.2"

/*
 *  Driving routines for the Thomson Components - Mostek Corp.
 *  MK48T02/12 time of day clock.
 *  Copyright, 1988,  Motorola, Inc.
 */


#include "sys/types.h"
#include "sys/todc.h"


/*
 * Time of Day Clock registers for the MK48T02/12 are viewed as
 *
 *  unsigned char todc[8]   where:
 *							bitmap of registers:
 *	todc[0] is the clock control/calibration reg. 	[W R S - - - - -]
 *	todc[1] is the seconds register (00-59)		[P - - - - - - -]
 * 	todc[2] is the minutes register (00-59)		[0 - - - - - - -]
 *	todc[3] is the hours register (00-23)		[K 0 - - - - - -]
 *	todc[4] is the weekday register (01-07)		[0 F 0 0 0 - - -]
 *	todc[5] is the date register (01-31)		[0 0 - - - - - -]
 *	todc[6] is the month register (01-12)		[0 0 0 - - - - -]
 *	todc[7] is the year register (00-99)		[- - - - - - - -]
 *						W=write, R=read, S=sign, P=stop
 *						K=kick start, F=frequency test
 */

#define READ	0x40
#define WRITE	0x80
#define STOP	0x80

#define TODC_ADDR	tdcaddr

static  int dmsize[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
static	unsigned char todcreg[8];

extern unsigned char *tdcaddr;


/*****************
 **  mk48access **
 *****************
 **
 **  FUNCTION:  Entry point for all work done with the MK48T02/12
 **		Time of Day clock.  
 **
 **  CALLED:	From within the kernel, specifically from the todc routine.
 **
 **  RETURNS:	mk48access returns an integer value of
 **			   0 on success 
 **			or
 **			   -1 on failure of action (for any reason)
 *****************/
mk48access(action, tdctime)
    int action;
    time_t *tdctime;
{
	unsigned char  *todc_ctl=TODC_ADDR;

	extern void	w_mk48tdc();
	extern time_t	r_mk48tdc();


	ENTRY("mk48access");

	switch (action) {
	case INIT:
		/*
		 * Calibration of MK48T02 todc is not currently being used.
		 * Zeroing the control register on init will effectively
		 * "turn off" all calibration functionality.  Then fall thru
		 * to do the initial read of todc.
		 */
		*todc_ctl = 0;
	case GETTIME:
		if ((*tdctime = r_mk48tdc()) < 0)
		{
			EXIT("mk48access");
			return(-1);
		}
		break;
	case SETTIME:
		w_mk48tdc(*tdctime);
		break;			/* assumes no error on set */
	default:
		EXIT("mk48access");
		return(-1);
	}
	EXIT("mk48access");
	return(0);

}  /* end mk48access */


/****************
 **  w_mk48tdc **
 ****************
 **
 **  FUNCTION:  Set time for the MK48T02 Time of Day clock.
 **
 **  CALLED:	From mk48access ONLY.
 **
 **  RETURNS:	none.
 ****************/
void
w_mk48tdc(ttime)
    register time_t ttime;
{
	register unsigned char  *todc_ctl=TODC_ADDR;
	register unsigned char  *tp;
	register int  i;
	int dayofweek,seconds,minutes,hours,day,month,year;

	extern void	start_mk48tdc();

	/*
	 * Check that todc is operational before attempting to set the time.
	 */
	ENTRY("w_mk48tdc");
	if (*(todc_ctl+1) & STOP)
		start_mk48tdc();

	/*
	 * Convert the "time" (seconds since 1 Jan 00:00:00 1970 GMT) 
	 * to year, month, date, day, hour, minutes, and seconds
	 */
	seconds = ttime % 60;
	ttime /= 60;
	minutes = ttime % 60;
	ttime /= 60;
	hours = ttime % 24;
	ttime /= 24;
	dayofweek = ((ttime + 7340036) %7) + 1;
	for(year=70; ttime >= (year%4 ? 365:366); year++)
		ttime -= (year%4 ? 365:366);
	if(!(year%4)) dmsize[1] = 29;
	for(month=0; ttime >dmsize[month]; month++)
		ttime -= dmsize[month];
	dmsize[1]=28;
	month++;
	day = ttime + 1;

	/*
	 * Convert to BCD representation and write to the clock registers.
	 * Updates must be inhibited before write and re-enabled afterwards.
	 * The "zero-valued required" bits, (e.g. kick start and frequency 
	 * test bits) are taken care of by the BCD conversion.
	 */
	todcreg[1] = ((seconds /10)<<4) | (seconds%10);
	todcreg[2] = ((minutes /10)<<4) | (minutes%10);
	todcreg[3] = ((hours /10)<<4) | (hours%10);
	todcreg[4] = dayofweek;
	todcreg[5] = ((day /10)<<4) | (day%10);
	todcreg[6] = ((month /10)<<4) | (month%10);
	todcreg[7] = ((year /10)<<4) | (year%10);

	*todc_ctl |= WRITE;	/* inhibit clkreg updates/enable writing */

	tp = todc_ctl + 1;
	for (i=1; i < 8; i++, tp++)
		*tp = todcreg[i];

	*todc_ctl &= ~WRITE; 	/* re-enable clkreg updates/disable writing */
	EXIT("w_mk48tdc");
}  /* end w_mk48tdc */


/****************
 **  r_mk48tdc **
 ****************
 **
 **  FUNCTION:  Get time from the MK48T02 Time of Day clock.
 **
 **  CALLED:	From mk48access ONLY.
 **
 **  RETURNS:	r_mk48tdc returns a "time_t" (long) value of
 **			   <time> on success (internal UNIX time format)
 **			or
 **			   -1 on failure of action (for any reason)
 **
 **  NOTES:	The weekday value in the time of day clock is not used.
 **
 ****************/
time_t
r_mk48tdc()
{
	register unsigned char  *todc_ctl=TODC_ADDR;
	register unsigned char  *tp;
	register int  i,tmptim;
	int  seconds,minutes,hours,day,month,year;

	/*
	 * Read the Time of Day Clock registers and convert to standard
	 * representation.  Update of the clock registers must be inhibited
	 * before the read and re-enabled afterwards.
	 */
	ENTRY("r_mk48tdc");
	*todc_ctl |= READ;		/* inhibit clkreg updates/enable read */

	tp = todc_ctl + 1;
	for (i=1; i < 8; i++,tp++)
		todcreg[i] = *tp;

	*todc_ctl &= ~READ;	/* re-enable clkreg updates/disable read */

	seconds = ((todcreg[1] >>4) * 10) + (todcreg[1] & 0xf);
	minutes = ((todcreg[2] >>4) * 10) + (todcreg[2] & 0xf);
	hours = ((todcreg[3] >>4) * 10) + (todcreg[3] & 0xf);
	day = ((todcreg[5] >>4) * 10) + (todcreg[5] & 0xf);
	month = ((todcreg[6] >>4) * 10) + (todcreg[6] & 0xf);
	year = ((todcreg[7] >>4) * 10) + (todcreg[7] & 0xf);

	/* 
	 * Validate resulting time values and convert to seconds since
	 * 1 Jan 00:00:00 1970 GMT.
	 */
	if ((year < 70) || (month > 12) || (month < 1) || (day > 31) ||
	    (day < 1) || (hours > 23) || (minutes >59) || (seconds > 59))
		return(-1);

	tmptim=0;
	year += 1900;
	for(i= 1970; i < year; i++)
		tmptim += i%4 ? 365 : 366;
	if(!(year % 4) && month >= 3)
		tmptim += 1;
	while(--month)
		tmptim += dmsize[month -1];
	tmptim += (day - 1);
	tmptim *= 24;
	tmptim += hours;
	tmptim *= 60;
	tmptim += minutes;
	tmptim *= 60;
	tmptim += seconds;
	EXIT("r_mk48tdc");
	return(tmptim);

}  /* end r_mk48tdc */


/********************
 **  start_mk48tdc **
 ********************
 **
 **  FUNCTION:  Start the Time of Day clock on the MK48T02 chip.
 **
 **  CALLED:	From w_mk48tdc ONLY.
 **
 **  RETURNS:	none.
 ********************/
void
start_mk48tdc()
{
	register unsigned char  *todc_ctl=TODC_ADDR;
	register unsigned char  *tp;

	/*
	 * Starting the MK48T02/12 requires a "kick start" bit to be set
	 * and reset in addition to "turning off" the stop bit.
	 */
	ENTRY("start_mk48tdc");
	tp = todc_ctl + 1;
	*todc_ctl |= WRITE;	/* enable writing */
	*tp &= ~STOP;		/* turn stop bit "off" */
	tp += 2;
	*tp |= STOP;		/*turn on "kick start" */
	*todc_ctl &= ~WRITE; 	/* disable writing */
	delay(120);		/* wait 2 seconds */
	*todc_ctl |= WRITE;	/* enable writing */
	*tp &= ~STOP;		/* reset "kick start" */

	/*
	 * return to w_mk48tdc to set correct time - note that write is still 
	 * enabled to prevent register updates from oscillator until the time
	 * is correctly set.
	 */

	EXIT("start_mk48tdc");
}  /* end of start_mk48tdc */


#if (TDC_DBG >= 4)

/*******************
 **  stop_mk48tdc **
 *******************
 **
 **  FUNCTION:  Stop the clock oscillator to prolong life of battery.
 **		Intended for use during longterm storage.
 **
 **  CALLED:	never.  (Debugging purposes only)
 **
 **  RETURNS:	none.
 **
 *******************/
void
stop_mk48tdc()
{
	register unsigned char  *todc_ctl=TODC_ADDR;
	register unsigned char  *tp;

	ENTRY("stop_todc");
	tp = todc_ctl + 1;
	*todc_ctl |= WRITE;	/* enable writing */
	*tp |= STOP;		/* stop clock */
	*todc_ctl &= ~WRITE; 	/* disable writing */

#if (TDC_DBG == 4)
	printf("\tseconds reg. after STOP: %x\n", *tp);
#endif

	EXIT("stop_mk48tdc");
}  /* end of stop_mk48tdc */
#endif
