/*	@(#)kern-port:io/mvme204.c	35.3	*/
/*
 * MVME204-1/-2 DUAL PORTED DYNAMIC MEMORY VMEmodule
 * related routines.  Also see "sys/mvme204.h" for
 * further information.
 *
 * Extended to also handle
 * MVME224-1/-2 4mb/8mb DRAM MEMORY MODULES
 *
 * COPYRIGHT 1986,1988 BY MOTOROLA INC.
 */

/*
 * Note:
 * Unless specifically noted otherwise, all comments and references
 * to MVME204 or 204 apply also to the MVME224.
 */

#ident	"@(#)kern-port:io/mvme204.c	35.3"

#include "sys/param.h"
#include "sys/types.h"
#include "sys/utsname.h"
#include "sys/elog.h"
#include "sys/erec.h"
#include "sys/cmn_err.h"
#include "sys/mvme204.h"

/* The following values are based on the
 * VME 2016 TOWER, BOARD STRAPPING AND CONFIGURATION GUIDE (Rev B)
 * and the MVME224 User's Manual.
 * 
 * Unlike the MVME204 which has independent CSR and address selection
 * switches, the MVME224 CSR location is a function of the base
 * VMEbus address selection. The CSR addresses in the left column 
 * below support the MVME224 for all VMEbus addresses. The column
 * on the right is for up to 16 MVME204's. (MVME204's may actually
 * be set to any of the addresses below but it will be easier to
 * avoid CSR collision if the MVME204's are restricted to the addresses
 * in the right column.)
 * Since these addresses are a superset of the previous set,
 * this driver remains backward compatible with existing system
 * configurations. However, it is recommended that the addresses
 * ffffbe01, ffffbe03, ffffbe05 and ffffbe09 no longer be used
 * for MVME204 boards.
 *
 * static unsigned char *csr_loc[] = {
 *	(unsigned char *) 0xffffbe01, (unsigned char *) 0xffffbe03,
 *	(unsigned char *) 0xffffbe05, (unsigned char *) 0xffffbe07,
 *	(unsigned char *) 0xffffbe09, (unsigned char *) 0xffffbe0b,
 *	(unsigned char *) 0xffffbe0d, (unsigned char *) 0xffffbe0f,
 *	(unsigned char *) 0xffffbe11, (unsigned char *) 0xffffbe13,
 *			.			.
 *			.			.
 *			.			.
 *	(unsigned char *) 0xffffbe3d, (unsigned char *) 0xffffbe3f,
 *	(unsigned char *) 0xffffbe41, (unsigned char *) 0xffffbe43,
 *	(unsigned char *) 0xffffbe45, (unsigned char *) 0xffffbe47,
 *	(unsigned char *) 0xffffbe49, (unsigned char *) 0xffffbe4b,
 *	(unsigned char *) 0xffffbe4d, (unsigned char *) 0xffffbe4f,
 *	(unsigned char *) 0xffffbe51,
 *	(unsigned char *) 0xffffbe55,
 *	(unsigned char *) 0xffffbe59,
 *	(unsigned char *) 0xffffbe5d,
 *	(unsigned char *) 0xffffbe61,
 *			.
 *			.
 *			.
 *	(unsigned char *) 0xffffbef1,
 *	(unsigned char *) 0xffffbef5,
 *	(unsigned char *) 0xffffbef9,
 *	(unsigned char *) 0xffffbefd, NULL
 *};
 */

struct csr_range {
	unsigned char * csr_frst;
	unsigned char * csr_last;
	int increment;
};

static struct csr_range csr_loc[] = {
	{ (unsigned char *)0xffffbe01, (unsigned char *)0xffffbefd, 4 },
	{ (unsigned char *)0xffffbe03, (unsigned char *)0xffffbe3f, 4 },
	{ 0, 0, 0 }
};

#define CSR_CNT 20
#define READPROBE -1

/* The pointers to the MVME204 CSR's are held here;
 * the last one is followed by NULL pointers until the array is full.
 * This array limits the number of memory boards in a system to CSR_CNT.
 */
unsigned char *m204csr[CSR_CNT+1];

extern int parityon;
static int parityoff;

/* m204init()
 *
 * Locate MVME204-1/-2 CSR's and save their addresses in
 * the m204csr[] array by sensing each possible MVME204
 * CSR location; if no bus error occurs, then clear any
 * previous parity error indication, set enable parity
 * error report bit, and, save the address in the
 * m204csr[] array.  The last valid entry in the m204csr[]
 * array is NULL.
 */
void m204init()
{
	extern int bprobe();	/* returns non-zero on error */
	static char *cnt_msg =
"m2x4init: Too many MVME2x4 memory boards, expecting up to %d\n";
	static char *boards_msg =
"Parity checking enabled on %d MVME2x4 memory board(s)\n";

	register unsigned char syn, *csr;
	register struct csr_range *csrrp;
	int j, boards;

	if( parityoff = (!parityon) )		/* boot time option only */
		return;

	boards = j = 0;
	csrrp = csr_loc;
	for (;csr = csrrp->csr_frst; csrrp++) {
		for(;csr <= csrrp->csr_last; csr += csrrp->increment) {
			if (!bprobe(csr, READPROBE)) {
				if(j == CSR_CNT) { /* should never happen */
					m204csr[j] = NULL;

					cmn_err(CE_WARN, cnt_msg, CSR_CNT);
					return;
				}
				syn = *csr;
				*csr = (syn & ~M204_PE) | M204_EPER;
				m204csr[j++] = csr;
				boards++;
			}
		}
	}
	m204csr[j] = NULL;
	if ( boards )
		cmn_err(CE_NOTE, boards_msg, boards);
	return;
}


/* m204parity()
 *
 * Check csr's on MVME-204 memory boards for parity error conditions.
 * If there is one (and there should at most one) then:
 *	- If the parity error is not a repeat of the last parity condition,
 *	  i.e., the board, syndrome bits, or, error location is different
 *	  from the last parity error.
 *		* save this error's board, syndrome bits, and location
 *		* reset parity error instance counter to zero.
 *	- Inform the console of the error condition if the parity error
 *	  instance counter is equivalent to zero modulo M204_MINCNT.
 *	- Increment parity error instance counter.
 *	- Log the error.
 *	- if this was a user mode parity error, then, attempt to fix the error:
 *		* Disable bus interrupt on parity error.
 *		* Read, then, write the 4 bytes starting at the error location.
 *		* Re-enable bus interrupt on parity error.
 *		* Clear the csr parity error bit.
 * Return non-zero if any csr's indicate a parity error condition.
 */
/* mode argument values */
#define KERNEL 0
#define USER 1

int m204parity(loc, mode)
char *loc;
int mode;
{
	extern void logmemory();

	register unsigned char syn, *csr;
	register char *p;
	register int i, j, ch, status;

	static char *modename[] = { "Kernel", "User" };
	static char *parity_msg =
		"MVME2x4 %s Mode Parity Error; csr addr = 0x%x; value = 0x%x\n";

      static unsigned char *lcsr = (unsigned char *)-1;	/* last csr address */
	static unsigned char lsyn = 0;			/* last syndrome bits */
	static char *lloc = (char *)-1;			/* last err location */
	static int count = 0;				/* similiar counter */

	status = 0;
	if (parityoff)
		return(status);

	for (i = 0; (csr = m204csr[i]) != NULL; i++)
		if ((syn = *csr) & M204_PE) {
			if ((lcsr != csr) || (lsyn != syn) || (lloc != loc)) {
				lcsr = csr;
				lsyn = syn;
				lloc = loc;
				count = 0;
			}
			if ((count & M204_MINCNT) == 0)
				cmn_err(CE_WARN, parity_msg,
					modename[mode], csr, syn);
			count++;

			logmemory(syn, csr, E_UME);

			if (mode == USER) {
				*csr = syn & ~M204_EPER;
			  	/* start on divisible by 4 */
				p = (char *)((int)loc & ~3);
				for (j = 0; j < 4; j++) {
					ch = fubyte(p);
					subyte(p, ch);
					p++;
				}
				*csr = (syn & ~M204_PE) | M204_EPER;
			}
			status--;
		}
	return(status);
}
